/*
 * CacheEntry.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using TopCoder.Util.Idempotency.Persistence;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Entry in the <see cref="IdempotencyManager"/>'s <B>Distributed Simple Cache</B> which stores the
    /// last access time for each <see cref="Request.Identifier"/>, as well as a flag indicating if any
    /// responses for requests with that <B>Identifier</B> are currently available.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This class is intended for internal use, but is marked public so that custom implementations of the
    /// <see cref="ICachePersistence"/> interface can instantiate objects of this type.
    /// </para><para>
    /// <B>NOTE:</B> This class is NOT thread-safe since its properties must be modified.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [Serializable]
    public class CacheEntry 
    {
        /// <summary>
        /// A timestamp indicating the last time this entry was used (request received or reponse stored.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and should be updated every time the entry is used.
        /// </remarks>
        private DateTime lastActivity; 

        /// <summary>
        /// <B>true</B> if there is a response stored in persistence for this request; otherwise, <B>false</B>.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and should be updated as responses are stored or removed.
        /// </remarks>
        private bool inStorage; 

        /// <summary>
        /// Gets of Sets the last activity time of this entry.
        /// </summary>
        /// <value>The last time this entry was used.</value>
        public DateTime LastActivity
        {
            get
            {
                return lastActivity;
            }
            set
            {
                lastActivity = value;
            }
        }

        /// <summary>
        /// Gets or Sets a value indicating if there is a response for this request in persistent storage.
        /// </summary>
        /// <value><B>true</B> if a response exsists in storage; otherwise, <B>false</B>.</value>
        public bool IsInStorage
        {
            get
            {
                return inStorage;
            }
            set
            {
                inStorage = value;
            }
        }

        /// <summary>
        /// Creates a new <see cref="CacheEntry"/> who's last activity time is <see cref="DateTime.Now"/>
        /// and which is marked as not exsisting is persistent storage (yet.)
        /// </summary>
        /// <remarks>
        /// This is equivalent to <c>CacheEntry(DateTime.Now, false)</c>.
        /// </remarks>
        internal CacheEntry()
        {
            // Set the instance fields.
            lastActivity = DateTime.Now;
            inStorage = false;
        }

        /**
         * <p>Creates a <strong><em>CacheEntry</em></strong> with specified 
         * lastActivity and inStorage. This will probably be used by custom 
         * implementation of <strong><em>ICachePersistence</em></strong> and is 
         * left public.</p>
         * 
         * <p>This constructor simply sets the arguments to instance variables.</p>
         * 
         * @param lastActivity the last time cache activity is on for the request
         * @param inStorage whether any response exists in persistence for the request
         */
        /// <summary>
        /// Creates a new <see cref="CacheEntry"/> which was last accessed at <paramref name="lastActivity"/>
        /// and which exsists in persistent storage iff <paramref name="inStorage"/> is <B>true</B>.
        /// </summary>
        /// <param name="lastActivity">The <see cref="LastActivity"/> time for the new entry.</param>
        /// <param name="inStorage">The <see cref="IsInStorage"/> flag for the new entry.</param>
        public CacheEntry(DateTime lastActivity, bool inStorage)
        {
            // Set the instance fields.
            this.lastActivity = lastActivity;
            this.inStorage = inStorage;
        }
    }
}
