/*
 * CallbackEntry.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Entry in the callback list which is used to notify applications when new requests become available.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This is temporary work-around to compensate for the fact that <B>Distributed Simple Cache</B> does
    /// not provide notifications when items are added or updated by a remove computer.
    /// </para><para>
    /// <B>NOTE:</B> This class is thread-safe since all of its fields are immutable.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    internal class CallbackEntry
    {
        /// <summary>
        /// The request that is awaiting a response object. It is used to poll the 
        /// <B>Distributed Simple Cache</B>.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly Request request; 

        /// <summary>
        /// The delegate that will be called once the response becomes available.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly ResponseAvailable callback; 

        /// <summary>
        /// The time after which the polling thread should stop waiting for the response and drop
        /// this callback entry from the polling list.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly DateTime timeout; 

        /// <summary>
        /// Gets the request object that is waiting for a response to be added to the cache.
        /// </summary>
        /// <value>The request which is waiting for a response.</value>
        public Request Request
        {
            get
            {
                return request;
            }
        }

        /// <summary>
        /// Gets the delegate which will be called once a response is available.
        /// </summary>
        /// <value>The callback which will be called when a response is available.</value>
        public ResponseAvailable Callback
        {
            get
            {
                return callback;
            }
        }

        /// <summary>
        /// Gets the time after which this request will timeout and be removed from the waiting queue.
        /// </summary>
        /// <value>The time when the component will stop polling for a response to this request.</value>
        public DateTime Timeout
        {
            get
            {
                return timeout;
            }
        }

        /// <summary>
        /// Creates a new <see cref="CallbackEntry"/> with the given parameters and the maximum timout (never.)
        /// </summary>
        /// <remarks>
        /// This is equivalent to <c>CallbackEntry(request, callback, TimeSpan.MaxValue)</c>.
        /// </remarks>
        /// <param name="request">The request which is awaiting a response.</param>
        /// <param name="callback">The delegate to call when a response becomes available.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="callback"/> is <B>null</B>.</exception>
        internal CallbackEntry(Request request, ResponseAvailable callback)
            : this(request, callback, TimeSpan.MaxValue)
        {
            // Everything is done by the delegation to the full constructor.
        }

        /// <summary>
        /// Creates a new <see cref="CallbackEntry"/> with the given parameters.
        /// </summary>
        /// <param name="request">The request which is awaiting a response.</param>
        /// <param name="callback">The delegate to call when a response becomes available.</param>
        /// <param name="timeout">The amout of time, from now, that the request should timeout.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="callback"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="timeout"/> is negative.</exception>
        internal CallbackEntry(Request request, ResponseAvailable callback, TimeSpan timeout)
        {
            // Check the parameters.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (callback == null)
            {
                throw new ArgumentNullException("callback");
            }
            if (timeout < TimeSpan.Zero)
            {
                throw new ArgumentOutOfRangeException("timeout", timeout, "The timeout must not be negative.");
            }

            // Set the instance fields.
            this.request = request;
            this.callback = callback;
            try
            {
                // There is no good way to test if adding timeout will cause an overflow, so we must catch
                // that exception explicitly and then set the timout to DateTime.MaxValue in that case.
                this.timeout = DateTime.Now + timeout;
            }
            catch (ArgumentOutOfRangeException)
            {
                this.timeout = DateTime.MaxValue;
            }
        }
    }
}
