/*
 * ConfigHelper.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using System.Diagnostics;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Helper methods to put a nicer face on the <B>Configuration Manager</B> component.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    internal class ConfigHelper
    {
        /// <summary>
        /// Prive constructor prevents instantiation.
        /// </summary>
        private ConfigHelper()
        {
        }

        /// <summary>
        /// Gets the namespace by the given name.
        /// </summary>
        /// <param name="namespaceName">The name of the namespace to retrieve.</param>
        /// <returns>The configuration namespace for the given name.</returns>
        /// <exception cref="InvalidConfigFileException"><paramref name="namespaceName"/> does not exsist.</exception>
        internal static Namespace GetNamespace(string namespaceName)
        {
            // Check the argument (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(namespaceName != null, "The namespaceName cannot be null.");
            Debug.Assert(namespaceName != string.Empty, "The namespaceName cannot be empty.");

            // Get the proper namespace.
            Namespace ns = ConfigManager.GetInstance().GetNamespace(namespaceName);

            // If the namespace doesn't exsist, throw an exception.
            if (ns == null)
            {
                throw new InvalidConfigFileException(namespaceName + " was not found.");
            }

            return ns;
        }

        /// <summary>
        /// Gets a property value from the given namespace.
        /// </summary>
        /// <param name="ns">The namespace to get the property from.</param>
        /// <param name="propName">The name of the property to retrieve.</param>
        /// <param name="throwNoProp"><B>true</B> if an exception should be thrown when the property does
        /// not exsist; otherwise, <B>null</B> will be returned instead.</param>
        /// <returns>The property's <see cref="String.Trim"/>ed value or <B>null</B> if the property 
        /// does not exsist and <paramref name="throwNoProp"/> is <B>false</B>.</returns>
        /// <exception cref="InvalidConfigFileException">The property does not exsist and
        /// <paramref name="throwNoProp"/> is <B>true</B>.</exception>
        internal static string GetValue(Namespace ns, string propName, bool throwNoProp)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");

            // Get the property.
            Property prop = ns.GetProperty(propName);

            // Does the property exsist?
            if (prop == null)
            {
                // If not, either throw and exception or just return null, depending on the throwNoProp flag.
                if (throwNoProp)
                {
                    throw new InvalidConfigFileException(propName + " was not found in " + ns.Name);
                }
                else
                {
                    return null;
                }
            }

            // If the property exsists, its value should never be null.
            Debug.Assert(prop.Value != null, "Property.Value should never be null.");

            // Return the value, trimming any whitespace in the process.
            return prop.Value.Trim();
        }

        /// <summary>
        /// Gets an array of property values from the given namespace.
        /// </summary>
        /// <param name="ns">The namespace to get the property from.</param>
        /// <param name="propName">The name of the property to retrieve.</param>
        /// <param name="throwNoProp"><B>true</B> if an exception should be thrown when the property does
        /// not exsist; otherwise, <B>null</B> will be returned instead.</param>
        /// <returns>An array of the property's <see cref="String.Trim"/>ed values or <B>null</B> if the 
        /// property does not exsist and <paramref name="throwNoProp"/> is <B>false</B>.</returns>
        /// <exception cref="InvalidConfigFileException">The property does not exsist and
        /// <paramref name="throwNoProp"/> is <B>true</B>.</exception>
        internal static string[] GetValues(Namespace ns, string propName, bool throwNoProp)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");

            // Get the property.
            Property prop = ns.GetProperty(propName);

            // Does the property exsist?
            if (prop == null)
            {
                // If not, either throw and exception or just return null, depending on the throwNoProp flag.
                if (throwNoProp)
                {
                    throw new InvalidConfigFileException(propName + " was not found in " + ns.Name);
                }
                else
                {
                    return null;
                }
            }

            // Get the values.
            string[] values = prop.Values;

            // If the property exsists, its value should never be null.
            Debug.Assert(values != null, "Property.Values should never be null.");
            Debug.Assert(values.Length > 0, "Property.Values should always contain at least one element.");

            // Trim all of the values before returning them.
            for (int i = 0; i < values.Length; ++i)
            {
                // None of the values should ever be null.
                Debug.Assert(values[i] != null, "Property.Values elements should never be null.");

                // Trim any whitespace from this value.
                values[i] = values[i].Trim();
            }

            // Return the values.
            return values;
        }

        /// <summary>
        /// Gets a property value from the given namespace and parses it as a <B>double</B>.
        /// </summary>
        /// <param name="ns">The namespace to get the property from.</param>
        /// <param name="propName">The name of the property to retrieve.</param>
        /// <param name="defaultValue">The value to return if the property does not exsist.</param>
        /// <returns>The parsed property value or the <paramref name="defaultValue"/> if the
        /// property does not exsist.</returns>
        /// <exception cref="InvalidConfigFileException">The property could not be parsed as a double.</exception>
        internal static double GetDoubleValue(Namespace ns, string propName, double defaultValue)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");

            // Get the string representation of this proprety.
            string value = ConfigHelper.GetValue(ns, propName, false);

            // If the property does not exsist, return the defult value.
            if (value == null)
            {
                return defaultValue;
            }

            // Parse the value as a double, wrapping any exceptions which are thrown in the process.
            try
            {
                return Double.Parse(value);
            }
            catch (Exception e)
            {
                throw new InvalidConfigFileException("The value of " + propName + " in " + ns.Name + 
                    " could not be parsed as a double.", e);
            }
        }

        /// <summary>
        /// Gets a property value from the given namespace and parses it as a <B>bool</B>.
        /// </summary>
        /// <param name="ns">The namespace to get the property from.</param>
        /// <param name="propName">The name of the property to retrieve.</param>
        /// <param name="defaultValue">The value to return if the property does not exsist.</param>
        /// <returns>The parsed property value or the <paramref name="defaultValue"/> if the
        /// property does not exsist.</returns>
        /// <exception cref="InvalidConfigFileException">The property could not be parsed as a bool.</exception>
        internal static bool GetBoolValue(Namespace ns, string propName, bool defaultValue)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");

            // Get the string representation of this proprety.
            string value = ConfigHelper.GetValue(ns, propName, false);

            // If the property does not exsist, return the defult value.
            if (value == null)
            {
                return defaultValue;
            }

            // Parse the value as a double, wrapping any exceptions which are thrown in the process.
            try
            {
                return Boolean.Parse(value);
            }
            catch (Exception e)
            {
                throw new InvalidConfigFileException("The value of " + propName + " in " + ns.Name + 
                    " could not be parsed as a bool.", e);
            }
        }

        /// <summary>
        /// Gets a property value from the given namespace and parses it as an <B>enum</B>.
        /// </summary>
        /// <param name="ns">The namespace to get the property from.</param>
        /// <param name="propName">The name of the property to retrieve.</param>
        /// <param name="enumType">The type of enum to parse the value as.</param>
        /// <returns>The parsed property value.</returns>
        /// <exception cref="InvalidConfigFileException">The property could not be parsed as an enum.</exception>
        internal static object GetEnumValue(Namespace ns, string propName, Type enumType)
        {
            // Create a dummy value for the default value.
            object dummy = new object();

            // Get the enum value (or null if its not defined.)
            object result = GetEnumValue(ns, propName, enumType, dummy);

            // If the dummy value was returned, the parameter was not defined, so throw an exception.
            if (result == dummy)
            {
                throw new InvalidConfigFileException(propName + " was not found in " + ns.Name);
            }

            // Since the result isn't our dummy value, return it.
            return result;
        }

        /// <summary>
        /// Gets a property value from the given namespace and parses it as an <B>enum</B>.
        /// </summary>
        /// <param name="ns">The namespace to get the property from.</param>
        /// <param name="propName">The name of the property to retrieve.</param>
        /// <param name="enumType">The type of enum to parse the value as.</param>
        /// <param name="defaultValue">The value to return if the property does not exsist.</param>
        /// <returns>The parsed property value or the <paramref name="defaultValue"/> if the
        /// property does not exsist.</returns>
        /// <exception cref="InvalidConfigFileException">The property could not be parsed as an enum.</exception>
        internal static object GetEnumValue(Namespace ns, string propName, Type enumType, object defaultValue)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");
            Debug.Assert(enumType != null, "The enumType cannot be null.");
            Debug.Assert(defaultValue != null, "The defaultValue cannot be null.");

            // Get the string representation of this proprety.
            string value = ConfigHelper.GetValue(ns, propName, false);

            // If the property does not exsist, return the defult value.
            if (value == null)
            {
                return defaultValue;
            }

            // Parse the value as a double, wrapping any exceptions which are thrown in the process.
            object result;
            try
            {
                // Try to parse the property's value into a CacheDecision value.
                result = Enum.Parse(enumType, value);
            }
            catch (Exception e)
            {
                // We should get here if parsing fails.
                throw new InvalidConfigFileException("The value of " + propName + " in " + ns.Name + 
                    " could not be parsed as a " + enumType.Name + " value.", e);
            }

            // If the parsed value is not one of the named constants, throw an exception.
            if (!Enum.IsDefined(enumType, result))
            {
                throw new InvalidConfigFileException("The value of " + propName + " in " + ns.Name +
                    " is not one of the defined " + enumType.Name + " values.");
            }

            // If we get this far, the result is good, so just return it.
            return result;
        }

        /// <summary>
        /// Instantiates a class based on the class name specified in a configuration file.
        /// </summary>
        /// <remarks>
        /// This method gets the <paramref name="propName"/> <see cref="Property"/> from the given
        /// <see cref="Namespace"/>. It then converts the class name into a <see cref="Type"/>
        /// object and checks to make sure that the type is derived from <paramref name="baseType"/>.
        /// Finally, an object of this type is instantiated and returned.
        /// </remarks>
        /// <param name="ns">The configuration namespace used to lookup the property value.</param>
        /// <param name="propName">The name of the property which contains the class name.</param>
        /// <param name="baseType">The specified type must inherit from this base type.</param>
        /// <param name="throwNoProp"><B>true</B> indicates an exception should be thrown if the property is not found;
        ///     <B>false</B> indicates that <B>null</B> will simply be returned in this case.</param>
        /// <returns>The newly instantiated object.</returns>
        /// <exception cref="InvalidConfigFileException">
        /// <paramref name="propName"/> was not found in <paramref name="ns"/> 
        /// and <paramref name="throwNoProp"/> is <B>true</B>.<BR/>
        /// -or-<BR/>
        /// <paramref name="propName"/> is a blank string.<BR/>
        /// -or-<BR/>
        /// <paramref name="propName"/> is not a valid <see cref="Type"/>.<BR/>
        /// -or-<BR/>
        /// The type is not derrived from <paramref name="baseType"/>.<BR/>
        /// -or-<BR/>
        /// An exception occured while instantiating the type (see the inner exception for details.)
        /// </exception>
        internal static object CreateTypesafeInstance(Namespace ns, string propName, Type baseType, bool throwNoProp)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");
            Debug.Assert(baseType != null, "The baseType cannot be null.");

            // Get the value (class name).
            string typeName = ConfigHelper.GetValue(ns, propName, throwNoProp);

            // Return early if typeName is null (the property did not exsist and throwNoProp was false.)
            if (typeName == null)
            {
                return null;
            }

            // Create a new instance of this type and return it.
            return CreateTypesafeInstance(ns, propName, typeName, baseType);
        }

        /// <summary>
        /// Instantiates a list of classes based on the class names specified in a configuration file.
        /// </summary>
        /// <remarks>
        /// This method gets the <paramref name="propName"/> <see cref="Property"/> from the given
        /// <see cref="Namespace"/>. It then converts the class names into <see cref="Type"/>
        /// objects and checks to make sure that the types are derived from <paramref name="baseType"/>.
        /// Finally, objects of these types are instantiated and returned.
        /// </remarks>
        /// <param name="ns">The configuration namespace used to lookup the property value.</param>
        /// <param name="propName">The name of the property which contains the class name.</param>
        /// <param name="baseType">The specified type must inherit from this base type.</param>
        /// <param name="throwNoProp"><B>true</B> indicates an exception should be thrown if the property is not found;
        ///     <B>false</B> indicates that <B>null</B> will simply be returned in this case.</param>
        /// <returns>The newly instantiated object.</returns>
        /// <exception cref="InvalidConfigFileException">
        /// <paramref name="propName"/> was not found in <paramref name="ns"/> 
        /// and <paramref name="throwNoProp"/> is <B>true</B>.<BR/>
        /// -or-<BR/>
        /// <paramref name="propName"/> is a blank string.<BR/>
        /// -or-<BR/>
        /// <paramref name="propName"/> is not a valid <see cref="Type"/>.<BR/>
        /// -or-<BR/>
        /// The types are not derrived from <paramref name="baseType"/>.<BR/>
        /// -or-<BR/>
        /// An exception occured while instantiating the types (see the inner exception for details.)
        /// </exception>
        internal static IList CreateTypesafeInstances(Namespace ns, string propName, Type baseType, bool throwNoProp)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");
            Debug.Assert(baseType != null, "The baseType cannot be null.");

            // Get the values (class names).
            string[] typeNames = ConfigHelper.GetValues(ns, propName, throwNoProp);

            // Return early if typeNames is null (the property did not exsist and throwNoProp was false.)
            if (typeNames == null)
            {
                return null;
            }

            // Create a new list of objects to return.
            IList objects = new ArrayList(typeNames.Length);

            // Instantiate each object and add it to the list.
            for (int i = 0; i < typeNames.Length; ++i)
            {
                objects.Add(CreateTypesafeInstance(ns, propName + "[" + i + "]", typeNames[i], baseType));
            }

            // Return the full list.
            return objects;
        }

        /// <summary>
        /// Instantiates a class based on the class name specified in a configuration file.
        /// </summary>
        /// <remarks>
        /// This method converts the <paramref name="typeName"/> into a <see cref="Type"/>
        /// object and checks to make sure that the type is derived from <paramref name="baseType"/>.
        /// Finally, an object of this type is instantiated and returned.
        /// </remarks>
        /// <param name="ns">The configuration namespace (used to generate helpful error messages.)</param>
        /// <param name="propName">The name of the property which contains the class name (for error messages.)</param>
        /// <param name="typeName">The fully-qualified name of the type to instantiate.</param>
        /// <param name="baseType">The specified type must inherit from this base type.</param>
        /// <returns>The newly instantiated object.</returns>
        /// <exception cref="InvalidConfigFileException">
        /// <paramref name="typeName"/> is a blank string.<BR/>
        /// -or-<BR/>
        /// <paramref name="typeName"/> is not a valid <see cref="Type"/>.<BR/>
        /// -or-<BR/>
        /// The type is not derrived from <paramref name="baseType"/>.<BR/>
        /// -or-<BR/>
        /// An exception occured while instantiating the type (see the inner exception for details.)
        /// </exception>
        private static object CreateTypesafeInstance(Namespace ns, string propName, string typeName, Type baseType)
        {
            // Check the arguments (this is used internally, so Debug.Assert is sufficient.)
            Debug.Assert(ns != null, "The ns cannot be null.");
            Debug.Assert(propName != null, "The propName cannot be null.");
            Debug.Assert(typeName != null, "The typeName cannot be null.");
            Debug.Assert(baseType != null, "The baseType cannot be null.");

            // If the class name is empty (after trimming), throw an exception.
            if (typeName == string.Empty)
            {
                throw new InvalidConfigFileException(propName + " in " + ns.Name + " is empty.");
            }

            // Convert the class name into a Type.
            Type type = Type.GetType(typeName);

            // If the type was not found, throw an exception.
            if (type == null)
            {
                throw new InvalidConfigFileException("Unable to load the type specified in the " + 
                    propName + " property of " + ns.Name);
            }

            // If the specified type is not derived from baseType, throw an exception.
            if (!baseType.IsAssignableFrom(type))
            {
                throw new InvalidConfigFileException("The type specified in the " + propName +
                    " property of " + ns.Name + " must be derived from " + baseType.FullName);
            }

            // Create and return an instance of the type. If this fails for any reason, the exception
            // is wrapped in an InvalidConfigFileException and then rethrown.
            try
            {
                return Activator.CreateInstance(type);
            }
            catch (Exception e)
            {
                throw new InvalidConfigFileException("Unable to instantiate " + type.FullName, e);
            }
        }
    }
}
