/*
 * GenericCacheManagement.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using System.Diagnostics;
using TopCoder.Cache.Distributed;
using TopCoder.Util.Idempotency.Persistence;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Implementation of the <see cref="ICacheManagement"/> interface.
    /// </summary>
    /// <remarks>
    /// The included operations may be fairly expensive, so they should not be used very often. This
    /// implementation is thread-safe since all of the operations are stateless.
    /// </remarks>
    public class GenericCacheManagement : ICacheManagement
    {
        /// <summary>
        /// Removes entries from the cache in an attempt to reduce the entry count to <paramref name="entryCount"/>.
        /// </summary>
        /// <remarks>
        /// The entries are removed in order by timestamp, from earliest to latest. This implementation
        /// is a best-effort attempt to reduce the cache to the given size. However, if enough entries
        /// are updates and/or added during this operation, it may not be possible. If the resulting size
        /// is not small enough, the caller may call this method again to attempt to reduce the size furthur.
        /// </remarks>
        /// <param name="cache">The cache to remove entries from.</param>
        /// <param name="entryCount">The target number of entries to reduce the cache to.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="entryCount"/> is negative.</exception>
        public void ShrinkCacheToSize(IDistributedSimpleCache cache, int entryCount)
        {
            // Check the parameters.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }
            if (entryCount < 0)
            {
                throw new ArgumentOutOfRangeException("entryCount", entryCount, "Must be non-negative.");
            }

            // Get a list of all the keys.
            ICollection keys = cache.Keys;

            // Figure out how many entries must be removed.
            int toRemove = keys.Count - entryCount;

            // If the cache is already small enough, quit early.
            if (toRemove <= 0)
            {
                return;
            }

            // Create a new array list for the keys. It will contain DictionaryEntries mapping cache keys to
            // the LastActivity values of their entries in the cache.
            ArrayList keyList = new ArrayList(keys.Count);

            // Loop through all the keys to get a copy of their last activity time. Simply making a copy of the
            // keys collection will not work since there is no way to lock the cache while we sort the keys.
            // Instead, we will save the current activity timestamp and use that for sorting purposed. We can
            // then use the ArrayList.Sort() method and take care of any updates after the sort is complete.
            foreach (string key in keys)
            {
                // Get the associated entry.
                CacheEntry entry = (CacheEntry) cache.Get(key);

                // If the entry still exsists, save the last activity time (so it does change while sorting.)
                if (entry != null)
                {
                    keyList.Add(new DictionaryEntry(key, entry.LastActivity));
                }
            }

            // Sort the list of key/activity elements (from lest-recent activity to most-recent.)
            keyList.Sort(new EntryComparer());

            // Loop through the items in the key list, removing any we can until we have removed enough.
            foreach (DictionaryEntry item in keyList)
            {
                // Get the associated entry.
                CacheEntry entry = (CacheEntry) cache.Get(item.Key);

                // If the entry still exsists, try and remove it.
                if (entry != null)
                {
                    // Since we can't lock things, we remove the entry first and hope for the best.
                    cache.Remove(item.Key);

                    // See if the entry was updated after we stored the LastActivity value in keyList.
                    if (entry.LastActivity != (DateTime) item.Value)
                    {
                        // If so, we should put it back in the cache (if locking were possible, we could
                        // check first and only remove if this check passed, but we can't so we have to
                        // remove--hoping for the best--and then put the entry back if we screwed up.)
                        cache.Put(item.Key, entry);
                    }
                    else
                    {
                        // If not, it was OK to remove this entry, so update the remove count.
                        --toRemove;

                        // If we've removed all that we needed to, then stop.
                        if (toRemove == 0)
                        {
                            return;
                        }
                    }
                }
            }

            // At this point, we've removed everything we can, but the more than entryCount items have been
            // accessed since we got the initial list of keys. We could try running this method again (and
            // again and again), but we may never finish and its easy enough for the application to call
            // this method again it its not satisfied with our "best-effort" results.
        }

        /// <summary>
        /// <see cref="IComparer"/> for a list of <B>string</B>/<B>DateTime</B> pairs by their <B>DateTime</B>.
        /// </summary>
        private class EntryComparer : IComparer
        {
            /// <summary>
            /// Compares two <see cref="DictionaryEntry"/>s which map identifier strings to <see cref="DateTime"/>s.
            /// </summary>
            /// <param name="x">The first entry.</param>
            /// <param name="y">The second entry.</param>
            /// <returns>A negative number if x's <B>DateTime</B> is &lt; y's; zero if they are equal; and
            /// positive if y's <B>DateTime</B> is less than x's.</returns>
            public int Compare(object x, object y)
            {
                // Make sure the objects are both DictionaryEntries with DateTime values.
                Debug.Assert(x is DictionaryEntry, "The first object must be a DictionaryEntry.");
                Debug.Assert(((DictionaryEntry) x).Value is DateTime, "The first object's value must be a DateTime.");

                Debug.Assert(y is DictionaryEntry, "The second object must be a DictionaryEntry.");
                Debug.Assert(((DictionaryEntry) y).Value is DateTime, "The second object's value must be a DateTime.");

                // Get the last activity time for each list entry.
                DateTime lastX = (DateTime) ((DictionaryEntry) x).Value;
                DateTime lastY = (DateTime) ((DictionaryEntry) y).Value;

                // Delegate to the DateTime.CompareTo method.
                return lastX.CompareTo(lastY);
            }
        }

        /// <summary>
        /// Removes any entries from the cache that have timestamps before <paramref name="timestamp"/>.
        /// </summary>
        /// <param name="cache">The cache to remove entries from.</param>
        /// <param name="timestamp">The earliest timestamp that should be kept in the cache.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        public void ShrinkCacheToDate(IDistributedSimpleCache cache, DateTime timestamp)
        {
            // Check the arguments.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }

            // Loop through all the keys, removing any who's access time is too old.
            foreach (string key in cache.Keys)
            {
                // Get the associated entry.
                CacheEntry entry = (CacheEntry) cache.Get(key);

                // If the entry still exsists but is older than 'timestamp', remove it.
                if ((entry != null) && (entry.LastActivity < timestamp))
                {
                    cache.Remove(key);
                }
            }
        }

        /// <summary>
        /// Removes any entries from persistent storage that do not have matching entries in the cache
        /// (since they could not possibly be accessed by this component anymore.)
        /// </summary>
        /// <param name="cache">The cache who's entries must exsist to access the stored responses.</param>
        /// <param name="responsePersistence">The response persistence to purge useless data from.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="responsePersistence"/> is <B>null</B>.</exception>
        public void PurgeResponsePersistence(IDistributedSimpleCache cache, IResponsePersistence responsePersistence)
        {
            // Check the arguments.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }
            if (responsePersistence == null)
            {
                throw new ArgumentNullException("responsePersistence");
            }

            // Check each identifier in persistent storage.
            IEnumerator it = responsePersistence.GetIdentifiers();
            while (it.MoveNext())
            {
                string id = (string) it.Current;

                // If the indentifier is not found in the cache, remove it from persistent storage.
                if (!cache.ContainsKey(id))
                {
                    responsePersistence.Delete(id);
                }
            }
        }
    }
}
