/*
 * ICacheManagement.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using TopCoder.Cache.Distributed;
using TopCoder.Util.Idempotency.Persistence;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Interface which provides some cache management functions.
    /// </summary>
    /// <remarks>
    /// The included operations may be fairly expensive, so they should not be used very often. It is
    /// recommended that implementations of this interface be thread-safe, however, this is not a strict
    /// requirement since the <see cref="IdempotencyManager"/> currently locks the cache before
    /// performing any operations.
    /// </remarks>
    public interface ICacheManagement 
    {
        /// <summary>
        /// Removes entries from the cache in an attempt to reduce the entry count to <paramref name="entryCount"/>.
        /// </summary>
        /// <remarks>
        /// The entries should be removed in order by timestamp, from earliest to latest.
        /// </remarks>
        /// <param name="cache">The cache to remove entries from.</param>
        /// <param name="entryCount">The target number of entries to reduce the cache to.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="entryCount"/> is negative.</exception>
        void ShrinkCacheToSize(IDistributedSimpleCache cache, int entryCount);

        /// <summary>
        /// Removes any entries from the cache that have timestamps before <paramref name="timestamp"/>.
        /// </summary>
        /// <param name="cache">The cache to remove entries from.</param>
        /// <param name="timestamp">The earliest timestamp that should be kept in the cache.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        void ShrinkCacheToDate(IDistributedSimpleCache cache, DateTime timestamp);

        /// <summary>
        /// Removes any entries from persistent storage that do not have matching entries in the cache
        /// (since they could not possibly be accessed by this component anymore.)
        /// </summary>
        /// <param name="cache">The cache who's entries must exsist to access the stored responses.</param>
        /// <param name="responsePersistence">The response persistence to purge useless data from.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="responsePersistence"/> is <B>null</B>.</exception>
        void PurgeResponsePersistence(IDistributedSimpleCache cache, IResponsePersistence responsePersistence);
    }
}
