/*
 * IdempotencyManager.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Threading;
using System.Diagnostics;
using TopCoder.Util.ConfigurationManager;
using TopCoder.Cache.Distributed;
using TopCoder.Util.Collection.List.LinkedList;
using TopCoder.Util.Idempotency.Logger;
using TopCoder.Util.Idempotency.Strategy;
using TopCoder.Util.Idempotency.Persistence;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Facade which exposes the functionality of the entire component to an application.
    /// </summary>
    /// <remarks>
    /// <para>
    /// When processing a request, an application should first create a <see cref="Request"/> wrapper
    /// for the request and use as a parameter to the <see cref="RequestArrival"/> method to signal
    /// the request's arrival and then use the <see cref="HasResponse"/> method to determine if there
    /// is already an response cached for this type of request.
    /// </para><para>
    /// If there is already a request cached, it may be retrieved by calling the <see cref="RetrieveResponse"/>
    /// method. Otherwise, the application should process the request normally and then can cache the
    /// result for future requests using the <see cref="StoreResponse"/> method.
    /// </para><para>
    /// This class also provides administrative functionality for the request cache and underlying response storage.
    /// </para><para>
    /// <B>NOTE:</B> This class is made thread-safe by locking the static instance during operations so that
    /// a consistent state is maintained throughout the operation. If multiple operations must be performed by
    /// an application atomically with this class, the instance could be locked externally to force this behavior.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class IdempotencyManager : IDisposable
    {
        /// <summary>
        /// The cache used to store a mapping from request identifiers to <see cref="CacheEntry"/>s.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter. The items stored in the
        /// cache, however, are mutable and may even be changed by a remote computer if this component
        /// is used in a distributed environment.
        /// </remarks>
        private readonly DistributedSimpleCache cache;

        /// <summary>
        /// The <see cref="ICacheManagement"/> instance used for management operations.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor but may be changed by the application at runtime. It may
        /// never be <B>null</B>, however.
        /// </remarks>
        private ICacheManagement cacheManagement;

        /// <summary>
        /// The <see cref="ICacheStrategy"/> used to determine if requests should be cached.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor, but may be changed by the application at runtime. It may
        /// also be set to <B>null</B> to signal that all requests may be cached.
        /// </remarks>
        private ICacheStrategy cacheStrategy;

        /// <summary>
        /// The <see cref="ICachePersistence"/> implementation used to save and load the request cache.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor, but may be changed by the application at runtime. It may
        /// also be set to <B>null</B>, in which case, the cache cannot be saved or loaded.
        /// </remarks>
        private ICachePersistence cachePersistence;

        /// <summary>
        /// The <see cref="IResponsePersistence"/> implementation used to store response objects.
        /// </summary>
        /// <remarks>
        /// This variable is set in the constructor but may be changed by the application at runtime. It may
        /// never be <B>null</B>, however.
        /// </remarks>
        private IResponsePersistence responsePersistence;

        /// <summary>
        /// The <see cref="IActivityLogger"/> implementation used to log status notifications.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor, but may be changed by the application at runtim. It may
        /// also be set to <B>null</B> to disable logging.
        /// </remarks>
        private IActivityLogger activityLogger;

        /// <summary>
        /// A list of <see cref="CallbackEntry"/>s, currently waiting for responses to be added to the cache.
        /// </summary>
        /// <remarks>
        /// <para>
        /// This value is set in the constructor and is immutable thereafter. The contents of the list,
        /// however, are modified. The asynchronous <see cref="RetrieveResponse"/> methods add entries to
        /// this list and the polling thread removes these entries when their responses are found or
        /// their timeouts expire.
        /// </para><para>
        /// All operations on this list lock <c>callbackList.SyncRoot</c> to ensure thread-safety.
        /// </para>
        /// </remarks>
        private readonly ILinkedList callbackList;

        /// <summary>
        /// Signals the polling thread if it should continue to run.
        /// </summary>
        /// <remarks>
        /// This value is initialized to <B>true</B> in the constructor and is changed to <B>false</B> in the
        /// <see cref="Dispose"/> method to signal the polling thread that it should stop running.
        /// </remarks>
        private bool threadRunning;

        /// <summary>
        /// The amount of time the polling thread should sleep between polling cycles.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly TimeSpan pollingInterval;

        /// <summary>
        /// The current instance of this class.
        /// </summary>
        /// <remarks>
        /// This class is not strictly a singleton since its possible to create multiple instances. The
        /// <see cref="Dispose"/> method sets this value back to null, allowing a new instance to be created.
        /// All old instances, however, must have been <B>Disposed</B> already, so only one of the instances
        /// of this class is ever useful at any given time.
        /// </remarks>
        private static IdempotencyManager instance = null;

        /// <summary>
        /// Gets the static instance of the <see cref="IdempotencyManager"/> class (or creates a new one.)
        /// </summary>
        /// <value>The current <B>IdempotencyManager</B> instance.</value>
        public static IdempotencyManager Instance
        {
            get
            {
                // Lock the type for static thread-safety.
                lock (typeof(IdempotencyManager))
                {
                    // Lazily create the instance and return it.
                    if (instance == null)
                    {
                        instance = new IdempotencyManager();
                    }

                    return instance;
                }
            }
        }

        /// <summary>
        /// Gets or Sets the <see cref="ICacheManagement"/> object used with this instance.
        /// </summary>
        /// <value>The current cache manager.</value>
        /// <exception cref="ArgumentNullException"><paramref name="value"/> is <B>null</B>.</exception>
        public ICacheManagement CacheManagement
        {
            get
            {
                lock (this)
                {
                    return cacheManagement;
                }
            }
            set
            {
                // This property cannot be set to null.
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                lock (this)
                {
                    cacheManagement = value;
                }
            }
        }

        /// <summary>
        /// Gets or Sets the <see cref="ICacheStrategy"/> object used with this instance.
        /// </summary>
        /// <value>The current cacheing strategy (may be <B>null</B>, indicating no filtering is done.)</value>
        public ICacheStrategy CacheStrategy
        {
            get
            {
                lock (this)
                {
                    return cacheStrategy;
                }
            }
            set
            {
                lock (this)
                {
                    cacheStrategy = value;
                }
            }
        }

        /// <summary>
        /// Gets or Sets the <see cref="ICachePersistence"/> object used with this instance.
        /// </summary>
        /// <value>The current cache storage (may be <B>null</B>, but then the cache cannot be stored.)</value>
        public ICachePersistence CachePersistence
        {
            get
            {
                lock (this)
                {
                    return cachePersistence;
                }
            }
            set
            {
                lock (this)
                {
                    cachePersistence = value;
                }
            }
        }

        /// <summary>
        /// Gets or Sets the <see cref="IResponsePersistence"/> object used with this instance.
        /// </summary>
        /// <value>The current response storage object.</value>
        /// <exception cref="ArgumentNullException"><paramref name="value"/> is <B>null</B>.</exception>
        public IResponsePersistence ResponsePersistence
        {
            get
            {
                lock (this)
                {
                    return responsePersistence;
                }
            }
            set
            {
                // This property cannot be set to null.
                if (value == null)
                {
                    throw new ArgumentNullException("value");
                }

                lock (this)
                {
                    responsePersistence = value;
                }
            }
        }

        /// <summary>
        /// Gets or Sets the <see cref="IActivityLogger"/> object used with this instance.
        /// </summary>
        /// <value>The current activity logger (may be <B>null</B>, indicating no logging is done.)</value>
        public IActivityLogger ActivityLogger
        {
            get
            {
                lock (this)
                {
                    return activityLogger;
                }
            }
            set
            {
                lock (this)
                {
                    activityLogger = value;
                }
            }
        }

        /// <summary>
        /// Creates a new <see cref="IdempotencyManager"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// <para>
        /// Configuration values should appear in the <c>TopCoder.Util.Idempotency.IdempotencyManager</c> 
        /// namespace. This namespace must contain a <c>cache_url</c> parameter and may contain a
        /// <c>start_url</c> parameter which will be used to construct the internal <B>Distributed Simple Cache</B>.
        /// </para><para>
        /// The configuration may contain a <c>cache_management</c> property which is used to instantiate the
        /// <see cref="CacheManagement"/> object. If this property is absent, then a 
        /// <see cref="GenericCacheManagement"/> object is created and used for this field. A
        /// <c>response_persistence</c> value is also required, but there is no fallback default for this parameter.
        /// </para><para>
        /// The <c>cache_strategy</c>, <c>cache_persistence</c>, and <c>activity_logger</c> parameters are
        /// all optional and may be used to specify the fully-qualified name of types to instantiate and
        /// use for these fields, but are set to <B>null</B> if these properties do not exsist.
        /// </para><para>
        /// Finally, a polling interval for the polling thread may be specified with the <c>polling_interval</c>
        /// property. The value should be a time in milliseconds to wait between polling cycles. If this
        /// parameter is not specified, then a value of 5 seconds will be used (5000ms.)
        /// </para><para>
        /// Once all the configuration is complete, the polling thread is started before construction is complete.
        /// </para>
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        private IdempotencyManager()
        {
            // Get the configuration namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Read the required cache_url property.
            string cacheUrl = ConfigHelper.GetValue(ns, "cache_url", true);

            // Read the start_url property (null wil be returned if the property does not exsist.)
            string startUrl = ConfigHelper.GetValue(ns, "start_url", false);

            // Attempt to create a new distributed cache.
            try
            {
                cache = new DistributedSimpleCache(cacheUrl, startUrl);
            }
            catch (Exception e)
            {
                throw new InvalidConfigFileException("Unable to create a DistributedSimpleCache with " +
                    "the cache_url and start_url configuration values in " + ns.Name, e);
            }

            // Read the cache_management property and use it to construct a new object
            // (returns null if the property does not exsist.)
            cacheManagement = (ICacheManagement) ConfigHelper.CreateTypesafeInstance(
                ns, "cache_management", typeof(ICacheManagement), false);

            // If the property was not defined, we default to use a GenericCacheManagement instance.
            if (cacheManagement == null)
            {
                cacheManagement = new GenericCacheManagement();
            }

            // Read the cache_strategy property and use it to construct a new object
            // (returns null if the property does not exsist.)
            cacheStrategy = (ICacheStrategy) ConfigHelper.CreateTypesafeInstance(
                ns, "cache_strategy", typeof(ICacheStrategy), false);

            // Read the cache_persistence property and use it to construct a new object
            // (returns null if the property does not exsist.)
            cachePersistence = (ICachePersistence) ConfigHelper.CreateTypesafeInstance(
                ns, "cache_persistence", typeof(ICachePersistence), false);

            // Read the response_persistence property and use it to construct a new object.
            responsePersistence = (IResponsePersistence) ConfigHelper.CreateTypesafeInstance(
                ns, "response_persistence", typeof(IResponsePersistence), true);

            // Read the activity_logger property and use it to construct a new object
            // (returns null if the property does not exsist.)
            activityLogger = (IActivityLogger) ConfigHelper.CreateTypesafeInstance(
                ns, "activity_logger", typeof(IActivityLogger), false);

            // Read the polling_interval property and parses its value as a double. If the property does not
            // exsist, then the default timeout of 5000 is returned instead.
            pollingInterval = TimeSpan.FromMilliseconds(ConfigHelper.GetDoubleValue(ns, "polling_interval", 5000));

            // Create a new callback list, set the polling thread's state to running, and then start the thread.
            callbackList = new DoubleLinkedList();
            threadRunning = true;
            new Thread(new ThreadStart(this.ExamineCallbackList));
        }

        /// <summary>
        /// Determines if a request should be rejected (not cached.)
        /// </summary>
        /// <remarks>
        /// This method should only be called inside a <c>lock (this) { ... }</c> block, since no
        /// locking is performed in this method (since it would be largely redundant.)
        /// </remarks>
        /// <param name="request">The request to check.</param>
        /// <returns><B>true</B> if there is a cache strategy that does not accept this request;
        /// otherwise, <B>false</B>.</returns>
        private bool ShouldReject(Request request)
        {
            return (cacheStrategy != null) && (!cacheStrategy.Accept(request));
        }

        /// <summary>
        /// Signals the arrival of a request.
        /// </summary>
        /// <remarks>
        /// This method should be called when the application recieves a request. If logging is enabled, the
        /// activity will be logged. If the request is accepted by the cache strategy (or there is no cache
        /// strategy), then the request is added to the cache or its entry is updated. If an entry already exsisted,
        /// then the time since its last access is returned; otherwise <see cref="TimeSpan.MaxValue"/> is returned.
        /// </remarks>
        /// <param name="request">The request which has just arrived.</param>
        /// <returns>The time since this request was last used (or <B>TimeSpan.MaxValue</B> if it is new.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        public TimeSpan RequestArrival(Request request)
        {
            // Check the argument.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            lock (this)
            {
                // If logging is enabled, log this request.
                if (activityLogger != null)
                {
                    activityLogger.RequestReceived(request);
                }

                // If the request should not be accepted, return early.
                if (ShouldReject(request))
                {
                    return TimeSpan.MaxValue;
                }

                // Get cached information about previous requests of this type (if any.)
                CacheEntry entry = (CacheEntry) cache.Get(request.Identifier);

                if (entry != null)
                {
                    // Calculate elapsed time since the last request for this identifier.
                    DateTime now = DateTime.Now;
                    TimeSpan elapsed = now - entry.LastActivity;

                    // Update the entry and then return the elapsed time.
                    entry.LastActivity = now;
                    return elapsed;
                }
                else
                {
                    // No match was found in the cache, so add a new entry and return TimeSpan.MaxValue.
                    cache.Put(request.Identifier, new CacheEntry(DateTime.Now, false));
                    return TimeSpan.MaxValue;
                }
            }
        }

        /// <summary>
        /// Determines if there are any responses stored with the identifier of the given request.
        /// </summary>
        /// <remarks>
        /// Just because there is a match does not mean a <see cref="RetrieveResponse"/> call will succede. If
        /// the stored response and the given request have different security keys, this method will still
        /// return <B>true</B>, but a <see cref="ResponseRetrievalException"/> will be thrown if an attempt
        /// is made to access the response.
        /// </remarks>
        /// <param name="request"></param>
        /// <returns><B>true</B> if there is at least one response in storage who's identifier matches
        /// the given request's ID; otherwise, <B>false</B>.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        public bool HasResponse(Request request)
        {
            // Check the argument.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            lock (this)
            {
                // If the request should not be accepted, return early.
                if (ShouldReject(request))
                {
                    return false;
                }

                // Get the cache entry (if any.)
                CacheEntry entry = (CacheEntry) cache.Get(request.Identifier);

                // Return true if there is an entry and it has a response in storage; otherwise, false.
                return (entry != null) && (entry.IsInStorage);
            }
        }

        /// <summary>
        /// Retrieves the stored response which matches both the identifier and security key of the given request.
        /// </summary>
        /// <remarks>
        /// If there is a caching strategy which rejects this request, <B>null</B> will be returned; otherwise,
        /// an attempt is made to retrieve the response from the <see cref="ResponsePersistence"/>. If this
        /// succedes and there is an activity logger, the action is logged.
        /// </remarks>
        /// <param name="request">The request who's response is desired.</param>
        /// <returns>The response for this request, or <B>null</B> if this request is not accepted.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        /// <exception cref="ResponseRetrievalException">No matching responses were found.</exception>
        /// <exception cref="PersistenceException">Some other error occured with the persistent storage.</exception>
        public Response RetrieveResponse(Request request)
        {
            // Check the argument.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            lock (this)
            {
                // If the request should not be accepted, return early.
                if (ShouldReject(request))
                {
                    return null;
                }

                // Get the response from persistent storage (this will throw an ResponseRetrievalException
                // if there is no matching response currently saved in the storage.)
                Response response = responsePersistence.Retrieve(request);

                // If logging is enabled, log the retreival before returning the response.
                if (activityLogger != null)
                {
                    activityLogger.ResponseRetrieved(request, response);
                }

                return response;
            }
        }

        /// <summary>
        /// Retrieves the stored response which matches both the identifier and security key of the given request.
        /// </summary>
        /// <remarks>
        /// If there is a caching strategy which rejects this request, <B>null</B> will be returned; otherwise,
        /// an attempt is made to retrieve the response from the <see cref="ResponsePersistence"/>. If this
        /// succedes and there is an activity logger, the action is logged. An attempt is made to access the
        /// response every 100ms until either a response if retrieved without errors, or the <paramref name="timout"/>
        /// expires.
        /// </remarks>
        /// <param name="request">The request who's response is desired.</param>
        /// <param name="timeout">The amout of time to keep trying to retrieve a response.</param>
        /// <returns>The response for this request, or <B>null</B> if this request is not accepted.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="timeout"/> is negative.</exception>
        /// <exception cref="ResponseRetrievalException">No matching responses were found.</exception>
        /// <exception cref="PersistenceException">Some other error occured with the persistent storage.</exception>
        public Response RetrieveResponse(Request request, TimeSpan timeout)
        {
            // Check the arguments.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (timeout < TimeSpan.Zero)
            {
                throw new ArgumentOutOfRangeException("timout", timeout, "The timeout must be non-negative.");
            }

            // Calculate the stopping time.
            DateTime endTime = DateTime.Now + timeout;

            // Loop until its the ending time.
            while (DateTime.Now < endTime)
            {
                lock (this)
                {
                    // If the request should not be accepted, return early.
                    if (ShouldReject(request))
                    {
                        return null;
                    }

                    // If there is a response now, return it.
                    if (HasResponse(request))
                    {
                        return RetrieveResponse(request);
                    }
                }

                // Sleep a little before the next try.
                Thread.Sleep(100);
            }

            // If we get here, time ran out without getting the response. We'll try retreiving the
            // response one last time, but this will likely throw a ResponseRetrievalException.
            return RetrieveResponse(request);
        }

        /// <summary>
        /// Asychronously retrieves the stored response which matches 
        /// both the identifier and security key of the given request.
        /// </summary>
        /// <remarks>
        /// This method is equivalent to <c>RetrieveResponse(request, callback, TimeSpan.MaxValue)</c>.
        /// </remarks>
        /// <param name="request">The request who's response is desired.</param>
        /// <param name="callback">The delegate to call when the response if finally found.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="callback"/> is <B>null</B>.</exception>
        public void RetrieveResponse(Request request, ResponseAvailable callback)
        {
            RetrieveResponse(request, callback, TimeSpan.MaxValue);
        }

        /// <summary>
        /// Asychronously retrieves the stored response which matches 
        /// both the identifier and security key of the given request.
        /// </summary>
        /// <remarks>
        /// First, a check is made to see if the response is already available. If it is, the callback is called
        /// directly and, if no exceptions are thrown, the method simply returns. If the response is not available
        /// or an exception occures while attempting the callback, a new <see cref="CallbackEntry"/> is created
        /// and added to the callback list. This entry will be processed the next time the polling thread wakes
        /// up to process pending requests.
        /// </remarks>
        /// <param name="request">The request who's response is desired.</param>
        /// <param name="callback">The delegate to call when the response if finally found.</param>
        /// <param name="timeout">The amount of time to wait for the response before giving up.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="callback"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="timeout"/> is negative.</exception>
        public void RetrieveResponse(Request request, ResponseAvailable callback, TimeSpan timeout)
        {
            // Check arguments.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (callback == null)
            {
                throw new ArgumentNullException("callback");
            }
            if (timeout < TimeSpan.Zero)
            {
                throw new ArgumentOutOfRangeException("timeout", timeout, "The timeout must be non-negative.");
            }

            lock (this)
            {
                // If this request should not be accepted, return early.
                if (ShouldReject(request))
                {
                    return;
                }

                // First, check to see if there is already a response that we can use. This is a minor
                // optimization so we don't have to wait for the next polling cycle if we don't need to.
                try
                {
                    if (HasResponse(request))
                    {
                        // Get the response and use it to call the callback.
                        callback(request, RetrieveResponse(request));
                        return;
                    }
                }
                catch
                {
                    // Ignore any exceptions.
                }

                // Either the response does not exsist yet, or an exception was thrown while retrieving
                // the response or calling the callback. In either case, we must add an entry to the
                // callbackList and let the polling thread handle it from here.
                lock (callbackList.SyncRoot)
                {
                    callbackList.Add(new CallbackEntry(request, callback, timeout));
                }
            }
        }

        /// <summary>
        /// Stores a response to the given request in persistent storage for later use if the request is repeated.
        /// </summary>
        /// <remarks>
        /// If there is a caching strategy which rejects this request, the method simply returns; otherwise,
        /// an attempt is made to store the response in the <see cref="ResponsePersistence"/>. If this
        /// succedes and there is an activity logger, the action is logged. Finally, if the proper entry does
        /// not exsist in the cache, it is created so that the response can be retrieved later.
        /// </remarks>
        /// <param name="request">The request who's response is being stored.</param>
        /// <param name="response">The response to store.</param>
        /// <exception cref="ArgumentNullException"><paramref name="reqeust"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">An error occured while trying to store the response.</exception>
        public void StoreResponse(Request request, Response response)
        {
            // Check the arguments.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            lock (this)
            {
                // If we should not accept this request, return early.
                if (ShouldReject(request))
                {
                    return;
                }

                // Store the response to persistent storage.
                responsePersistence.Store(request, response);

                // If logging is enabled, log the store operation.
                if (activityLogger != null)
                {
                    activityLogger.ResponseStored(request, response);
                }

                // Get the cached entry (if any.)
                CacheEntry entry = (CacheEntry) cache.Get(request.Identifier);

                // If there is no entry in the cache yet, create a new one.
                if (entry == null)
                {
                    entry = new CacheEntry();
                }

                // Update the activity time and persistent storage flag.
                entry.LastActivity = DateTime.Now;
                entry.IsInStorage = true;

                // Update the cached entry.
                cache.Put(request.Identifier, entry);
            }
        }

        /// <summary>
        /// Removes all entries from the cache and persistent storage with the given identifier.
        /// </summary>
        /// <param name="identifier">The identifier who's responses should be deleted.</param>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">An error occured while trying to delete the entries.</exception>
        public void DropEntry(string identifier)
        {
            // Check the argument.
            if (identifier == null)
            {
                throw new ArgumentNullException("identifier");
            }
            if (identifier == string.Empty)
            {
                throw new ArgumentException("The empty string is not a valid identifier.", "identifier");
            }

            lock (this)
            {
                // Remove from persistent storage.
                responsePersistence.Delete(identifier);

                // Remove from the cache.
                cache.Remove(identifier);
            }
        }

        /// <summary>
        /// Clears all entries from the cache and persistent storage.
        /// </summary>
        /// <exception cref="PersistenceException">An error occured while trying to clear the entries.</exception>
        public void ClearCache()
        {
            lock (this)
            {
                // Clear persistent storage.
                responsePersistence.Clear();

                // Clear the cache.
                cache.Clear();
            }
        }

        /// <summary>
        /// Loads the cache from an external persistent storage.
        /// </summary>
        /// <exception cref="PersistenceException">The <see cref="CachePersistence"/> is <B>null</B>.<BR/>-or-<BR/>
        /// An error occured while trying to load the entries.</exception>
        public void LoadCache()
        {
            lock (this)
            {
                // Make sure we have a cachePersistence to delegate to.
                if (cachePersistence == null)
                {
                    throw new PersistenceException("CachePersistence is null.");
                }

                // Delegate to cachePersistence.
                cachePersistence.LoadCache(cache);
            }
        }

        /// <summary>
        /// Saves the cache to an external persistent storage.
        /// </summary>
        /// <exception cref="PersistenceException">The <see cref="CachePersistence"/> is <B>null</B>.<BR/>-or-<BR/>
        /// An error occured while trying to save the entries.</exception>
        public void SaveCache()
        {
            lock (this)
            {
                // Make sure we have a cachePersistence to delegate to.
                if (cachePersistence == null)
                {
                    throw new PersistenceException("CachePersistence is null.");
                }

                // Delegate to cachePersistence.
                cachePersistence.SaveCache(cache);
            }
        }

        /// <summary>
        /// Attempts to shrink the cache to the given size, using the <see cref="CacheManagement"/>.
        /// </summary>
        /// <param name="entryCount">The desired number of entries in the cache.</param>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="entryCount"/> is negative.</exception>
        public void ShrinkCacheToSize(int entryCount)
        {
            lock (this)
            {
                // Delegate to cacheManagement.
                cacheManagement.ShrinkCacheToSize(cache, entryCount);
            }
        }

        /// <summary>
        /// Shrinks the cache by removing all entries which haven't been used since the given timestamp, 
        /// using the <see cref="CacheManagement"/>.
        /// </summary>
        /// <param name="timestamp">The cutoff point before which responses should not be kept.</param>
        public void ShrinkCacheToDate(DateTime timestamp)
        {
            lock (this)
            {
                // Delegate to cacheMangement.
                cacheManagement.ShrinkCacheToDate(cache, timestamp);
            }
        }

        /// <summary>
        /// Cleans up persistent storage by removing any responses that are no longer useful since they have
        /// no matching entries in the cache, using the <see cref="CacheManagement"/>.
        /// </summary>
        /// <exception cref="PersistenceException">If removing outdated responses from storage fails.</exception>
        public void PurgeResponsePersistence()
        {
            lock (this)
            {
                // Delegate to cacheManagement.
                cacheManagement.PurgeResponsePersistence(cache, responsePersistence);
            }
        }

        /// <summary>
        /// Disposes of any resources used by the <see cref="IdempotencyManager"/> and sets the static instance
        /// field back to <B>null</B>.
        /// </summary>
        /// <remarks>
        /// Disposes of any of the fields which are not <B>null</B>, as well as signals the polling thread
        /// to stop process and disposes the cache. The object should not be used after this method is called.
        /// </remarks>
        public void Dispose()
        {
            // Get the instance lock.
            lock (typeof(IdempotencyManager))
            {
                // Only the current instance should be disposed.
                Debug.Assert(IdempotencyManager.instance == this, "Should only dispose the current instance.");

                // This object is no longer the default instance.
                IdempotencyManager.instance = null;
            }

            // Stop the polling thread (eventually.)
            threadRunning = false;

            // Prevent modifications to the instance properties.
            lock (this)
            {
                // responsePersistence should never be null.
                Debug.Assert(responsePersistence != null, "responsePersistence should never be null.");

                // Dispose of responsePersistence.
                responsePersistence.Dispose();

                // Dispose of cachePersistence if there is one.
                if (cachePersistence != null)
                {
                    cachePersistence.Dispose();
                }

                // Dispose of activityLogger if there is one.
                if (activityLogger != null)
                {
                    activityLogger.Dispose();
                }

                // Dispose of the cache (this will disconnect it from the rest of the peers and then clear it.)
                cache.Dispose();
            }
        }

        /// <summary>
        /// Polling thread worker method which continuously checks the cache for new responses that
        /// have been requested.
        /// </summary>
        /// <remarks>
        /// Continues processing until the <see cref="IdempotencyManager"/> is disposed. Each polling cycle
        /// consists of iterating through the list of <see cref="CallbackEntry"/>s waiting for their responses.
        /// Each entry is checked to see if its response is available yet and, if so, the associated callback
        /// is called. If no response is available and the entry's timout has expired, it is simply removed
        /// from the waiting list. Once all the entries have been processed, the thread sleeps for the
        /// specified polling interval before repeating the polling process.
        /// </remarks>
        private void ExamineCallbackList()
        {
            // Run until we are signaled to stop.
            while (threadRunning)
            {
                // Prevent concurrent modifications while iterating.
                lock (callbackList.SyncRoot)
                {
                    // Make sure the iterator is reset.
                    callbackList.Reset();

                    // Loop through all the callback entries.
                    while (callbackList.MoveNext())
                    {
                        // Get the current entry.
                        CallbackEntry entry = (CallbackEntry) callbackList.Current;

                        // Check to see if there is a response to this request yet.
                        bool hasResponse = HasResponse(entry.Request);

                        // If there is a response or if the timeout has expired, remove this entry.
                        if (hasResponse || (entry.Timeout >= DateTime.Now))
                        {
                            callbackList.Remove();

                            // If there is no response (the timeout just expired) then skip to the next entry.
                            if (!hasResponse)
                            {
                                continue;
                            }
                        }

                        // Allow exceptions to be ignored.
                        try
                        {
                            // Get the response and use it to call the callback.
                            // (Logging is not needed here since RetrieveResponse will perform the proper logging.)
                            entry.Callback(entry.Request, RetrieveResponse(entry.Request));
                        }
                        catch
                        {
                            // Ignore any exceptions.
                        }
                    }
                }

                // Wait until the next polling time.
                Thread.Sleep(pollingInterval);
            }
        }
    }
}
