/*
 * ActivityBroadcaster.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// <see cref="IActivityLogger"/> implementation which forwards notifications to a list of other
    /// <see cref="IActivityLogger"/>s.
    /// </summary>
    /// <remarks>
    /// <para>
    /// <B>NOTE:</B> This class is inherently thread-safe since its internal <see cref="IList"/> is immutable
    /// and all <see cref="IActivityLogger"/>s in that list should be thread-safe as well.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class ActivityBroadcaster : IActivityLogger 
    {
        /// <summary>
        /// A list of <see cref="IActivityLogger"/>s to forward notifications to.
        /// </summary>
        /// <remarks>
        /// This value is set by the constructor and is immutable thereafter.
        /// </remarks>
        private readonly IList activityLoggers;

        /// <summary>
        /// Creates a new <see cref="ActivityBroadcaster"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// Configuration values should appear in the <c>TopCoder.Util.Idempotency.Logger.ActivityBroadcaster</c>
        /// namespace. This namespace must contain an <c>activity_loggers</c> parameter which should contain
        /// a list of fully-qualified type names of <see cref="IActivityLogger"/> classes. These classes must
        /// all have default constructors which will be used to instantiate the class. All the newly instantiated
        /// objects will be added to a list of loggers that will be invoked during each event notification.
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        public ActivityBroadcaster()
        {
            // Get the proper namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Create the list of activity loggers.
            activityLoggers = ConfigHelper.CreateTypesafeInstances(
                ns, "activity_loggers", typeof(IActivityLogger), true);
        }

        /// <summary>
        /// Creates a new <see cref="ActivityBroadcaster"/> instance based on the given list of loggers.
        /// </summary>
        /// <remarks>
        /// To prevent external modifications an ensure object immutability, a shallow copy of the
        /// <paramref name="activityLoggers"/> parameter is made to fill the internal list of loggers.
        /// </remarks>
        /// <param name="activityLoggers">A list of non-<B>null</B> <see cref="IActivityLogger"/>s.</param>
        /// <exception cref="ArgumentNullException"><paramref name="activityLoggers"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="activityLoggers"/> is empty.<BR/>-or-<BR/>
        /// <paramref name="activityLoggers"/> contains a <B>null</B> or non-<see cref="IActivityLogger"/> 
        /// instance.</exception>
        public ActivityBroadcaster(IList activityLoggers)
        {
            // Check argument.
            if (activityLoggers == null)
            {
                throw new ArgumentNullException("activityLoggers");
            }

            // Copy the parameter to prevent external modifications.
            this.activityLoggers = new ArrayList(activityLoggers);

            // Verify that the list has at least one logger.
            if (this.activityLoggers.Count <= 0)
            {
                throw new ArgumentException("The list of loggers must not be empty.", "activityLoggers");
            }

            // Make sure all elements in the list are non-null IActivityLoggers.
            for (int i = 0; i < this.activityLoggers.Count; ++i)
            {
                if (!(this.activityLoggers[i] is IActivityLogger))
                {
                    throw new ArgumentException("List items must be non-null IActivityLoggers.", 
                        "activityLoggers[" + i + "]");
                }
            }
        }

        /// <summary>
        /// Calls <c>RequestReceived(request)</c> on each of the loggers in the internal list.
        /// </summary>
        /// <remarks>
        /// No argument checking is done by this method, but nested loggers will likely throw
        /// <see cref="ArgumentNullException"/> if <paramref name="request"/> is <B>null</B>.
        /// </remarks>
        /// <param name="request">The request to log.</param>
        public void RequestReceived(Request request)
        {
            foreach (IActivityLogger logger in activityLoggers)
            {
                logger.RequestReceived(request);
            }
        }

        /// <summary>
        /// Calls <c>ResponseStored(request, response)</c> on each of the loggers in the internal list.
        /// </summary>
        /// <remarks>
        /// No argument checking is done by this method, but nested loggers will likely throw
        /// <see cref="ArgumentNullException"/> if either <paramref name="request"/> or <paramref name="response"/>
        /// are <B>null</B>.
        /// </remarks>
        /// <param name="request">The request to log.</param>
        /// <param name="response">The response that was stored for the request.</param>
        public void ResponseStored(Request request, Response response)
        {
            foreach (IActivityLogger logger in activityLoggers)
            {
                logger.ResponseStored(request, response);
            }
        }

        /// <summary>
        /// Calls <c>ResponseRetrieved(request, response)</c> on each of the loggers in the internal list.
        /// </summary>
        /// <remarks>
        /// No argument checking is done by this method, but nested loggers will likely throw
        /// <see cref="ArgumentNullException"/> if either <paramref name="request"/> or <paramref name="response"/>
        /// are <B>null</B>.
        /// </remarks>
        /// <param name="request">The request to log.</param>
        /// <param name="response">The response that was retrieved for the request.</param>
        public void ResponseRetrieved(Request request, Response response)
        {
            foreach (IActivityLogger logger in activityLoggers)
            {
                logger.ResponseRetrieved(request, response);
            }
        }

        /// <summary>
        /// Calls <c>ExceptionOccured(request, exception)</c> on each of the loggers in the internal list.
        /// </summary>
        /// <remarks>
        /// No argument checking is done by this method, but nested loggers will likely throw
        /// <see cref="ArgumentNullException"/> if either <paramref name="request"/> or <paramref name="exception"/>
        /// are <B>null</B>.
        /// </remarks>
        /// <param name="request">The request to log.</param>
        /// <param name="exception">The exception that was thrown.</param>
        public void ExceptionOccurred(Request request, Exception exception)
        {
            foreach (IActivityLogger logger in activityLoggers)
            {
                logger.ExceptionOccurred(request, exception);
            }
        }

        /// <summary>
        /// Calls <c>Disponse()</c> on each of the loggers in the internal list.
        /// </summary>
        public void Dispose()
        {
            foreach (IActivityLogger logger in activityLoggers)
            {
                logger.Dispose();
            }
        }
    }
}
