/*
 * ActivityLogging.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Diagnostics;
using TopCoder.LoggingWrapper;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// <see cref="IActivityLogger"/> implementation which logs notifications using the <B>Logging Wrapper</B>.
    /// </summary>
    /// <remarks>
    /// <para>
    /// Initialization of the <B>Logging Wrapper</B> is the application's responsibility. The logging format,
    /// however, is a configurable parameter. It is also configurable whether security keys should be allowed
    /// to appear in log messages or not.
    /// </para><para>
    /// <B>NOTE:</B> This class is inherently thread-safe since all its instance variables are immutable,
    /// and the <B>Logging Wrapper</B> itself is thread-safe.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class ActivityLogging : IActivityLogger
    {
        /// <summary>
        /// The format of the log message.
        /// </summary>
        /// <remarks>
        /// This value is set by the constructor and is immutable thereafter.
        /// </remarks>
        private readonly string format;

        /// <summary>
        /// <B>true</B> if security keys are allowed to appear in log messages; otherwise, <B>false</B>.
        /// </summary>
        /// <remarks>
        /// This value is set by the constructor and is immutable thereafter.
        /// </remarks>
        private readonly bool logSecurityKey;

        /// <summary>
        /// Creates a new <see cref="ActivityLogging"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// Configuration values should appear in the <c>TopCoder.Util.Idempotency.Logger.ActivityLogging</c>
        /// namespace. This namespace must contain a <c>format</c> and <c>log_security_key</c> parameter. The
        /// <c>log_security_key</c>'s value shoule be parsable as a boolean value--either <c>"true"</c> or
        /// <c>"false"</c> (case-insensitive.) If this parameter is <B>true</B>, then the <c>format</c> parameter
        /// should be a format string (sutable for use in <see cref="String.Format"/>) which may contain upto 
        /// 4 parameter references (<see cref="Request.Identifier"/>, <see cref="Request.SecurityKey"/>, 
        /// <see cref="Request.Type"/>, and <c>DateTime.Now</c>.) If the <c>log_security_key</c> parameter 
        /// is <B>false</B>, then the <B>SecurityKey</B> parameter is omitted and the format string may have 
        /// only 3 parameter references.
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        public ActivityLogging()
        {
            // Get the proper namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Get the format string.
            format = ConfigHelper.GetValue(ns, "format", true);

            // Get the logSecurityKey value.
            logSecurityKey = ConfigHelper.GetBoolValue(ns, "log_security_key", true);

            // Test the format string.
            Exception e = TestFormatString();

            // If the format test didn't work, wrap the exception and rethrow it.
            if (e != null)
            {
                throw new InvalidConfigFileException(typeof(ActivityLogging).FullName + 
                    ".format is not a valid format string for the given parameters.", e);
            }
        }

        /// <summary>
        /// Creates a new <see cref="ActivityLogging"/> instance that may log security keys and uses
        /// the given formatting string.
        /// </summary>
        /// <remarks>
        /// This constructor is equivalent to <c>ActivityLogging(<paramref name="format"/>, <B>true</B>)</c>.
        /// </remarks>
        /// <param name="format">The string format to use for log messages.</param>
        /// <exception cref="ArgumentNullException"><paramref name="format"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="format"/> is invalid.</exception>
        public ActivityLogging(string format)
            : this(format, true)
        {
            // Do nothing (the full constructor does all the work.)
        }

        /// <summary>
        /// Creates a new <see cref="ActivityLogging"/> instance that may or may not log security keys 
        /// and uses the given formatting string.
        /// </summary>
        /// <remarks>
        /// If <paramref name="logSecurityKey"/> is <B>true</B>, then <paramref name="format"/>
        /// should be a format string (sutable for use in <see cref="String.Format"/>) which may contain upto 
        /// 4 parameter references (<see cref="Request.Identifier"/>, <see cref="Request.SecurityKey"/>, 
        /// <see cref="Request.Type"/>, and <c>DateTime.Now</c>.) If the <paramref name="logSecurityKey"/> 
        /// is <B>false</B>, then the <B>SecurityKey</B> parameter is omitted and the format string may have 
        /// only 3 parameter references.
        /// </remarks>
        /// <param name="format">The string format to use for log messages.</param>
        /// <param name="logSecurityKey"><B>true</B> if security keys may be logged; otherwise, <B>false</B>.</param>
        /// <exception cref="ArgumentNullException"><paramref name="format"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="format"/> is invalid.</exception>
        public ActivityLogging(string format, bool logSecurityKey)
        {
            // Check the parameters.
            if (format == null)
            {
                throw new ArgumentNullException("format");
            }

            // Set the instance fields.
            this.format = format;
            this.logSecurityKey = logSecurityKey;

            // Test the format string.
            Exception e = TestFormatString();

            // If the format test didn't work, wrap the exception and rethrow it.
            if (e != null)
            {
                throw new ArgumentException("The format is not valid with the given parameters.", "format", e);
            }
        }

        /// <summary>
        /// Tests the current parameters using <see cref="String.Format"/> with dummy argument to ensure that
        /// <see cref="format"/> is a valid formatting string based on the value of <see cref="logSecurityKey"/>.
        /// </summary>
        /// <remarks>
        /// This method is called by the constructors to validate the format string before any logging.</remarks>
        /// <returns><B>null</B> if the format is valid; 
        /// otherwise, the exception throw while attempting the test formatting.</returns>
        private FormatException TestFormatString()
        {
            try
            {
                // Test the format string (with or without a security key parameter.)
                if (logSecurityKey)
                {
                    String.Format(format, "id", "key", "type", DateTime.Now);
                }
                else
                {
                    String.Format(format, "id", "type", DateTime.Now);
                }

                // Formatting worked.
                return null;
            }
            catch (FormatException e)
            {
                return e;
            }
        }

        /// <summary>
        /// Write a message to the <B>Logging Wrapper</B>.
        /// </summary>
        /// <remarks>
        /// If <see cref="logSecurityKey"/> is <B>true</B>, then the <see cref="format"/> string is used
        /// with all 4 logging parameters; otherwise, only 3 parameters are used. <B>format</B> is first
        /// modified by adding the prefix and postfix strings to create the full formatting string.
        /// </remarks>
        /// <param name="level">The severity of the logging message.</param>
        /// <param name="prefix">A string to prepend to the logged message.</param>
        /// <param name="postfix">A string to append to the logged message.</param>
        /// <param name="request">The request to log.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        private void Log(Level level, string prefix, string postfix, Request request)
        {
            // Check arguments (this method is private, so Debug.Asserts work just fine.)
            Debug.Assert(Enum.IsDefined(typeof(Level), level), "The level must be a named constant.");
            Debug.Assert(prefix != null, "The prefix may be empty, but not null.");
            Debug.Assert(postfix != null, "The postfix may be empty, but not null.");

            // Check the request parameter (since it is passed straight into the method.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            // Construct the message.
            string message = prefix + format + postfix;

            if (logSecurityKey)
            {
                // Log with the security key.
                LogManager.Log(level, message, request.Identifier, request.SecurityKey, request.Type, DateTime.Now);
            }
            else
            {
                // Log without the security key.
                LogManager.Log(level, message, request.Identifier, request.Type, DateTime.Now);
            }
        }

        /// <summary>
        /// Logs a request-received event at log-level <see cref="Level.INFO"/> with a <c>"RECEIVED: "</c> prefix.
        /// </summary>
        /// <param name="request">The request to log.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        public void RequestReceived(Request request)
        {
            Log(Level.INFO, "RECEIVED: ", "", request);
        }

        /// <summary>
        /// Logs a response-stored event at log-level <see cref="Level.INFO"/> with a <c>"STORED: "</c> prefix.
        /// </summary>
        /// <param name="request">The request to log.</param>
        /// <param name="response">The response that is being stored (Ignored.)</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        public void ResponseStored(Request request, Response response)
        {
            // Check the response parameter.
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            Log(Level.INFO, "STORED: ", "", request);
        }

        /// <summary>
        /// Logs a response-retrieved event at log-level <see cref="Level.INFO"/> with a <c>"RETRIEVED: "</c> prefix.
        /// </summary>
        /// <param name="request">The request to log.</param>
        /// <param name="response">The response that was retrieved (Ignored.)</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        public void ResponseRetrieved(Request request, Response response)
        {
            // Check the response parameter.
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            Log(Level.INFO, "RETRIEVED: ", "", request);
        }

        /// <summary>
        /// Logs an exception that occured at log-level <see cref="Level.ERROR"/> with an <c>"ERROR: "</c> prefix
        /// and <see cref="Exception.Message"/> postfix.
        /// </summary>
        /// <param name="request">The request to log.</param>
        /// <param name="exception">The exception that was thrown.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="exception"/> is <B>null</B>.</exception>
        public void ExceptionOccurred(Request request, Exception exception)
        {
            // Check the exception parameter.
            if (exception == null)
            {
                throw new ArgumentNullException("exception");
            }

            Log(Level.ERROR, "ERROR: ", exception.Message, request);
        }

        /// <summary>
        /// Does nothing (no cleanup needed in this implementation.)
        /// </summary>
        public void Dispose()
        {
            // Do nothing.
        }
    }
}
