/*
 * ActivityStatistics.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using System.Diagnostics;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// <see cref="IActivityLogger"/> implementation which stores notification statistics in in-memory collections.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This is a rather simple statistics plugin with the following limitations:
    /// <list type="bullet">
    ///     <item>No external persistence is used which means the statistics are lost when the program exits.</item>
    ///     <item>Since the <see cref="IdempotencyManager"/> only logs local actions, the statistics will
    ///         represent only the activity within the current application domain. It distributed statistics
    ///         are desired, then an application must either provide a custom <see cref="IActivityLogger"/>
    ///         which can be used in a distributed manner, or simply save the statistics from each location
    ///         and then combine the result to obtain global statistics.</item>
    /// </list>
    /// </para><para>
    /// <B>NOTE:</B> Thread-safety is accomplished by locking the individual internal collections when they are used.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class ActivityStatistics : IActivityLogger 
    {
        /// <summary>
        /// Maps request identifiers to a count of the number of times that ID has been received.
        /// </summary>
        /// <remarks>
        /// All operations which access or modify this collection should lock its <B>SyncRoot</B>.
        /// </remarks>
        private IDictionary receiveCounts = new Hashtable();

        /// <summary>
        /// Maps request identifiers to a count of the number of times that ID has been stored.
        /// </summary>
        /// <remarks>
        /// All operations which access or modify this collection should lock its <B>SyncRoot</B>.
        /// </remarks>
        private IDictionary storeCounts = new Hashtable();

        /// <summary>
        /// Maps request identifiers to a count of the number of times that ID has been retrieved.
        /// </summary>
        /// <remarks>
        /// All operations which access or modify this collection should lock its <B>SyncRoot</B>.
        /// </remarks>
        private IDictionary retrieveCounts = new Hashtable();

        /// <summary>
        /// Gets the current receive count statistics.
        /// </summary>
        /// <value>
        /// Copy of the internal mapping from request identifier to an integer count of the number of times
        /// a request with that identifier has been received.
        /// </value>
        public IDictionary ReceiveCounts
        {
            get
            {
                lock (receiveCounts.SyncRoot)
                {
                    return new Hashtable(receiveCounts);
                }
            }
        }

        /// <summary>
        /// Gets the current store count statistics.
        /// </summary>
        /// <value>
        /// Copy of the internal mapping from request identifier to an integer count of the number of times
        /// a request with that identifier has been stored.
        /// </value>
        public IDictionary StoreCounts
        {
            get
            {
                lock (storeCounts.SyncRoot)
                {
                    return new Hashtable(storeCounts);
                }
            }
        }

        /// <summary>
        /// Gets the current retrieve count statistics.
        /// </summary>
        /// <value>
        /// Copy of the internal mapping from request identifier to an integer count of the number of times
        /// a request with that identifier has been retrieved.
        /// </value>
        public IDictionary RetrieveCounts
        {
            get
            {
                lock (retrieveCounts.SyncRoot)
                {
                    return new Hashtable(retrieveCounts);
                }
            }
        }

        /// <summary>
        /// Creates a new <see cref="ActivityStatistics"/> instance with no statistics recorded yet.
        /// </summary>
        public ActivityStatistics()
        {
            // Do nothing (all instance variables are set above.)
        }

        /// <summary>
        /// Increments the count for the <c>request.Identifier</c> entry of <paramref name="table"/>.
        /// </summary>
        /// <param name="table">The table of statistics to update.</param>
        /// <param name="request">The request to record.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        private static void Increment(IDictionary table, Request request)
        {
            // Check the arguments.
            Debug.Assert(table != null, "A dictionary must be specified.");
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            lock (table.SyncRoot)
            {
                if (table.Contains(request.Identifier))
                {
                    // Increment the count for this ID.
                    table[request.Identifier] = 1 + (int) table[request.Identifier];
                }
                else
                {
                    // Add a new count if this is the first time this ID has been seen.
                    table[request.Identifier] = (int) 1;
                }
            }
        }

        /// <summary>
        /// Gets the count for the <paramref name="identifier"/> entry of <paramref name="table"/>.
        /// </summary>
        /// <param name="table">The table of statistics to lookup <paramref name="identifier"/> in.</param>
        /// <param name="identifier">The identifier to lookup.</param>
        /// <returns>The value of <c>table[identifier]</c> 
        /// or 0 if <paramref name="identifer"/> is not in the table.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        private static int GetCount(IDictionary table, string identifier)
        {
            // Check the arguments.
            Debug.Assert(table != null, "A dictionary must be specified.");
            if (identifier == null)
            {
                throw new ArgumentNullException("identifier");
            }
            if (identifier == string.Empty)
            {
                throw new ArgumentException("The identifier cannot be the empty string.", "identifier");
            }

            // Get the count for this identifier.
            lock (table.SyncRoot)
            {
                // If the identifier exsists in the table, return its count; otherwise, return zero.
                return table.Contains(identifier) ? (int) table[identifier] : 0;
            }
        }

        /// <summary>
        /// Increases the receive count for the request's identifier.
        /// </summary>
        /// <param name="request">The incoming request.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        public void RequestReceived(Request request)
        {
            ActivityStatistics.Increment(receiveCounts, request);
        }

        /// <summary>
        /// Increases the store count for the request's identifier.
        /// </summary>
        /// <param name="request">The request that <paramref name="response"/> was bound to.</param>
        /// <param name="response">Ignored.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        public void ResponseStored(Request request, Response response)
        {
            // Check the response argument (even though its not used.)
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            ActivityStatistics.Increment(storeCounts, request);
        }

        /// <summary>
        /// Increases the retrieve count for the request's identifier.
        /// </summary>
        /// <param name="request">The request that is being retrieved.</param>
        /// <param name="response">Ignored.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        public void ResponseRetrieved(Request request, Response response)
        {
            // Check the response argument (even though its not used.)
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            ActivityStatistics.Increment(retrieveCounts, request);
        }

        /// <summary>
        /// Does nothing in this implementation (doesn't even check its argument for <B>null</B>.)
        /// </summary>
        /// <param name="request">Ignored.</param>
        /// <param name="exception">Ignored.</param>
        public void ExceptionOccurred(Request request, Exception exception)
        {
            // Do nothing.
        }

        /// <summary>
        /// Gets the receive count for the <paramref name="identifier"/>.
        /// </summary>
        /// <param name="identifier">The identifier to lookup.</param>
        /// <returns>The value of <c>ReceiveCounts[identifier]</c> 
        /// or 0 if <paramref name="identifer"/> is not in the table.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        public int GetReceiveCount(string identifier)
        {
            return GetCount(receiveCounts, identifier);
        }

        /// <summary>
        /// Gets the store count for the <paramref name="identifier"/>.
        /// </summary>
        /// <param name="identifier">The identifier to lookup.</param>
        /// <returns>The value of <c>StoreCounts[identifier]</c> 
        /// or 0 if <paramref name="identifer"/> is not in the table.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        public int GetStoreCount(string identifier)
        {
            return GetCount(storeCounts, identifier);
        }

        /// <summary>
        /// Gets the retrieve count for the <paramref name="identifier"/>.
        /// </summary>
        /// <param name="identifier">The identifier to lookup.</param>
        /// <returns>The value of <c>RetrieveCounts[identifier]</c> 
        /// or 0 if <paramref name="identifer"/> is not in the table.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        public int GetRetrieveCount(string identifier)
        {
            return GetCount(retrieveCounts, identifier);
        }

        /// <summary>
        /// Resets all counts for <paramref name="identifier"/> to 0.
        /// </summary>
        /// <param name="identifier">The identifier to reset.</param>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        public void ResetCount(string identifier)
        {
            // Check the argument.
            if (identifier == null)
            {
                throw new ArgumentNullException("identifier");
            }
            if (identifier == string.Empty)
            {
                throw new ArgumentException("The identifier cannot be the empty string.", "identifier");
            }

            // Remove this identifier from all tables (effectively resets them to 0.)
            lock (receiveCounts.SyncRoot)
            {
                receiveCounts.Remove(identifier);
            }
            lock (storeCounts.SyncRoot)
            {
                storeCounts.Remove(identifier);
            }
            lock (retrieveCounts.SyncRoot)
            {
                retrieveCounts.Remove(identifier);
            }
        }

        /// <summary>
        /// Resets all counts for all identifiers to 0.
        /// </summary>
        public void ResetAll()
        {
            // Clear all the tables (effectively resets all counts to 0.)
            lock (receiveCounts.SyncRoot)
            {
                receiveCounts.Clear();
            }
            lock (storeCounts.SyncRoot)
            {
                storeCounts.Clear();
            }
            lock (retrieveCounts.SyncRoot)
            {
                retrieveCounts.Clear();
            }
        }

        /// <summary>
        /// Does nothing (no special cleanup is required for this implementation.)
        /// </summary>
        public void Dispose()
        {
            // Do nothing.
        }
    }
}
