/*
 * IActivityLogger.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// Defines a notification interface for <see cref="IdempotencyManager"/> activities.
    /// </summary>
    /// <remarks>
    /// <para>
    /// There are 4 types of notifications:
    /// <list type="number">
    ///     <item><see cref="RequestReceived"/> - a request has been received by the managed.</item>
    ///     <item><see cref="ResponseStored"/> - a response object has been stored for the request.</item>
    ///     <item><see cref="ResponseRetrieved"/> - a previously stored response was retrieved.</item>
    ///     <item><see cref="ExceptionOccurred"/> - and exception occured in the manager.</item>
    /// </list>
    /// </para><para>
    /// This component currently provides two main implementation of this interface, the
    /// <see cref="ActivityLogging"/> class which records event notifications using the <B>Logging Wrapper</B>,
    /// and the <see cref="ActivityStatistics"/> class which keeps track of statistics in-memory for each
    /// request type. The <see cref="ActivityBroadcaster"/> is also provided and functions as a composite
    /// by forwarding notifications to any number of other <B>IActivityLogger</B>s.
    /// </para><para>
    /// <B>NOTE:</B> Implementations of this interface should be thread-safe.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public interface IActivityLogger : IDisposable 
    {
        /// <summary>
        /// Callback that is invoked when a request is received by the <see cref="IdempotencyManager"/>.
        /// </summary>
        /// <remarks>
        /// This method is invoked by the <see cref="IdempotencyManager.RequestArrival"/> before custom filtering.
        /// </remarks>
        /// <param name="request">The incoming request.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        void RequestReceived(Request request);

        /// <summary>
        /// Callback that is invoked when a response is stored by the <see cref="IdempotencyManager"/>.
        /// </summary>
        /// <remarks>
        /// This method is invoked by the <see cref="IdempotencyManager.StoreResponse"/> 
        /// if the response is sucessfully stored (no persistence errors.)
        /// </remarks>
        /// <param name="request">The request that <paramref name="response"/> was bound to.</param>
        /// <param name="response">The respone which was stored.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        void ResponseStored(Request request, Response response);

        /// <summary>
        /// Callback that is invoked when a response is retrieved from storage by the <see cref="IdempotencyManager"/>.
        /// </summary>
        /// <remarks>
        /// This method is invoked by the <see cref="IdempotencyManager.RetrieveResponse"/> methods either
        /// synchronously before the response is returned, or asynchronously before the response callback
        /// is invoked.
        /// </remarks>
        /// <param name="request">The request that is being retrieved.</param>
        /// <param name="response">The response which was retrieved.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        void ResponseRetrieved(Request request, Response response);

        /// <summary>
        /// Callback that is invoked when the <see cref="IdempotencyManager.ResponsePersistence"/>
        /// throws an exception during one of its operations.
        /// </summary>
        /// <remarks>
        /// The current implementation of <see cref="IdempotencyManager"/> does not catch any exceptions
        /// from the persistence layer. It simply allows such exceptions to propogate to the caller, so
        /// this method is never called by the current version.
        /// </remarks>
        /// <param name="request">The request which whose processing lead to this exception.</param>
        /// <param name="exception">The exception received by the manager.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="exception"/> is <B>null</B>.</exception>
        void ExceptionOccurred(Request request, Exception exception);
    }
}
