/*
 * CachedResponsePersistence.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using System.Diagnostics;
using TopCoder.Cache.Distributed;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// <see cref="IResponsePersistence"/> implementation which uses an in-memory <B>Distributed Simple Cache</B>
    /// to store responses and may optionally decorate another <B>IResponsePersistence</B> in order to provide
    /// caching for that implementation.
    /// </summary>
    /// <remarks>
    /// <para>
    /// Since this implementation stores responses in-memory, it is not technically persistent unless it
    /// is being used as a decorator on top of some other persistent storage (such as a database.)
    /// </para><para>
    /// <B>NOTE:</B> Thread-safety is achieved by locking the instance for all operations.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class CachedResponsePersistence : IResponsePersistence 
    {
        /// <summary>
        /// The distributed cache where responses are stored. It maps <see cref="Request.Identifier"/> strings
        /// to <see cref="IDictionary"/>s which, themselves, contain a mapping from <see cref="Request.SecurityKey"/>s
        /// to the actual <see cref="Request"/> objects.
        /// </summary>
        /// <remarks>
        /// <para>
        /// This value is set in the constructor and is immutable thereafter (although the object's internal
        /// state is still mutable.)
        /// </para><para>
        /// <B>FIXME:</B> This field should really be a <see cref="IDistributedSimpleCache"/> instead of the
        /// concrete <see cref="DistributedSimpleCache"/>, but the interface is not <see cref="IDisposable"/>
        /// in the current version, so we must use the concrete class for now.
        /// </para>
        /// </remarks>
        private readonly DistributedSimpleCache cache;

        /// <summary>
        /// An optional persistence that this class decorates. If not-null, this instance acts as a caching decorator.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly IResponsePersistence persistence;

        /// <summary>
        /// Creates a new <see cref="CachedResponsePersistence"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// Configuration values should appear in the 
        /// <c>TopCoder.Util.Idempotency.Persistence.CachedResponsePersistence</c> namespace. This namespace 
        /// must contain a <c>cache_url</c> property and may contain a <c>start_url</c> parameter which 
        /// are used to construct a <see cref="DistributedSimpleCache"/>. It may also contain a
        /// <c>response_persistence</c> property which should be the fully-qualified typename of an
        /// <see cref="IResponsePersistence"/> with a default constructor that will be instantiated and used
        /// as the inner persistence uppon which this cache is based. If no inner persistence is specified,
        /// this this class simply uses its cache as the "persistence" storage.
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        public CachedResponsePersistence()
        {
            // Ge the namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Read the cache_url property.
            string cacheUrl = ConfigHelper.GetValue(ns, "cache_url", true);

            // Read the start_url property (null wil be returned if the property does not exsist.)
            string startUrl = ConfigHelper.GetValue(ns, "start_url", false);

            // Attempt to create a new distributed cache.
            try
            {
                cache = new DistributedSimpleCache(cacheUrl, startUrl);
            }
            catch (Exception e)
            {
                // If there was a problem, wrap the exception and rethrow.
                throw new InvalidConfigFileException("Unable to create a DistributedSimpleCache with " +
                    "the cache_url and start_url configuration values in " + ns.Name, e);
            }

            // Load the optional response_persistence object.
            persistence = (IResponsePersistence) ConfigHelper.CreateTypesafeInstance(
                ns, "response_persistence", typeof(IResponsePersistence), false);

        }

        /// <summary>
        /// Creates a new <see cref="CachedResponsePersistence"/> instance based on the given parameters.
        /// </summary>
        /// <param name="cache">The pre-constructed cache to use for "persistence" storage.</param>
        /// <param name="persistence">An optional backing store which this instance will function as a
        /// cache for. <B>null</B> if the cache should be used as the only storage.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        public CachedResponsePersistence(IDistributedSimpleCache cache, IResponsePersistence persistence)
        {
            // Check the arguments.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }

            // Set the instance fields.
            this.cache = (DistributedSimpleCache) cache; // FIXME: blind cast for now, but should change with new DSC
            this.persistence = persistence;
        }

        /// <summary>
        /// Stores a response to the given request in the cache and, optionally, in the wrapped persistence.
        /// </summary>
        /// <param name="request">The request this response is stored for.</param>
        /// <param name="response">The response to store.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The response could not be stored.</exception>
        public void Store(Request request, Response response)
        {
            // Check the arguments.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            lock (this)
            {
                // If there is an inner persistence, store the response there first.
                if (persistence != null)
                {
                    persistence.Store(request, response);
                }

                // Get all responses for this ID.
                IDictionary dict = (IDictionary) cache.Get(request.Identifier);

                // If there are no responses for this ID yet, create a new table for them and add it to the cache.
                if (dict == null)
                {
                    dict = new Hashtable();
                    cache.Put(request.Identifier, dict);
                }

                // Add the response to this ID's table.
                dict[request.SecurityKey] = response;
            }
        }

        /// <summary>
        /// Retrieves a response to the given request from the cache. If it it not found in the cache, but there
        /// is a wrapped persistence, an attempt will be made to return it from the inner persistence.
        /// </summary>
        /// <param name="request">The request to retrieve a response for.</param>
        /// <returns>The response that was most recently stored for this request.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        /// <exception cref="ResponseRetrievalException">No response was found for the given request.</exception>
        /// <exception cref="PersistenceException">The response could not be retrieved 
        /// for some other reason.</exception>
        public Response Retrieve(Request request)
        {
            // Check the argument.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            lock (this)
            {
                // Get the list of all responses for this ID.
                IDictionary dict = (IDictionary) cache.Get(request.Identifier);
                Response response;

                // If there are responses for this ID, continue...
                if (dict != null)
                {
                    // Get the response specific to this security key.
                    response = (Response) dict[request.SecurityKey];

                    // If there is a response, return it now.
                    if (response != null)
                    {
                        return response;
                    }
                }

                // If there is no cached value and no inner persistence, give up and throw an exception.
                if (persistence == null)
                {
                    throw new ResponseRetrievalException(request);
                }
                
                // Get the response from the inner persistence.
                response = persistence.Retrieve(request);

                // The inner persistence should never return null (but it may throw a ResponseRetrivalException.)
                Debug.Assert(response != null, "A persistence should never return a null response.");

                // If there is no cache entry for this ID, add one.
                if (dict == null)
                {
                    dict = new Hashtable();
                    cache.Put(request.Identifier, dict);
                }

                // Add the response to the cache's dictionary for this ID.
                dict[request.SecurityKey] = response;

                // Return the response.
                return response;
            }
        }

        /// <summary>
        /// Deletes all stored responses to requests with the given identifer from the cache and from
        /// the inner persistence, if defined.
        /// </summary>
        /// <remarks>
        /// There may be several responses with the same <paramref name="identifer"/>, but different
        /// security keys. In this case, all of these responses are be removed from both places.
        /// </remarks>
        /// <param name="identifier">The identifier of responses to delete.</param>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        /// <exception cref="PersistenceException">The responses could not be deleted.</exception>
        public void Delete(string identifier)
        {
            // Check the argument.
            if (identifier == null)
            {
                throw new ArgumentNullException("identifier");
            }
            if (identifier == string.Empty)
            {
                throw new ArgumentException("The identifier cannot be empty.", "identifier");
            }

            lock (this)
            {
                // If there is an inner persistence, delete the responses in there first.
                if (persistence != null)
                {
                    persistence.Delete(identifier);
                }

                // Delete the cached responses.
                cache.Remove(identifier);
            }
        }

        /// <summary>
        /// Gets an enumeration of all the identifiers for responses currently stored in the inner persistence,
        /// if any, or the cache if there is no inner persistence.
        /// </summary>
        /// <remarks>
        /// No identifer will appear more than once in this enumeration, even if there are multiple responses
        /// (with different security keys) stored for that identifier.
        /// </remarks>
        /// <returns>An enumeration of the set of identifiers for which responses are currently stored.</returns>
        /// <exception cref="PersistenceException">The enumeration could not be returned.</exception>
        public IEnumerator GetIdentifiers()
        {
            lock (this)
            {
                // If there is an inner persistence, return its identifiers.
                if (persistence != null)
                {
                    return persistence.GetIdentifiers();
                }

                // Otherwise, return the IDs of the cached responses.
                return new ArrayList(cache.Keys).GetEnumerator();
            }
        }

        /// <summary>
        /// Clears the inner persistentce (if any) as well as the cache.
        /// </summary>
        /// <exception cref="PersistenceException">The entries could not be cleared.</exception>
        public void Clear()
        {
            lock (this)
            {
                // If there is an inner persistence, clear it first.
                if (persistence != null)
                {
                    persistence.Clear();
                }

                // Clear the cache.
                cache.Clear();
            }
        }

        /// <summary>
        /// If there is an inner persistence, it is disposed and then the cache is disposed.
        /// </summary>
        public void Dispose()
        {
            lock (this)
            {
                // If there is an inner persisitence, dispose of it first.
                if (persistence != null)
                {
                    persistence.Dispose();
                }

                // Dispose of the cache.
                cache.Dispose();
            }
        }
    }
}
