/*
 * DatabasePersistence.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.IO;
using System.Data;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using TopCoder.Cache.Distributed;
using TopCoder.Util.ConfigurationManager;
using TopCoder.Data.ConnectionFactory;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// <see cref="ICachePersistence"/> and <see cref="IResponsePersistence"/> implementation which uses a 
    /// database to store the request cache and responses.
    /// </summary>
    /// <remarks>
    /// <para>
    /// If this class is to work properly in a distributed environment, all instances shoud connect to the
    /// same database.
    /// </para><para>
    /// <B>NOTE:</B> Transactions are used to execute operations which require more than a single SQL statement
    /// so that the database remains consistent in a multi-threaded and distributed environment. In addition,
    /// the connection object is locked to ensure that only one thread uses it at a time. (Future version of
    /// the component may wish to try improving performance by creating multiple connections to the database
    /// to allow concurrent reads in a multi-threaded environment.)
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class DatabasePersistence : IResponsePersistence, ICachePersistence
    {
        /// <summary>
        /// A connection to the database, used by all of the methods.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter. Since the underlying ADO.NET
        /// libraries provide thread pooling, there is no reason to leave the connection open between uses.
        /// Each method opens the connection when it starts, and then closes it when it is done. This releases
        /// the connection back to the pool so that other parts of the system may reuse that same connection
        /// for the database if needed.
        /// </remarks>
        private readonly IDbConnection connection; 

        /// <summary>
        /// Creates a new <see cref="DatabasePersistence"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// Configuration values should appear in the 
        /// <c>TopCoder.Util.Idempotency.Persistence.DatabasePersistence</c> namespace. This namespace 
        /// must contain a <c>predefined_connection</c> parameter which is used in the <B>Connection Manager</B>
        /// component to create a new database connection.
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        public DatabasePersistence()
        {
            // Get the proper namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Read the predefined connection name.
            string connectionName = ConfigHelper.GetValue(ns, "predefined_connection", true);

            // Attempt to create a connection object using the Connection Manager.
            try
            {
                connection = ConnectionManager.Instance.CreatePredefinedDbConnection(connectionName);
            }
            catch (Exception e)
            {
                // If there was a problem, report it by wrapping the exception and rethrowing.
                throw new InvalidConfigFileException("Unable to create the predefined connection.", e);
            }
        }

        /// <summary>
        /// Creates a new <see cref="DatabasePersistence"/> instance based on the given parameters.
        /// </summary>
        /// <remarks>
        /// It is encouraged that the connection not be used externally after it is passed to this
        /// constructor. The connection should also be in a <see cref="ConnectionState.Closed"/> state since
        /// the it will be opened every time a method is invoked. If, however, the connection must continue
        /// to be used, locking the connection object will synchronize external use with the internal usage.
        /// If this is done, the connection must always be closed again before the lock is released to
        /// prevent an exception from being thrown when the next method of this class is invoked.
        /// </remarks>
        /// <param name="connection">The pre-constructed connection to the database.</param>
        /// <exception cref="ArgumentNullException"><paramref name="connection"/> is <B>null</B>.</exception>
        public DatabasePersistence(IDbConnection connection)
        {
            // Check the argument.
            if (connection == null)
            {
                throw new ArgumentNullException("connection");
            }

            // Set the instance field.
            this.connection = connection;
        }

        /// <summary>
        /// Creates a new <see cref="IDbDataParameter"/> for the given <see cref="IDbCommand"/>,
        /// fills it with the given parameters, and adds it to the command's parameter list.
        /// </summary>
        /// <param name="cmd">The command used to create and add the parameter.</param>
        /// <param name="name">The name of the parameter (sets <see cref="IDataParameter.ParameterName"/>.)</param>
        /// <param name="type">The datatype of the parameter (sets <see cref="IDataParameter.DbType"/>.)</param>
        /// <returns>The newly created parameter object (for furthur modification, if needed.)</returns>
        private IDbDataParameter NewParam(IDbCommand cmd, string name, DbType type)
        {
            // Create the new parameter and add it to the parameter list.
            IDbDataParameter param = cmd.CreateParameter();
            cmd.Parameters.Add(param);

            // Set the parameters values.
            param.ParameterName = name;
            param.DbType = type;

            // Return the parameter.
            return param;
        }

        /// <summary>
        /// Creates a new <see cref="IDbDataParameter"/> for the given <see cref="IDbCommand"/>,
        /// fills it with the given parameters, and adds it to the command's parameter list.
        /// </summary>
        /// <param name="cmd">The command used to create and add the parameter.</param>
        /// <param name="name">The name of the parameter (sets <see cref="IDataParameter.ParameterName"/>.)</param>
        /// <param name="type">The datatype of the parameter (sets <see cref="IDataParameter.DbType"/>.)</param>
        /// <param name="size">The size of the parameter (sets <see cref="IDbDataParameter.Size"/>.)</param>
        /// <returns>The newly created parameter object (for furthur modification, if needed.)</returns>
        private IDbDataParameter NewParam(IDbCommand cmd, string name, DbType type, int size)
        {
            // Create the new parameter with the given setting.
            IDbDataParameter param = NewParam(cmd, name, type);

            // In addition, set the size before returning.
            param.Size = size;
            return param;
        }

        /// <summary>
        /// Stores a response to the given request in the database.
        /// </summary>
        /// <remarks>
        /// This operation is transactioned since any old entries for this request must be removed first.
        /// </remarks>
        /// <param name="request">The request this response is stored for.</param>
        /// <param name="response">The response to store.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The response could not be stored.</exception>
        public void Store(Request request, Response response)
        {
            // Check the arguments.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }
            if (response == null)
            {
                throw new ArgumentNullException("response");
            }

            lock (connection)
            {
                IDbTransaction transaction = null;
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Create a transaction with the highest isolation level.
                    transaction = connection.BeginTransaction(IsolationLevel.Serializable);

                    // Clear the cache_image table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.Transaction = transaction;
                    cmd.CommandText = "DELETE FROM response WHERE identifier = @id AND security_key = @key";

                    // Setup and add the parameters.
                    NewParam(cmd, "@id", DbType.AnsiString, 255).Value = request.Identifier;
                    NewParam(cmd, "@key", DbType.AnsiString, 255).Value = request.SecurityKey;

                    // Delete any exsisting rows.
                    cmd.ExecuteNonQuery();

                    // Setup the command for adding cache entries.
                    cmd = connection.CreateCommand();
                    cmd.Transaction = transaction;
                    cmd.CommandText = 
                        "INSERT INTO response (identifier, security_key, response_object, time_stamp, life_time) " +
                        "VALUES (@id, @key, @response, @time_stamp, @life_time)";

                    // Serialize the response.
                    MemoryStream mem = new MemoryStream();
                    IFormatter fmt = new BinaryFormatter();
                    fmt.Serialize(mem, response);

                    // Setup and add the parameters.
                    NewParam(cmd, "@id", DbType.AnsiString, 255).Value = request.Identifier;
                    NewParam(cmd, "@key", DbType.AnsiString, 255).Value = request.SecurityKey;
                    NewParam(cmd, "@response", DbType.Binary, (int) mem.Length).Value = mem.ToArray();
                    NewParam(cmd, "@time_stamp", DbType.DateTime).Value = response.Timestamp;
                    NewParam(cmd, "@life_time", DbType.Int64).Value = response.Life.Ticks;

                    // Perform the insertion and commit the changes.
                    cmd.ExecuteNonQuery();
                    transaction.Commit();
                }
                catch (Exception e)
                {
                    // Rollback the transaction.
                    if (transaction != null)
                    {
                        transaction.Rollback();
                    }

                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to save the cache to the database.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Retrieves a response to the given request from the database.
        /// </summary>
        /// <param name="request">The request to retrieve a response for.</param>
        /// <returns>The response that was most recently stored for this request.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        /// <exception cref="ResponseRetrievalException">No response was found for the given request.</exception>
        /// <exception cref="PersistenceException">The response could not be retrieved 
        /// for some other reason.</exception>
        public Response Retrieve(Request request)
        {
            // Check the argument.
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            lock (connection)
            {
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Clear the cache_image table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.CommandText = "SELECT response_object FROM response WHERE identifier = @id AND security_key = @key";

                    // Setup and add the parameters.
                    NewParam(cmd, "@id", DbType.AnsiString, 255).Value = request.Identifier;
                    NewParam(cmd, "@key", DbType.AnsiString, 255).Value = request.SecurityKey;

                    using (IDataReader reader = cmd.ExecuteReader())
                    {
                        // Make sure there is a response available.
                        if (!reader.Read())
                        {
                            throw new ResponseRetrievalException(request);
                        }

                        // Extract the object data.
                        byte[] bytes = (byte[]) reader.GetValue(0);

                        // Deserialize the response and return it.
                        MemoryStream mem = new MemoryStream(bytes);
                        IFormatter fmt = new BinaryFormatter();
                        return (Response) fmt.Deserialize(mem);
                    }
                }
                catch (ResponseRetrievalException)
                {
                    // Just let these exceptions propogate without wrapping.
                    throw;
                }
                catch (Exception e)
                {
                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to retrive a response from the database.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Deletes all stored responses to requests with the given identifer from the database.
        /// </summary>
        /// <remarks>
        /// There may be several responses with the same <paramref name="identifer"/>, but different
        /// security keys. In this case, all of these responses are be removed from the database.
        /// </remarks>
        /// <param name="identifier">The identifier of responses to delete.</param>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        /// <exception cref="PersistenceException">The responses could not be deleted.</exception>
        public void Delete(string identifier)
        {
            // Check the argument.
            if (identifier == null)
            {
                throw new ArgumentNullException("identifier");
            }
            if (identifier == string.Empty)
            {
                throw new ArgumentException("The identifier cannot be empty.", "identifier");
            }

            lock (connection)
            {
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Remove all rows with the given ID from the response table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.CommandText = "DELETE FROM response WHERE identifier = @id";

                    // Setup and add the ID parameter and execute.
                    NewParam(cmd, "@id", DbType.AnsiString, 255).Value = identifier;
                    cmd.ExecuteNonQuery();
                }
                catch (Exception e)
                {
                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to delete rows from the response table.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Gets an enumeration of all the identifiers for responses currently stored in the database.
        /// </summary>
        /// <remarks>
        /// No identifer will appear more than once in this enumeration, even if there are multiple responses
        /// (with different security keys) stored for that identifier.
        /// </remarks>
        /// <returns>An enumeration of the set of identifiers for which responses are currently stored.</returns>
        /// <exception cref="PersistenceException">The enumeration could not be returned.</exception>
        public IEnumerator GetIdentifiers()
        {
            lock (connection)
            {
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Get all of the unique identifiers in the response table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.CommandText = "SELECT DISTINCT identifier FROM response";
                    using (IDataReader reader = cmd.ExecuteReader())
                    {
                        // Extract all the results and put them in a new list.
                        IList keys = new ArrayList();
                        while (reader.Read())
                        {
                            keys.Add(reader.GetString(0));
                        }

                        // Return an enumerator for that list.
                        return keys.GetEnumerator();
                    }
                }
                catch (Exception e)
                {
                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to clear the response table.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Deletes all responses from the database.
        /// </summary>
        /// <exception cref="PersistenceException">The entries could not be cleared.</exception>
        public void Clear()
        {
            lock (connection)
            {
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Remove all rows from the response table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.CommandText = "TRUNCATE TABLE response";
                    cmd.ExecuteNonQuery();
                }
                catch (Exception e)
                {
                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to clear the response table.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Loads the stored cache entries from the database into the given cache.
        /// </summary>
        /// <remarks>
        /// If an entry already exsists with the same identifier (cache key) as an entry being loaded from
        /// the database, the exsisting entry is not overwritten (the stored entry is ignored.)
        /// </remarks>
        /// <param name="cache">The cache to load stored entries into.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The entries could not be loaded.</exception>
        public void LoadCache(IDistributedSimpleCache cache)
        {
            // Check the argument.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }

            lock (connection)
            {
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Get all the rows in the cache_image table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.CommandText = "SELECT * FROM cache_image";
                    using (IDataReader reader = cmd.ExecuteReader())
                    {
                        // Iterate through all rows in the result set.
                        while (reader.Read())
                        {
                            // Extract the data from the current record.
                            string key = reader.GetString(0);
                            CacheEntry entry = new CacheEntry(reader.GetDateTime(1), reader.GetBoolean(2));

                            // If there is not already an entry for the key, add this entry.
                            if (cache.Get(key) == null)
                            {
                                cache.Put(key, entry);
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to load the cache from the database.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Saves all cache entries to the database.
        /// </summary>
        /// <remarks>
        /// This operation is transactioned since it must first delete any previously stored items before it
        /// adds each of the cache entries. Because of this, the operation is potentially very slow so it
        /// should only be used at system shutdown.
        /// </remarks>
        /// <param name="cache">The cache to save entries from.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The entries could not be saved.</exception>
        public void SaveCache(IDistributedSimpleCache cache)
        {
            // Check the argument.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }

            lock (connection)
            {
                IDbTransaction transaction = null;
                try
                {
                    // Open a connection (or get one of the pooled connections.)
                    connection.Open();

                    // Create a transaction with the highest isolation level.
                    transaction = connection.BeginTransaction(IsolationLevel.Serializable);

                    // Clear the cache_image table.
                    IDbCommand cmd = connection.CreateCommand();
                    cmd.Transaction = transaction;
                    cmd.CommandText = "TRUNCATE TABLE cache_image";
                    cmd.ExecuteNonQuery();

                    // Setup the command for adding cache entries.
                    cmd.CommandText = 
                        "INSERT INTO cache_image (identifier, last_activity, in_storage) VALUES (@id, @last, @in)";

                    // Setup and add the parameters.
                    IDbDataParameter idParam = NewParam(cmd, "@id", DbType.AnsiString, 255);
                    IDbDataParameter lastParam = NewParam(cmd, "@last", DbType.DateTime);
                    IDbDataParameter inParam = NewParam(cmd, "@in", DbType.Boolean);

                    // Loop through all the cached items, adding them to the database.
                    IList keys = new ArrayList(cache.Keys);
                    foreach (string key in keys)
                    {
                        // Get this entry.
                        CacheEntry entry = (CacheEntry) cache.Get(key);

                        // If the entry is still in the cache, add it to the database.
                        if (entry != null)
                        {
                            idParam.Value = key;
                            lastParam.Value = entry.LastActivity;
                            inParam.Value = entry.IsInStorage;
                            cmd.ExecuteNonQuery();
                        }
                    }

                    // Commit the changes.
                    transaction.Commit();
                }
                catch (Exception e)
                {
                    // Rollback the transaction.
                    if (transaction != null)
                    {
                        transaction.Rollback();
                    }

                    // Wrap the error and rethrow.
                    throw new PersistenceException("Unable to save the cache to the database.", e);
                }
                finally
                {
                    // Make sure the connection gets closed.
                    connection.Close();
                }
            }
        }

        /// <summary>
        /// Does nothing (there are no special resources to dispose of in this implementation.)
        /// </summary>
        /// <remarks>
        /// Since the underlying ADO.NET implementation provides connection pooling, there is no reason
        /// to leave connection objects open between operations so there is no cleanup that must
        /// be done in this method.
        /// </remarks>
        public void Dispose()
        {
            // Do nothing (the connection is closed after every operation.)
        }
    }
}
