/*
 * FileSystemCachePersistence.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using TopCoder.Cache.Distributed;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// <see cref="ICachePersistence"/> implementation which uses a file to load and save the cache.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This implementation is not distributed (unless the underlying filesystem is distributed.)
    /// </para><para>
    /// <B>NOTE:</B> File locking is used to ensure that this class is thread-safe (and to protect it
    /// against other attempts at concurrent modifications to the save files used.)
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class FileSystemCachePersistence : ICachePersistence
    {
        /// <summary>
        /// The location of the file used to store cache entries (if the file does not exsist, it will be created.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly string filename;

        /// <summary>
        /// Creates a new <see cref="FileSystemCachePersistence"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// Configuration values should appear in the 
        /// <c>TopCoder.Util.Idempotency.Persistence.FileSystemCachePersistence</c> namespace. This namespace 
        /// must contain a <c>filename</c> parameter which should contain the path where the cache should
        /// be saved and/or loaded from.
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        public FileSystemCachePersistence()
        {
            // Get the configuration namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Get the filename parameter.
            filename = ConfigHelper.GetValue(ns, "filename", true);
        }

        /// <summary>
        /// Creates a new <see cref="FileSystemCachePersistence"/> instance based on the given parameter.
        /// </summary>
        /// <param name="filename">The path where the cache should be saved and/or loaded from.</param>
        /// <exception cref="ArgumentNullException"><paramref name="filename"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="filename"/> is an empty string.</exception>
        public FileSystemCachePersistence(string filename)
        {
            // Check the argument.
            if (filename == null)
            {
                throw new ArgumentNullException("filename");
            }
            if (filename == string.Empty)
            {
                throw new ArgumentException("The filename cannot be an empty string.", "filename");
            }

            // Set the instance field.
            this.filename = filename;
        }

        /// <summary>
        /// Loads the stored cache entries from a file into the given cache.
        /// </summary>
        /// <remarks>
        /// If an entry already exsists with the same identifier (cache key) as an entry being loaded from
        /// the file, the exsisting entry is not overwritten (the stored entry is ignored.)
        /// </remarks>
        /// <param name="cache">The cache to load stored entries into.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The entries could not be loaded.</exception>
        public void LoadCache(IDistributedSimpleCache cache)
        {
            // Check the argument.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }

            FileInfo file = new FileInfo(filename);

            // If the file does not exsist, throw an exception.
            if (!file.Exists)
            {
                throw new PersistenceException("Unable to load the cache from '" + 
                    filename + "' (the file does not exsist.)");
            }

            try
            {
                // Get a read-only, locked (only other readers allowed) FileStream and deserialize its contents.
                using (Stream stream = file.Open(FileMode.Open, FileAccess.Read, FileShare.Read))
                {
                    // The data is stored in binary format.
                    IFormatter formatter = new BinaryFormatter();

                    // While there is still data in the stream, read another cache entry.
                    while (stream.Position < stream.Length)
                    {
                        // Deserialize the cache key (identifier) and the cache value (CacheEntry.)
                        string key = (string) formatter.Deserialize(stream);
                        CacheEntry value = (CacheEntry) formatter.Deserialize(stream);

                        // Add the value to the cache if there is not already an entry with that ID.
                        if (cache.Get(key) == null)
                        {
                            cache.Put(key, value);
                        }
                    }
                }
            }
            catch (FileNotFoundException)
            {
                // The file must have been deleted between our check and the creation of the FileStream.
                // Unfortuately, there's no way to prevent this possibility so we have to duplicate the
                // exception generation here as well.
                throw new PersistenceException("Unable to load the cache from '" + 
                    filename + "' (the file does not exsist.)");
            }
            catch (Exception e)
            {
                // If anything goes wrong, wrap the exception in a PersistenceException and throw that.
                throw new PersistenceException("Unable to Load the cache.", e);
            }
        }

        /// <summary>
        /// Saves all cache entries to a file (serializes them.)
        /// </summary>
        /// <param name="cache">The cache to save entries from.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The entries could not be saved.</exception>
        public void SaveCache(IDistributedSimpleCache cache)
        {
            // Check the argument.
            if (cache == null)
            {
                throw new ArgumentNullException("cache");
            }

            try
            {
                FileInfo file = new FileInfo(filename);

                // Get a write-only, locked FileStream and serialize the cache to it.
                using (Stream stream = file.Open(FileMode.OpenOrCreate, FileAccess.Write, FileShare.None))
                {
                    // Store the data in binay format.
                    IFormatter formatter = new BinaryFormatter();

                    // Get all the keys in the cache right now.
                    IList keys = new ArrayList(cache.Keys);

                    // Serialize each of the key/value pairs to the file.
                    foreach (string key in keys)
                    {
                        // Get the entry for the current key.
                        CacheEntry entry = (CacheEntry) cache.Get(key);

                        // If the entry hasn't been removed from the cache, add this key/value to the file.
                        if (entry != null)
                        {
                            formatter.Serialize(stream, key);
                            formatter.Serialize(stream, entry);
                        }
                    }
                }
            }
            catch (Exception e)
            {
                // If anything goes wrong, wrap the exception in a PersistenceException and throw that.
                throw new PersistenceException("Unable to Save the cache.", e);
            }
        }

        /// <summary>
        /// Does nothing (there are no special resources to dispose of in this implementation.)
        /// </summary>
        public void Dispose()
        {
            // Do nothing.
        }
    }
}
