/*
 * ICachePersistence.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using TopCoder.Cache.Distributed;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// Interface for any class wishing to provide <see cref="CacheEntry"/> storage (persistence) for the 
    /// <see cref="IdempotencyManager"/>.
    /// </summary>
    /// <remarks>
    /// <para>
    /// A component of this type is not required for the <see cref="IdempotencyManager"/> to run, but may
    /// be useful since it allows the request cache to be saved and loaded instead of rebuilt each time
    /// the component is used.
    /// </para><para>
    /// This component currently provides two main implementation of this interface, the
    /// <see cref="DatabasePersistence"/> class which saves the cache to an external database,
    /// and the <see cref="FileSystemCachePersistence"/> class which serialized the cache entries to a file.
    /// </para><para>
    /// <B>NOTE:</B> Implementations of this interface should be thread-safe, but need not be distributed
    /// (since the cache, itself, is distributed.)
    /// Also, though thread-safety is required, instances are unlikely to be used in a multi-thread
    /// environment since they should only be used during startup and shutdown.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public interface ICachePersistence : IDisposable 
    {
        /// <summary>
        /// Loads the stored cache entries into the given cache.
        /// </summary>
        /// <remarks>
        /// If an entry already exsists with the same identifier (cache key) as an entry being loaded from
        /// storage, the exsisting entry should not be overwritten (simply skip loading that stored entry.)
        /// </remarks>
        /// <param name="cache">The cache to load stored entries into.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The implementation was unable to load the entries.</exception>
        void LoadCache(IDistributedSimpleCache cache);

        /// <summary>
        /// Saves all cache entries to persistent storage.
        /// </summary>
        /// <remarks>
        /// The persistent storage should be cleared of any old saved data, if needed, before saving the
        /// new cache entries.
        /// </remarks>
        /// <param name="cache">The cache to save entries from.</param>
        /// <exception cref="ArgumentNullException"><paramref name="cache"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The implementation was unable to save the entries.</exception>
        void SaveCache(IDistributedSimpleCache cache);
    }
}
