/*
 * IResponsePersistence.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// Interface for any class wishing to provide response storage (persistence) for the 
    /// <see cref="IdempotencyManager"/>.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This component currently provides two main implementation of this interface, the
    /// <see cref="DatabasePersistence"/> class which saves responses to an external database,
    /// and the <see cref="CachedResponsePersistence"/> class which uses a <B>Distributed Simple Cache</B>
    /// to save responses in memory.
    /// </para><para>
    /// <B>NOTE:</B> Implementations of this interface should be thread-safe and distributed (all instances
    /// of these classes within a single Idempotency group should return identical results from a query.)
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public interface IResponsePersistence : IDisposable 
    {
        /// <summary>
        /// Stores a response to the given request in persistent storage.
        /// </summary>
        /// <remarks>
        /// The persistence should use both the <see cref="Request.Identifier"/> and <see cref="Request.SecurityKey"/>,
        /// but not the <see cref="Request.Type"/> to differentiate the responses.
        /// </remarks>
        /// <param name="request">The request this response is stored for.</param>
        /// <param name="response">The response to store.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.<BR/>-or<BR/>
        /// <paramref name="response"/> is <B>null</B>.</exception>
        /// <exception cref="PersistenceException">The implementation was unable to store the response.</exception>
        void Store(Request request, Response response);

        /// <summary>
        /// Retrieves a response to the given request from persistent storage.
        /// </summary>
        /// <remarks>
        /// The persistence should use both the <see cref="Request.Identifier"/> and <see cref="Request.SecurityKey"/>,
        /// but not the <see cref="Request.Type"/> to differentiate the responses.
        /// </remarks>
        /// <param name="request">The request to retrieve a response for.</param>
        /// <returns>The response that was most recently stored for this request.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        /// <exception cref="ResponseRetrievalException">The implementation could not find a response
        /// for the given request.</exception>
        /// <exception cref="PersistenceException">The implementation was unable to retrieve the response
        /// for any other reason.</exception>
        Response Retrieve(Request request);

        /// <summary>
        /// Deletes all stored responses to requests with the given identifer from storage.
        /// </summary>
        /// <remarks>
        /// There may be several responses with the same <paramref name="identifer"/>, but different
        /// security keys. In this case, all of these responses should be removed from storage.
        /// </remarks>
        /// <param name="identifier">The identifier of responses to delete.</param>
        /// <exception cref="ArgumentNullException"><paramref name="identifier"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        /// <exception cref="PersistenceException">The implementation was unable to delete the responses.</exception>
        void Delete(string identifier);

        /// <summary>
        /// Gets an enumeration of all the identifiers for responses currently stored in this persistence.
        /// </summary>
        /// <remarks>
        /// No identifer should appear more than once in this enumeration, even if there are multiple responses
        /// (with different security keys) stored for an identifier.
        /// </remarks>
        /// <returns>An enumeration of the set of identifiers for which responses are currently stored.</returns>
        /// <exception cref="PersistenceException">The implementation was unable to create the enumeration.</exception>
        IEnumerator GetIdentifiers();

        /// <summary>
        /// Clears the persistent storage (removes all stored responses.)
        /// </summary>
        /// <exception cref="PersistenceException">The implementation was unable to clear the responses.</exception>
        void Clear();
    }
}
