/*
 * PersistenceException.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// This exception is used to wrap errors caused by persistent storage operations.
    /// </summary>
    /// <remarks>
    /// <B>NOTE:</B> This exception should be thrown for any persistence problems other than non-exsistent
    /// responses (use the <see cref="ResponseRetrievalException"/> in that case.)
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class PersistenceException : Exception
    {
        /// <summary>
        /// Creates a new <see cref="PersistenceException"/> with the given error <paramref name="message"/>.
        /// </summary>
        /// <param name="message">A description of the cause of this exception.</param>
        /// <exception cref="ArgumentNullException"><paramref name="message"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="message"/> is an empty string.</exception>
        public PersistenceException(string message)
            : base(message)
        {
            // Check the argument (after the fact.)
            if (message == null)
            {
                throw new ArgumentNullException("message");
            }
            if (message == string.Empty)
            {
                throw new ArgumentException("The message cannot be empty.", "message");
            }
        }

        /// <summary>
        /// Creates a new <see cref="PersistenceException"/> with the given error <paramref name="message"/>
        /// and <paramref name="cause"/> (<c>innerException</c>.)
        /// </summary>
        /// <param name="message">A description of the cause of this exception.</param>
        /// <param name="cause">The exception that caused this exception (<c>innerException</c>.)</param>
        /// <exception cref="ArgumentNullException"><paramref name="message"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="cause"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="message"/> is an empty string.</exception>
        public PersistenceException(string message, Exception cause)
            : base(message, cause)
        {
            // Check the arguments (after the fact.)
            if (message == null)
            {
                throw new ArgumentNullException("message");
            }
            if (message == string.Empty)
            {
                throw new ArgumentException("The message cannot be empty.", "message");
            }
            if (cause == null)
            {
                throw new ArgumentNullException("cause");
            }
        }
    }
}
