/*
 * ResponseRetrievalException.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// This exception is thrown when an attempt is made to retrieve a response from persistence that
    /// is not stored there.
    /// </summary>
    /// <remarks>
    /// <para>
    /// Since the <see cref="IdempotencyManager"/> always checks the persistence to ensure that it has
    /// at least one response for the request's identifier, this exception usually means that the
    /// security key did not match any stored responses (which could indicate an unauthorized access.)
    /// </para><para>
    /// <B>NOTE:</B> This exception is only thrown if the response does not exsist. Any other problem
    /// that occurs during a response retrieval from persistent storage should throw a 
    /// <see cref="PersistenceException"/>.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class ResponseRetrievalException : Exception
    {
        /// <summary>
        /// The request who's response could not be found.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly Request request; 

        /// <summary>
        /// Gets the request which caused this exception.
        /// </summary>
        /// <value>
        /// The request who's response could not be found.
        /// </value>
        public Request Request
        {
            get
            {
                return request;
            }
        }

        /// <summary>
        /// Creates a new <see cref="ResponseRetrievalException"/> for the given <paramref name="request"/>.
        /// </summary>
        /// <param name="request">The request which caused the exception.</param>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        public ResponseRetrievalException(Request request)
            : base(ResponseRetrievalException.ErrorMessage(request))
        {
            // Set the instance field.
            this.request = request;
        }

        /// <summary>
        /// Checks the argument and then uses it to build an error message.
        /// </summary>
        /// <remarks>
        /// This method is required so that the <paramref name="request"/> argument can be checked before its
        /// identifier is used to construct the message to pass to the base constructor in this class' constructor.
        /// </remarks>
        /// <param name="request">The request which caused the exception.</param>
        /// <returns>An exception message containing the request's 
        /// <see cref="Idempotency.Request.Identifier"/>.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        private static string ErrorMessage(Request request)
        {
            // Check the argument (since the constructor can't.)
            if (request == null)
            {
                throw new ArgumentNullException("request");
            }

            // Create a message containing the request's ID, but *NOT* the security key.
            return "Unable to retrieve a response for this request: " + request.Identifier;
        }
    }
}
