/*
 * Request.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using TopCoder.Util.Idempotency.Strategy;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// A lightweight proxy for a request who's response should be cached in a distributed fasion.
    /// </summary>
    /// <remarks>
    /// <param>
    /// A request is defined primarily by an identifier, which must not be <B>null</B> or empty. In addition,
    /// a security key may be used to provide different responses depending on some level of "security clearance."
    /// If such security measures are not needed, the security key may be left blank (but not <B>null</B>.)
    /// A type may also be specified, along with an arbitrary table of attributes, which may be used by a
    /// custom <see cref="ICacheStrategy"/> to determine which requests should be cached and which should not.
    /// Since the <B>Request</B> objects are transmitted over the network to keep the distributed caches up to
    /// date, they attribute keys and values must all be serializable. In addition, it is best to keep these
    /// attributes to a bare minimum to prevent unnecessary network traffic.
    /// </param><param>
    /// <B>NOTE:</B> This class is thread-safe since most of the fields are immutable. The attributes collection,
    /// the only field that is modified, is always locked during modifications to ensure thread-safety.
    /// </param>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [Serializable]
    public class Request 
    {
        /// <summary>
        /// The identifier for this request (may not be null or empty.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly string identifier; 

        /// <summary>
        /// The security key for this request (may be empty, but not null.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly string securityKey; 

        /// <summary>
        /// The type of this request (may be empty, but not null.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly string type; 

        /// <summary>
        /// A table mapping attribute keys to their values.
        /// </summary>
        /// <remarks>
        /// Due to the distributed nature of the request cache, all keys and values must be serializable.
        /// </remarks>
        private readonly IDictionary attributes = new Hashtable();

        /// <summary>
        /// Gets the identifier for this request.
        /// </summary>
        /// <value>A string representation of this request's identifier.</value>
        public string Identifier
        {
            get
            {
                return identifier;
            }
        }

        /// <summary>
        /// Gets the security key for this request.
        /// </summary>
        /// <value>A string representing the security key for this request (may be empty.)</value>
        public string SecurityKey
        {
            get
            {
                return securityKey;
            }
        }

        /// <summary>
        /// Gets the type of this request.
        /// </summary>
        /// <value>A string representing the type of this request (may be empty.)</value>
        public string Type
        {
            get
            {
                return type;
            }
        }

        /// <summary>
        /// Gets a copy of the set of attribute keys currently defined for this request.
        /// </summary>
        /// <value>An enumerator over a copy of all the defined attribute keys.</value>
        public IEnumerator AtrtributeKeys
        {
            get
            {
                lock (attributes.SyncRoot)
                {
                    return new ArrayList(attributes.Keys).GetEnumerator();
                }
            }
        }

        /// <summary>
        /// Creates a new <see cref="Request"/> object with the given values.
        /// </summary>
        /// <remarks>
        /// This is equivalent to <c>Request(identifier, securityKey, "")</c>.
        /// </remarks>
        /// <param name="identifier">The identifier for this request.</param>
        /// <param name="securityKey">The security key used to differentiate requests with the same ID.</param>
        /// <exception cref="ArgumentNullException"><paramref name="key"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="securityKey"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        public Request(string identifier, string securityKey)
            : this(identifier, securityKey, "")
        {
            // All work is delegated to the full constructor.
        }

        /// <summary>
        /// Creates a new <see cref="Request"/> object with the given values.
        /// </summary>
        /// <param name="identifier">The identifier for this request.</param>
        /// <param name="securityKey">The security key used to differentiate requests with the same ID.</param>
        /// <param name="type">The type of request this is (used for custom filtering.)</param>
        /// <exception cref="ArgumentNullException"><paramref name="key"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="securityKey"/> is <B>null</B>.<BR/>-or-<BR/>
        /// <paramref name="type"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException"><paramref name="identifier"/> is an empty string.</exception>
        public Request(string identifier, string securityKey, string type)
        {
            // Check the arguments.
            if (identifier == null)
            {
                throw new ArgumentNullException("identifier");
            }
            if (identifier == string.Empty)
            {
                throw new ArgumentException("The identifier cannot be the empty string.", "identifier");
            }
            if (securityKey == null)
            {
                throw new ArgumentNullException("securityKey");
            }
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }

            // Set the instance fields.
            this.identifier = identifier;
            this.securityKey = securityKey;
            this.type = type;
        }

        /// <summary>
        /// Gets the value of the given attribute.
        /// </summary>
        /// <param name="key">The attribute key to retrieve.</param>
        /// <returns>The attribute value for the given key, or <B>null</B> if it is not defined.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="key"/> is <B>null</B>.</exception>
        public object GetAttribute(object key)
        {
            lock (attributes.SyncRoot)
            {
                // Delegate to the IDictionary getter (which will throw ANE if key is null.)
                return attributes[key];
            }
        }

        /// <summary>
        /// Sets the value of the attribute to the given value.
        /// </summary>
        /// <remarks>
        /// Both the key and value must be <c>[Serializable]</c>.
        /// </remarks>
        /// <param name="key">The attribute key to set.</param>
        /// <param name="value">The new value for the attribute.</param>
        /// <exception cref="ArgumentNullException"><paramref name="key"/> is <B>null</B>.</exception>
        public void SetAttribute(object key, object value)
        {
            lock (attributes.SyncRoot)
            {
                // Delegate to the IDictionary setter (which will throw ANE if key is null.)
                attributes[key] = value;
            }
        }

        /// <summary>
        /// Removes an attribute and its value from the table.
        /// </summary>
        /// <param name="key">The attribute key to remove.</param>
        /// <exception cref="ArgumentNullException"><paramref name="key"/> is <B>null</B>.</exception>
        public void RemoveAttribute(object key)
        {
            lock (attributes.SyncRoot)
            {
                // Delegate to the IDictionary.Remove method (which will throw ANE if key is null.)
                attributes.Remove(key);
            }
        }

        /// <summary>
        /// Checks to see if an attribute has a value defined.
        /// </summary>
        /// <param name="key">The attribute key to check.</param>
        /// <returns><B>true</B> if the attribute has a defined value; otherwise, <B>false</B>.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="key"/> is <B>null</B>.</exception>
        public bool ContainsAttribute(object key)
        {
            lock (attributes.SyncRoot)
            {
                // Delegate to the IDictionary.Contains method (which will throw ANE if key is null.)
                return attributes.Contains(key);
            }
        }

        /// <summary>
        /// Clears the attribute table.
        /// </summary>
        public void ClearAttributes()
        {
            lock (attributes.SyncRoot)
            {
                // Delegate to the IDictionary.Clear method.
                attributes.Clear();
            }
        }
    }
}
