/*
 * Response.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// A generic response to a request, used by the <see cref="IdempotencyManager"/>.
    /// </summary>
    /// <remarks>
    /// <param>
    /// The <B>Response</B> object wraps a generic response object to allow maximum flexibility. The
    /// only requirement is that the response object be serializable so that it can be saved to
    /// persistent storage. The response also includes a timestamp which indicates when the response
    /// was originally created as well as a lifetime <see cref="TimeSpan"/> which can help applications
    /// determine when cached responses should be refreshed and/or removed.
    /// </param><param>
    /// <B>NOTE:</B> This class is thread-safe since all its fields are immutable.
    /// </param>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [Serializable]
    public class Response 
    {
        /// <summary>
        /// The wrapped response object (must be serializable.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly object responseObject; 

        /// <summary>
        /// The time at which the response object was created.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly DateTime timestamp; 

        /// <summary>
        /// The lifetime of the response object (ignored by the <see cref="IdempotencyManager"/>.)
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly TimeSpan life; 

        /// <summary>
        /// Gets the inner response object.
        /// </summary>
        /// <value>The response object.</value>
        public object ResponseObject
        {
            get
            {
                return responseObject;
            }
        }

        /// <summary>
        /// Gets the time at which this object was created.
        /// </summary>
        /// <value>The time this <see cref="Response"/> was created.</value>
        public DateTime Timestamp
        {
            get
            {
                return timestamp;
            }
        }

        /// <summary>
        /// Gets the lifetime of the inner response object (for use by the application.)
        /// </summary>
        /// <value>The lifetime the response should be used for.</value>
        public TimeSpan Life
        {
            get
            {
                return life;
            }
        }

        /// <summary>
        /// Creates a new <see cref="Response"/> object which wraps the given parameter.
        /// </summary>
        /// <remarks>
        /// This is equivalent to <c>Response(responseObject, DateTime.Now, TimeSpan.MaxValue)</c>.
        /// </remarks>
        /// <param name="responseObject">The inner response object (must be serializable.)</param>
        /// <exception cref="ArgumentNullException"><paramref name="responseObject"/> is <B>null</B>.</exception>
        public Response(object responseObject)
            : this(responseObject, DateTime.Now, TimeSpan.MaxValue)
        {
            // All work is delegated to the full constructor.
        }

        /// <summary>
        /// Creates a new <see cref="Response"/> object from the given parameters.
        /// </summary>
        /// <remarks>
        /// This is equivalent to <c>Response(responseObject, DateTime.Now, life)</c>.
        /// </remarks>
        /// <param name="responseObject">The inner response object (must be serializable.)</param>
        /// <param name="life">The lifetime of the response object.</param>
        /// <exception cref="ArgumentNullException"><paramref name="responseObject"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="life"/> is negative.</exception>
        public Response(object responseObject, TimeSpan life)
            : this(responseObject, DateTime.Now, life)
        {
            // All work is delegated to the full constructor.
        }

        /// <summary>
        /// Creates a new <see cref="Response"/> object from the given parameters.
        /// </summary>
        /// <param name="responseObject">The inner response object (must be serializable.)</param>
        /// <param name="timestamp">The time at which the response was created.</param>
        /// <param name="life">The lifetime of the response object.</param>
        /// <exception cref="ArgumentNullException"><paramref name="responseObject"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentOutOfRangeException"><paramref name="life"/> is negative.</exception>
        public Response(object responseObject, DateTime timestamp, TimeSpan life)
        {
            // Check the arguments.
            if (responseObject == null)
            {
                throw new ArgumentNullException("responseObject");
            }
            if (life < TimeSpan.Zero)
            {
                throw new ArgumentOutOfRangeException("life", life, "The TimeSpan must not be negative.");
            }

            // Set the instance fields.
            this.responseObject = responseObject;
            this.timestamp = timestamp;
            this.life = life;
        }
    }
}
