/*
 * TypeMappingCacheStrategy.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using System.Diagnostics;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency.Strategy
{
    /// <summary>
    /// <see cref="ICacheStrategy"/> implementation which determines acceptance based on the request type.
    /// </summary>
    /// <remarks>
    /// <para>
    /// A mapping from <see cref="Request.Type"/> to <see cref="CacheDecision"/> is used to determine if
    /// a request should be allowed to be cached or not. There is also a default <B>CacheDecision</B> which 
    /// is used to determine if a request should be cached when its type does not appear in the mapping.
    /// </para><para>
    /// <B>NOTE:</B> This class is thread-safe since all internal data is immutable.
    /// </para>
    /// </remarks>
    /// <author>WishingBone</author>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class TypeMappingCacheStrategy : ICacheStrategy
    {
        /// <summary>
        /// The mapping from <see cref="Request.Type"/>s to <see cref="CacheDecision"/>s.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter.
        /// </remarks>
        private readonly IDictionary decisions;

        /// <summary>
        /// The default decision which is used when the mapping above does not contain an entry for a given type.
        /// </summary>
        /// <remarks>
        /// This value is set in the constructor and is immutable thereafter. It may be set to either
        /// <see cref="CacheDecision.Accept"/> or <see cref="CacheDecision.Deny"/>, but not
        /// <see cref="CacheDecision.Default"/> (since it is the default value.)
        /// </remarks>
        private readonly CacheDecision defaultDecision;

        /// <summary>
        /// Creates a new <see cref="TypeMappingCacheStrategy"/> instance based on config file parameters.
        /// </summary>
        /// <remarks>
        /// Configuration values should appear in the 
        /// <c>TopCoder.Util.Idempotency.Strategy.TypeMappingCacheStrategy</c> namespace. This namespace 
        /// may contain either a <c>known_types</c> property, which should be a list of <see cref="Request.Type"/>
        /// string, or a <c>default</c> property which contains a <see cref="CacheDecision"/> value that will be
        /// used as the default decision, or both. If the <c>known_types</c> property is present, then there must
        /// also be properties for each value in its value list. These property names should be the type name,
        /// followed by the <c>"_decision"</c> suffix (for example, the <c>"GET"</c> value would require an
        /// associated property with the name <c>"GET_decision"</c>.) These associated properties should have a
        /// <B>CacheDecision</B> value which is added to the decision map along with the decison type name.
        /// </remarks>
        /// <exception cref="InvalidConfigFileException">The configuration is invalid.</exception>
        public TypeMappingCacheStrategy()
        {
            // Create a decisions table to add CacheDecisions to.
            decisions = new Hashtable();

            // Get the namespace.
            Namespace ns = ConfigHelper.GetNamespace(GetType().FullName);

            // Get the "known_types" property and, if it exsists, use it to fill the decisions Hashtable.
            string[] knownTypes = ConfigHelper.GetValues(ns, "known_types", false);

            if (knownTypes != null)
            {
                foreach (string type in knownTypes)
                {
                    // Make sure the 'type' is a valid string.
                    if (type == string.Empty)
                    {
                        throw new InvalidConfigFileException("known_types in " + GetType().FullName +
                            " contains an Empty string.");
                    }

                    // Get the "XXX_decision" property for the current type and add it to the hashtable.
                    decisions.Add(type, ConfigHelper.GetEnumValue(ns, type + "_decision", typeof(CacheDecision)));
                }
            }

            // Get the default decision (or use Accept if there is no "default" property.)
            defaultDecision = (CacheDecision) ConfigHelper.GetEnumValue(
                ns, "default", typeof(CacheDecision), CacheDecision.Accept);
        }

        /// <summary>
        /// Creates a new <see cref="TypeMappingCacheStrategy"/> instance based on the given parameters.
        /// </summary>
        /// <param name="decisions">A mapping from <see cref="Request.Type"/> to <see cref="CacheDecision"/>.</param>
        /// <param name="defaultDecision">The decision to use when a type is not found in the 
        /// <paramref name="decisions"/> mapping.</param>
        /// <exception cref="ArgumentNullException"><paramref name="decisions"/> is <B>null</B>.</exception>
        /// <exception cref="ArgumentException">
        /// <paramref name="defaultDecision"/> is neither <B>Accept</B> nor <B>Deny</B>.<BR/>-or-<BR/>
        /// <paramref name="decisions"/> contains an entry with an empty or non-<B>string</B> key.<BR/>-or-<BR/>
        /// <paramref name="decisions"/> contains an entry with a non-<B>CacheDecision</B> value.</exception>
        public TypeMappingCacheStrategy(IDictionary decisions, CacheDecision defaultDecision)
        {
            // Perform basic argument checks.
            if (decisions == null)
            {
                throw new ArgumentNullException("decisions", "A IDictionary must be specified.");
            }
            if ((defaultDecision != CacheDecision.Accept) && 
                (defaultDecision != CacheDecision.Deny))
            {
                throw new ArgumentException("The default decision must be either Accept or Deny.", 
                    "defaultDecision");
            }

            // Copy the dictionary before checking its internals.
            // (This prevents outside modifications during iteration.)
            this.decisions = new Hashtable(decisions);
            this.defaultDecision = defaultDecision;

            // Now check the contents of the dictionary.
            foreach (DictionaryEntry entry in this.decisions)
            {
                // Check the key. The keys must be non-empty string.
                if (!(entry.Key is string))
                {
                    throw new ArgumentException("Keys must be strings.", "decisions");
                }
                if (((string) entry.Key) == string.Empty)
                {
                    throw new ArgumentException("Empty key strings are not allowed.", "decisions");
                }

                // Check the value. It must be one of the defined enumerants in CacheDecision.
                if (entry.Value == null)
                {
                    throw new ArgumentException("The value is null.", "decisions[" + entry.Key + "]");
                }
                if (!Enum.IsDefined(typeof(CacheDecision), entry.Value))
                {
                    throw new ArgumentException("The value is not defined in CacheDecision.",
                        "decisions[" + entry.Key + "]");
                }
            }
        }

        /// <summary>
        /// Determines if the given request should be cached or not.
        /// </summary>
        /// <remarks>
        /// <see cref="Request.Type"/> is looked up in the decisions mapping and, if found, the result
        /// is used to determine if the request should be accepted or denied. If the type maps to
        /// <see cref="CacheDecision.Default"/>, or if there is no mapping at all, then the default
        /// decision is used to determine if the request should be accepted.
        /// </remarks>
        /// <param name="request">The request to determine if it should be cached.</param>
        /// <returns><B>true</B> if the request may be cached; otherwise, <B>false</B>.</returns>
        /// <exception cref="ArgumentNullException"><paramref name="request"/> is <B>null</B>.</exception>
        public bool Accept(Request request)
        {
            // Check the argument.
            if (request == null)
            {
                throw new ArgumentNullException("request", "A request must be specified.");
            }

            // Lookup the identifier in our decisions table.
            object value = decisions[request.Type];

            // If there is no entry, use the default decision; otherwise, cast the value.
            CacheDecision decision = (value == null) ? defaultDecision : (CacheDecision) value;

            // If the entry was set to Default, change the decision to the default decision.
            if (decision == CacheDecision.Default)
            {
                decision = defaultDecision;
            }

            // At this point, the decision value should be either Accept for Deny.
            Debug.Assert((decision == CacheDecision.Accept) || (decision == CacheDecision.Deny),
                "Either a decisions entry was not a defined CacheDecision value, " +
                "or the defaultDecision was something other than Accept or Deny.");

            // Return true if the decision is Accept; otherwise, false.
            return decision == CacheDecision.Accept;
        }
    }
}
