/*
 * Demo.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.IO;
using System.Data;
using NUnit.Framework;
using TopCoder.Data.ConnectionFactory;
using TopCoder.Util.Idempotency.Logger;
using TopCoder.Util.Idempotency.Persistence;
using TopCoder.Util.Idempotency.Strategy;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Demonstration tests of this component.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class Demo
    {
        /// <summary>
        /// Before all tests make sure to clear the config.
        /// </summary>
        [TestFixtureSetUp]
        public void RunBeforeAllTests()
        {
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// Load the standard connection factory configuration and the demo config.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            TestHelper.Config.LoadFile("../../test_files/CF-standard.xml");
            TestHelper.Config.LoadFile("../../test_files/Demo.xml");
        }

        /// <summary>
        /// Cleanup by disposing the IdempotencyManager instance, deleting the database file, and clearin the config.
        /// </summary>
        [TearDown]
        public void RunAfterEachTest()
        {
            IdempotencyManager.Instance.Dispose();
            File.Delete("../../test_files/demo.db");
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// The designer's demo section, slightly modified to compile and rearanged so that
        /// responses are actually retrieved.
        /// </summary>
        [Test]
        public void DesignerDemo()
        {
            // Get Idempotency Manager
            IdempotencyManager manager = IdempotencyManager.Instance;

            // New request arrives
            Request request = new Request("/importal/readme.txt", "", "GET");
            TimeSpan elapsed = manager.RequestArrival(request);
            bool exists = manager.HasResponse(request);

            // Store response
            Response response = new Response("hello world", DateTime.Now, TimeSpan.FromDays(1));
            manager.StoreResponse(request, response);

            // Retrieve response
            response = manager.RetrieveResponse(request, TimeSpan.FromSeconds(5));
            manager.RetrieveResponse(request, new ResponseAvailable(ResponseHandler), TimeSpan.FromMinutes(10));

            // Manage cache
            manager.DropEntry("/importal/readme.txt");
            manager.ClearCache();
            manager.ShrinkCacheToSize(50);
            manager.ShrinkCacheToDate(DateTime.Now - TimeSpan.FromDays(5));
            manager.PurgeResponsePersistence();
            manager.SaveCache();
            manager.LoadCache();

            // Manage pluggable sources
            IDbConnection conn = ConnectionManager.Instance.CreatePredefinedDbConnection("standard");
            DatabasePersistence persistence = new DatabasePersistence(conn);
            manager.ResponsePersistence = persistence;
            manager.CachePersistence = persistence;
            manager.CacheStrategy = new TypeMappingCacheStrategy();
            manager.CacheManagement = new GenericCacheManagement();
            manager.ActivityLogger = new ActivityStatistics();
        }

        /// <summary>
        /// Dummy callback used in the demo method above.
        /// </summary>
        /// <param name="request">Ignored.</param>
        /// <param name="response">Ignored.</param>
        private void ResponseHandler(Request request, Response response)
        {
            // Do nothing.
        }
    }
}
