/*
 * GenericCacheManagementTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using NUnit.Framework;
using TopCoder.Cache.Distributed;
using TopCoder.Util.Idempotency.Persistence;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Tests of the <see cref="GenericCacheManagement"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class GenericCacheManagementTests
    {
        /// <summary>
        /// An instance of the class being tested. Since it has no internal state, there is no need to
        /// recreate the object in a SetUp method before each test.
        /// </summary>
        private readonly GenericCacheManagement gcm = new GenericCacheManagement();

        /// <summary>
        /// Trying to shrink a null cache should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestShrinkCacheToSizeNullArg()
        {
            gcm.ShrinkCacheToSize(null, 5);
        }

        /// <summary>
        /// Trying to shrink a cache to a negative size should throw an AOORE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestShrinkCacheToSizeNegArg()
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                gcm.ShrinkCacheToSize(cache, -1);
            }
        }

        /// <summary>
        /// Shrinking a cache with valid values should work and remove the oldest element(s).
        /// </summary>
        [Test]
        public void TestShrinkCacheToSize()
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                cache.Put("key1", new CacheEntry(new DateTime(1000), false));
                cache.Put("key2", new CacheEntry(new DateTime(5000), false));

                Assert.AreEqual(2, cache.Keys.Count, "Wrong number of cache entries.");

                gcm.ShrinkCacheToSize(cache, 1);

                // FIXME: We can't just check the Keys.Count again since the key set is not updated properly.
                //Assert.AreEqual(1, cache.Keys.Count);

                Assert.IsNull(cache.Get("key1"), "'key1' should have been removed.");
                Assert.IsNotNull(cache.Get("key2"), "'key2' should NOT have been removed.");
            }
        }

        /// <summary>
        /// Trying to shrink a null cache should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestShrinkCacheToDateNullArg()
        {
            gcm.ShrinkCacheToDate(null, DateTime.Now);
        }

        /// <summary>
        /// Shrinking a cache with valid values should work and remove all elements older than the parameter.
        /// </summary>
        [Test]
        public void TestShrinkCacheToDate()
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                cache.Put("key1", new CacheEntry(new DateTime(1000), false));
                cache.Put("key2", new CacheEntry(new DateTime(5000), false));

                Assert.AreEqual(2, cache.Keys.Count, "Wrong number of cache entries.");

                gcm.ShrinkCacheToDate(cache, new DateTime(2500));

                // FIXME: We can't just check the Keys.Count again since the key set is not updated properly.
                //Assert.AreEqual(1, cache.Keys.Count);

                Assert.IsNull(cache.Get("key1"), "'key1' should have been removed.");
                Assert.IsNotNull(cache.Get("key2"), "'key2' should NOT have been removed.");
            }
        }

        /// <summary>
        /// Trying to purge a null cache should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestPurgeResponsePersistenceArg1Null()
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                using (CachedResponsePersistence persistence = 
                           new CachedResponsePersistence(new DistributedSimpleCache("tcp://inner", null), null))
                {
                    gcm.PurgeResponsePersistence(null, persistence);
                }
            }
        }

        /// <summary>
        /// Trying to purge a null persistence should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestPurgeResponsePersistenceArg2Null()
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                using (CachedResponsePersistence persistence = 
                           new CachedResponsePersistence(new DistributedSimpleCache("tcp://inner", null), null))
                {
                    gcm.PurgeResponsePersistence(cache, null);
                }
            }
        }

        /// <summary>
        /// Purging with valid parameters should remove the stored responses which do not appear in the cache.
        /// </summary>
        [Test]
        public void TestPurgeResponsePersistence()
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                using (CachedResponsePersistence persistence = 
                           new CachedResponsePersistence(new DistributedSimpleCache("tcp://inner", null), null))
                {
                    // Add an entry to the persistence only.
                    persistence.Store(new Request("id", ""), new Response("obj"));

                    // Make sure the proper key is enumerated.
                    IEnumerator it = persistence.GetIdentifiers();
                    Assert.IsTrue(it.MoveNext(), "There should be one ID in the collection.");
                    Assert.AreEqual("id", it.Current, "Wrong ID in the collection.");
                    Assert.IsFalse(it.MoveNext(), "There are too many IDs in the collection.");

                    // Purge the persistence.
                    gcm.PurgeResponsePersistence(cache, persistence);

                    // Make sure the key has been purged.
                    // FIXME: Again, the DSC.Keys bug appears to prevent this simple test from working.
                    // Instead, we will have to actually try and get the response which will cause a
                    // ResponseRetrievalException to be thrown, thus proving the purge worked.
                    //it = persistence.GetIdentifiers();
                    //Assert.IsFalse(it.MoveNext(), "The collection should not have any keys anymore.");
                    try
                    {
                        persistence.Retrieve(new Request("id", ""));
                        Assert.Fail("An exception should have been thrown.");
                    }
                    catch (ResponseRetrievalException)
                    {
                        // The response should no longer be there, so this exception is expected.
                    }
                }
            }
        }
    }
}
