/*
 * IdempotencyManagerTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using NUnit.Framework;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Tests of the <see cref="IdempotencyManager"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class IdempotencyManagerTests
    {
        /// <summary>
        /// Before all tests make sure to clear the config setup.
        /// </summary>
        [TestFixtureSetUp]
        public void RunBeforeAllTests()
        {
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// Load the standard connection factory configuration and the test config.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            TestHelper.Config.LoadFile("../../test_files/CF-standard.xml");
            TestHelper.Config.LoadFile("../../test_files/IdempotencyManagerTests.xml");
        }

        /// <summary>
        /// Cleanup by disposing of the manager and clearing the config.
        /// </summary>
        [TearDown]
        public void RunAfterEachTest()
        {
            IdempotencyManager.Instance.Dispose();
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// Get the instance (which forces construction.)
        /// </summary>
        [Test]
        public void TestCtor()
        {
            IdempotencyManager dummy = IdempotencyManager.Instance;
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRequestArrivalNullArg()
        {
            IdempotencyManager.Instance.RequestArrival(null);
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestHasResponseNullArg()
        {
            IdempotencyManager.Instance.HasResponse(null);
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRetrieveResponse1NullArg()
        {
            IdempotencyManager.Instance.RetrieveResponse(null);
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRetrieveResponse2NullArg()
        {
            IdempotencyManager.Instance.RetrieveResponse(null, TimeSpan.MaxValue);
        }

        /// <summary>
        /// Calling the method with a negative timespan should throw an AOORE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestRetrieveResponse2NegSpan()
        {
            IdempotencyManager.Instance.RetrieveResponse(new Request("id", ""), TimeSpan.MinValue);
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRetrieveResponse3Arg1Null()
        {
            IdempotencyManager.Instance.RetrieveResponse(
                null, new ResponseAvailable(this.ResponseHandler), TimeSpan.MaxValue);
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRetrieveResponse3Arg2Null()
        {
            IdempotencyManager.Instance.RetrieveResponse(new Request("id", ""), null, TimeSpan.MaxValue);
        }

        /// <summary>
        /// Calling the method with a negative timespan should throw an AOORE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestRetrieveResponse3NegSpan()
        {
            IdempotencyManager.Instance.RetrieveResponse(
                new Request("id", ""), new ResponseAvailable(this.ResponseHandler), TimeSpan.MinValue);
        }

        /// <summary>
        /// Dummy callback used in the tests above.
        /// </summary>
        /// <param name="request">Ignored.</param>
        /// <param name="response">Ignored.</param>
        private void ResponseHandler(Request request, Response response)
        {
            // Do nothing.
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestStoreResponseArg1Null()
        {
            IdempotencyManager.Instance.StoreResponse(null, new Response("obj"));
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestStoreResponseArg2Null()
        {
            IdempotencyManager.Instance.StoreResponse(new Request("id", ""), null);
        }

        /// <summary>
        /// Calling the method with a null argument should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestDropEntryArgNull()
        {
            IdempotencyManager.Instance.DropEntry(null);
        }

        /// <summary>
        /// Calling the method with an empty argument should throw an AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestDropEntryArgEmpty()
        {
            IdempotencyManager.Instance.DropEntry("");
        }

        /// <summary>
        /// Calling the method with a negative count should throw an AOORE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestShrinkCacheToSizeNegCount()
        {
            IdempotencyManager.Instance.ShrinkCacheToSize(-1);
        }
    }
}
