/*
 * ActivityBroadcasterTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using NUnit.Framework;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// Tests of the <see cref="ActivityBroadcaster"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class ActivityBroadcasterTests
    {
        /// <summary>
        /// The logger which is tested in several methods below.
        /// </summary>
        private IActivityLogger logger;

        /// <summary>
        /// A test logger which keeps track of if its various methods have been called.
        /// </summary>
        private LogTester tester;

        /// <summary>
        /// Before all tests and after each one make sure to clear the config manager.
        /// </summary>
        [TestFixtureSetUp, TearDown]
        public void RunBeforeAllTestsAndAfterEachTest()
        {
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// Recreate the tester and logger objects for each test.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            tester = new LogTester();
            logger = new ActivityBroadcaster(new object[]{tester});
        }

        /// <summary>
        /// Using the default constructor with a valid config should work fine.
        /// </summary>
        [Test]
        public void TestCtor0()
        {
            // Load a config file and ensure that the default constructor works with it.
            TestHelper.Config.LoadFile("../../test_files/ActivityBroadcasterTests.xml");

            new ActivityBroadcaster();
        }

        /// <summary>
        /// Using the 1-arg constructor with a null param should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor1NullArg()
        {
            new ActivityBroadcaster(null);
        }

        /// <summary>
        /// Using the 1-arg constructor with an empty list should throw AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor1Empty()
        {
            new ActivityBroadcaster(new object[]{});
        }

        /// <summary>
        /// Using the 1-arg constructor with a non-IActivityLogger should throw AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor1NonLogger()
        {
            new ActivityBroadcaster(new object[]{null});
        }

        /// <summary>
        /// Using the 1-arg constructor with a proper list should work fine.
        /// </summary>
        [Test]
        public void TestCtor1()
        {
            new ActivityBroadcaster(new object[]{new ActivityStatistics()});
        }

        /// <summary>
        /// Make sure the broadcaster really broadcasts to the correct underlying loggers without checking parameters.
        /// </summary>
        [Test]
        public void TestRequestReceived()
        {
            Assert.IsFalse(tester.WasReceived, "A request has already been received.");

            logger.RequestReceived(null);

            Assert.IsTrue(tester.WasReceived, "The request was not received.");
        }

        /// <summary>
        /// Make sure the broadcaster really broadcasts to the correct underlying loggers without checking parameters.
        /// </summary>
        [Test]
        public void TestRequestStored()
        {
            Assert.IsFalse(tester.WasStored, "A request has already been stored.");

            logger.ResponseStored(null, null);

            Assert.IsTrue(tester.WasStored, "The request was not stored.");
        }

        /// <summary>
        /// Make sure the broadcaster really broadcasts to the correct underlying loggers without checking parameters.
        /// </summary>
        [Test]
        public void TestResponseRetrieved()
        {
            Assert.IsFalse(tester.WasRetrieved, "A response has already been retrieved.");

            logger.ResponseRetrieved(null, null);

            Assert.IsTrue(tester.WasRetrieved, "The response was not retrieved.");
        }

        /// <summary>
        /// Make sure the broadcaster really broadcasts to the correct underlying loggers without checking parameters.
        /// </summary>
        [Test]
        public void TestExceptionOccured()
        {
            Assert.IsFalse(tester.WasException, "An exception has already been received.");

            logger.ExceptionOccurred(null, null);

            Assert.IsTrue(tester.WasException, "The exception was not received.");
        }

        /// <summary>
        /// Make sure the broadcaster really broadcasts to the correct underlying loggers without checking parameters.
        /// </summary>
        [Test]
        public void TestDispose()
        {
            Assert.IsFalse(tester.WasDisposed, "The object has already been disposed.");

            logger.Dispose();

            Assert.IsTrue(tester.WasDisposed, "The object was not disposed.");
        }

        /// <summary>
        /// Simple logger which just keeps track of which methods have been called.
        /// </summary>
        private class LogTester : IActivityLogger
        {
            /// <summary>
            /// Set to true when the RequestReceived method is called.
            /// </summary>
            public bool WasReceived = false;

            /// <summary>
            /// Set to true when the RequestStored method is called.
            /// </summary>
            public bool WasStored = false;

            /// <summary>
            /// Set to true when the RequestRetrieved method is called.
            /// </summary>
            public bool WasRetrieved = false;

            /// <summary>
            /// Set to true when the ExceptionOccurred method is called.
            /// </summary>
            public bool WasException = false;

            /// <summary>
            /// Set to true when the Dispose method is called.
            /// </summary>
            public bool WasDisposed = false;

            /// <summary>
            /// Sets WasReceived to true.
            /// </summary>
            /// <param name="request"></param>
            public void RequestReceived(Request request)
            {
                WasReceived = true;
            }

            /// <summary>
            /// Sets WasStored to true.
            /// </summary>
            /// <param name="request">Ignored.</param>
            /// <param name="response">Ignored.</param>
            public void ResponseStored(Request request, Response response)
            {
                WasStored = true;
            }

            /// <summary>
            /// Sets WasRetrieved to true.
            /// </summary>
            /// <param name="request">Ignored.</param>
            /// <param name="response">Ignored.</param>
            public void ResponseRetrieved(Request request, Response response)
            {
                WasRetrieved = true;
            }

            /// <summary>
            /// Sets WasException to true.
            /// </summary>
            /// <param name="request">Ignored.</param>
            /// <param name="exception">Ignored.</param>
            public void ExceptionOccurred(Request request, Exception exception)
            {
                WasException = true;
            }

            /// <summary>
            /// Sets WasDisposed to true.
            /// </summary>
            public void Dispose()
            {
                WasDisposed = true;
            }
        }
    }
}
