/*
 * ActivityLoggingTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Diagnostics;
using NUnit.Framework;
using TopCoder.LoggingWrapper;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// Tests of the <see cref="ActivityLoggingTests"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class ActivityLoggingTests
    {
        /// <summary>
        /// The log name.
        /// </summary>
        private const string LogName = "ActivityLogging";

        /// <summary>
        /// The EventLog instance.
        /// </summary>
        private EventLog logger = null;

        /// <summary>
        /// An instance of the test class used in most tests.
        /// </summary>
        private ActivityLogging logging;

        /// <summary>
        /// A dummy request to log.
        /// </summary>
        private Request request;

        /// <summary>
        /// A dummy response to log.
        /// </summary>
        private Response response;

        /// <summary>
        /// Before all tests and after each one make sure to clear the config manager.
        /// </summary>
        [TestFixtureSetUp, TearDown]
        public void RunBeforeAllTestsAndAfterEachTest()
        {
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// Before each test, we setup the Logging Wrapper again, just to be safe.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            // Delete any old event logs.
            if (EventLog.Exists(LogName))
            {
                EventLog.Delete(LogName);
            }

            // Create a new debug configuration for the logger.
            LogManager.Configuration = new System.Collections.Specialized.NameValueCollection();
            LogManager.Configuration.Add(LogManager.CLASS_PARAMETER, "TopCoder.LoggingWrapper.DiagnosticImpl");
            LogManager.Configuration.Add("Source", LogName);
            LogManager.Configuration.Add("LogName", LogName);
            LogManager.LoadConfiguration();

            // Create the log.
            LogManager.Log("dummy");
            logger = new EventLog(LogName);
            logger.Clear();

            // Create a new logger.
            logging = new ActivityLogging("|{0}|{1}|{2}|{3}|", true);

            // Create a dummy request.
            request = new Request("id", "key", "type");

            // Create a dummy response.
            response = new Response("obj");
        }

        /// <summary>
        /// Test the default constructor with a valid config file to ensure there are no errors.
        /// </summary>
        [Test]
        public void TestCtor0()
        {
            // Load a config file and ensure that the default constructor works with it.
            TestHelper.Config.LoadFile("../../test_files/ActivityLoggingTests.xml");

            new ActivityLogging();
        }

        /// <summary>
        /// Test the 1-arg constructor with a null format (should throw ANE.)
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor1NullArg()
        {
            new ActivityLogging(null);
        }

        /// <summary>
        /// Test the 1-arg constructor with an invalid format (too many params--should throw an AE.)
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor1Invalid()
        {
            new ActivityLogging("{0}{1}{2}{3}{4}");
        }

        /// <summary>
        /// Test the 1-arg constructor with a valid format that uses all the args.
        /// </summary>
        [Test]
        public void TestCtor1()
        {
            new ActivityLogging("{0}{1}{2}{3}");
        }

        /// <summary>
        /// Test the 2-arg constructor with a null format (should throw ANE.)
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor2NullArg()
        {
            new ActivityLogging(null, true);
        }

        /// <summary>
        /// Test the 2-arg constructor with an invalid format (too many params--should throw an AE.)
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2Invalid()
        {
            new ActivityLogging("{0}{1}{2}{3}", false);
        }

        /// <summary>
        /// Test the 2-arg constructor with a valid format that uses all the args.
        /// </summary>
        [Test]
        public void TestCtor2()
        {
            new ActivityLogging("{0}{1}{2}{3}", true);
        }

        /// <summary>
        /// Trying to log a null request should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRequestReceivedNullArg()
        {
            logging.RequestReceived(null);
        }

        /// <summary>
        /// Test logging a valid request received message.
        /// </summary>
        [Test]
        public void TestRequestReceived()
        {
            logging.RequestReceived(request);

            CheckLogMessage("RECEIVED: ", "");
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseStoredNullArg1()
        {
            logging.ResponseStored(null, response);
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseStoredNullArg2()
        {
            logging.ResponseStored(request, null);
        }

        /// <summary>
        /// Test logging a valid response stored message.
        /// </summary>
        [Test]
        public void TestResponseStored()
        {
            logging.ResponseStored(request, response);

            CheckLogMessage("STORED: ", "");
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseRetrievedNullArg1()
        {
            logging.ResponseRetrieved(null, response);
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseRetrievedNullArg2()
        {
            logging.ResponseRetrieved(request, null);
        }

        /// <summary>
        /// Test logging a valid response retrieved message.
        /// </summary>
        [Test]
        public void TestResponseRetrieved()
        {
            logging.ResponseRetrieved(request, response);

            CheckLogMessage("RETRIEVED: ", "");
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestExceptionOccurredNullArg1()
        {
            logging.ExceptionOccurred(null, new Exception());
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestExceptionOccurredNullArg2()
        {
            logging.ExceptionOccurred(request, null);
        }

        /// <summary>
        /// Test logging a valid exception message.
        /// </summary>
        [Test]
        public void TestExceptionOccurred()
        {
            Exception e = new Exception("message");

            logging.ExceptionOccurred(request, e);

            CheckLogMessage("ERROR: ", e.Message);
        }

        /// <summary>
        /// Checks the event log to ensure the proper log message is present.
        /// </summary>
        /// <remarks>
        /// This method ensures that the event log contains only one entry (the one we are validating.) It
        /// then takes the log message and splits it into its 6 parts and verifies each one separately. The
        /// timestamp is validated by ensuring that it is from less than a minute before the current time.
        /// </remarks>
        /// <param name="prefix">The expected prefix message.</param>
        /// <param name="postfix">The expected postfix message.</param>
        private void CheckLogMessage(string prefix, string postfix)
        {
            // Check the length.
            Assert.AreEqual(1, logger.Entries.Count, "Wrong number of log entries.");

            // Split the message into its parts.
            string[] parts = logger.Entries[0].Message.Split('|');

            // There should be 6 parts.
            Assert.AreEqual(6, parts.Length, "Wrong number of message parts.");
            
            // Check each of the 6 parts.
            Assert.AreEqual(prefix, parts[0], "Wrong log prefix.");
            Assert.AreEqual(request.Identifier, parts[1], "Wrong logged identifier.");
            Assert.AreEqual(request.SecurityKey, parts[2], "Wrong logged security key.");
            Assert.AreEqual(request.Type, parts[3], "Wrong logged type.");
            Assert.IsTrue(DateTime.Now - DateTime.Parse(parts[4]) < TimeSpan.FromMinutes(1), 
                "The timestamp is off by too much.");
            Assert.AreEqual(postfix, parts[5], "Wrong log postfix.");
        }
    }
}
