/*
 * ActivityStatisticsTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using NUnit.Framework;

namespace TopCoder.Util.Idempotency.Logger
{
    /// <summary>
    /// Tests of the <see cref="ActivityStatistics"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class ActivityStatisticsTests
    {
        /// <summary>
        /// An instance of the test class used in most tests.
        /// </summary>
        private ActivityStatistics logging;

        /// <summary>
        /// A dummy request to log.
        /// </summary>
        private Request request;

        /// <summary>
        /// A dummy response to log.
        /// </summary>
        private Response response;

        /// <summary>
        /// Before all tests and after each one make sure to clear the config manager.
        /// </summary>
        [TestFixtureSetUp, TearDown]
        public void RunBeforeAllTestsAndAfterEachTest()
        {
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// Before each test, we reinitialize all variables (just to be safe.)
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            // Create a new logger.
            logging = new ActivityStatistics();

            // Create a dummy request.
            request = new Request("id", "key", "type");

            // Create a dummy response.
            response = new Response("obj");
        }

        /// <summary>
        /// Trying to log a null request should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestRequestReceivedNullArg()
        {
            logging.RequestReceived(null);
        }

        /// <summary>
        /// Test logging a valid request received message.
        /// </summary>
        [Test]
        public void TestRequestReceived()
        {
            Assert.AreEqual(0, logging.GetReceiveCount(request.Identifier), "A request is already logged.");
            Assert.AreEqual(0, logging.ReceiveCounts.Count, "Other requests are logged.");

            logging.RequestReceived(request);

            Assert.AreEqual(1, logging.GetReceiveCount(request.Identifier), "The reqeust was not logged.");
            Assert.AreEqual(1, logging.ReceiveCounts.Count, "Other requests were logged.");
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseStoredNullArg1()
        {
            logging.ResponseStored(null, response);
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseStoredNullArg2()
        {
            logging.ResponseStored(request, null);
        }

        /// <summary>
        /// Test logging a valid response stored message.
        /// </summary>
        [Test]
        public void TestResponseStored()
        {
            Assert.AreEqual(0, logging.GetStoreCount(request.Identifier), "A response is already logged.");
            Assert.AreEqual(0, logging.StoreCounts.Count, "Other responses are logged.");

            logging.ResponseStored(request, response);

            Assert.AreEqual(1, logging.GetStoreCount(request.Identifier), "The store was not logged.");
            Assert.AreEqual(1, logging.StoreCounts.Count, "Other responses were logged.");
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseRetrievedNullArg1()
        {
            logging.ResponseRetrieved(null, response);
        }

        /// <summary>
        /// Logging with a null parameter should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResponseRetrievedNullArg2()
        {
            logging.ResponseRetrieved(request, null);
        }

        /// <summary>
        /// Test logging a valid response retrieved message.
        /// </summary>
        [Test]
        public void TestResponseRetrieved()
        {
            Assert.AreEqual(0, logging.GetRetrieveCount(request.Identifier), "A response is already logged.");
            Assert.AreEqual(0, logging.RetrieveCounts.Count, "Other responses are logged.");

            logging.ResponseRetrieved(request, response);

            Assert.AreEqual(1, logging.GetRetrieveCount(request.Identifier), "The response was not logged.");
            Assert.AreEqual(1, logging.RetrieveCounts.Count, "Other responses were logged.");
        }

        /// <summary>
        /// Logging with null parameter should work here since statistics are not kept on exceptions.
        /// </summary>
        [Test]
        public void TestExceptionOccurredNullArgs()
        {
            logging.ExceptionOccurred(null, null);
        }

        /// <summary>
        /// Trying to reset a null counter should throw ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestResetCountNullArg()
        {
            logging.ResetCount(null);
        }

        /// <summary>
        /// Trying to reset an empty counter should throw an AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestResetCountEmptyArg()
        {
            logging.ResetCount("");
        }

        /// <summary>
        /// Tests the ResetCount method to make sure that it resets the counts of the given ID and
        /// only of that ID.
        /// </summary>
        [Test]
        public void TestResetCount()
        {
            // Update some counters.
            logging.RequestReceived(request);
            logging.ResponseStored(request, response);
            logging.ResponseRetrieved(request, response);

            Request request2 = new Request("id2", "key", "type");

            // Update some more counters with different IDs.
            logging.RequestReceived(request2);
            logging.ResponseStored(request2, response);
            logging.ResponseRetrieved(request2, response);

            // There should be 2 of each counter type.
            Assert.AreEqual(2, logging.ReceiveCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(2, logging.StoreCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(2, logging.RetrieveCounts.Count, "Wrong number of counts.");

            // Reset the request2.
            logging.ResetCount("id2");

            // There should now be only 1 of each counter type.
            Assert.AreEqual(1, logging.ReceiveCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(1, logging.StoreCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(1, logging.RetrieveCounts.Count, "Wrong number of counts.");

            // The counters should all be for the "id" request.
            Assert.AreEqual(1, logging.ReceiveCounts["id"], "Wrong 'id' count.");
            Assert.AreEqual(1, logging.StoreCounts["id"], "Wrong 'id' count.");
            Assert.AreEqual(1, logging.RetrieveCounts["id"], "Wrong 'id' count.");
        }

        /// <summary>
        /// Tests the ResetAll method to make sure it resets all counts, regardless of their IDs.
        /// </summary>
        [Test]
        public void TestResetAll()
        {
            // Update some counters.
            logging.RequestReceived(request);
            logging.ResponseStored(request, response);
            logging.ResponseRetrieved(request, response);

            Request request2 = new Request("id2", "key", "type");

            // Update some more counters with different IDs.
            logging.RequestReceived(request2);
            logging.ResponseStored(request2, response);
            logging.ResponseRetrieved(request2, response);

            // There should be 2 of each counter type.
            Assert.AreEqual(2, logging.ReceiveCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(2, logging.StoreCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(2, logging.RetrieveCounts.Count, "Wrong number of counts.");

            // Reset the request2.
            logging.ResetAll();

            // There should not be any more counters of either type.
            Assert.AreEqual(0, logging.ReceiveCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(0, logging.StoreCounts.Count, "Wrong number of counts.");
            Assert.AreEqual(0, logging.RetrieveCounts.Count, "Wrong number of counts.");
        }
    }
}
