/*
 * CacheResponsePersistenceTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Data;
using System.Collections;
using NUnit.Framework;
using TopCoder.Cache.Distributed;
using TopCoder.Data.ConnectionFactory;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// Tests of the <see cref="CachedResponsePersistence"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class CachedResponsePersistenceTests
    {
        /// <summary>
        /// The persistence object used in most of the tests.
        /// </summary>
        private CachedResponsePersistence persistence;

        /// <summary>
        /// The cache used to create the persistence.
        /// </summary>
        private DistributedSimpleCache persistenceCache = null;

        /// <summary>
        /// A nested persistence used for some tests.
        /// </summary>
        private DatabasePersistence nested = null;

        /// <summary>
        /// This is a test framework for IResponsePersistence implementations.
        /// </summary>
        private IResponsePersistenceTestsHelper responseHelper;

        /// <summary>
        /// Before all tests and after each one, the config manager is cleared and the
        /// cache is disposed of (if it exsists.)
        /// </summary>
        [TestFixtureSetUp, TearDown]
        public void RunBeforeAllTestsAndAfterEachTest()
        {
            TestHelper.Config.Clear(false);
            if (persistenceCache != null)
            {
                persistenceCache.Dispose();
            }
        }

        /// <summary>
        /// Load config files and create common instances fields used in the tests.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            // Load the Connection Factor config and create a database connection with it.
            TestHelper.Config.LoadFile("../../test_files/CF-standard.xml");
            IDbConnection conn = ConnectionManager.Instance.CreatePredefinedDbConnection("standard");

            // Create helper objects for creating the actual persistence.
            persistenceCache = new DistributedSimpleCache("http://persistentCache", null);
            Assert.AreEqual(0, persistenceCache.Keys.Count, "There are already keys in the cache.");
            nested = new DatabasePersistence(conn);

            // Create the persistence for testing, as well as a test framework to run the tests.
            persistence = new CachedResponsePersistence(persistenceCache, null);
            responseHelper = new IResponsePersistenceTestsHelper(persistence,
                new IResponsePersistenceTestsHelper.ReadResponses(this.ReadResponsesFromCache));
        }

        /// <summary>
        /// Calling the default constructor with a valid config should not throw any exceptions.
        /// </summary>
        [Test]
        public void TestCtor0()
        {
            // Load the config.
            TestHelper.Config.LoadFile("../../test_files/CachedResponsePersistenceTests.xml");

            // Create using the config and then dispose of it immediately.
            new CachedResponsePersistence().Dispose();
        }

        /// <summary>
        /// Calling the 2-arg constructor with a null cache should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor2NullArg()
        {
            new CachedResponsePersistence(null, nested);
        }

        /// <summary>
        /// Run a full test of the store operations with two elements.
        /// </summary>
        [Test]
        public void TestStoreTwoElements()
        {
            responseHelper.DoStoreTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Run a full test of the retrieve operation with two elements.
        /// </summary>
        [Test]
        public void TestRetrieveTwoElements()
        {
            responseHelper.DoRetrieveTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Run a full test of the delete operation with two elements.
        /// </summary>
        [Test]
        public void TestDeleteTwoElements()
        {
            responseHelper.DoDeleteTest(responseHelper.TwoElements, "id1");
        }

        /// <summary>
        /// Run a full test of the get identifiers method with two elements.
        /// </summary>
        [Test]
        public void TestGetIdentifiersTwoElements()
        {
            responseHelper.DoGetIdentifiersTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Run a full test of the clear method with two elements.
        /// </summary>
        [Test]
        public void TestClearTwoElements()
        {
            responseHelper.DoClearTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Helper callback used by the testing framework. It must fill the given dictionary with the
        /// "stored" (in this case cached) response objects.
        /// </summary>
        /// <param name="dict">The dictionary to fill with the stored objects.</param>
        private void ReadResponsesFromCache(IDictionary dict)
        {
            // The first level of the cache maps keys to dictionaries.
            foreach (string id in persistenceCache.Keys)
            {
                // FIXME: The DSC.Remove() method currently doesn't remove keys from the DSC.Keys set, once
                // this is handled properly, we shouldn't have to double-check for entry exsistence here.
                if (persistenceCache.Get(id) == null)
                {
                    continue;
                }

                // The second level (the dictionaries) map security keys to response objects.
                foreach (DictionaryEntry entry in (IDictionary) persistenceCache.Get(id))
                {
                    dict[new Request(id, (string) entry.Key)] = entry.Value;
                }
            }
        }
    }
}
