/*
 * DatabasePersistenceTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.IO;
using System.Data;
using System.Data.SqlClient;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using NUnit.Framework;
using TopCoder.Cache.Distributed;
using TopCoder.Util.ConfigurationManager;
using TopCoder.Data.ConnectionFactory;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// Tests of the <see cref="DatabasePersistence"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class DatabasePersistenceTests
    {
        /// <summary>
        /// A connection to the database.
        /// </summary>
        private IDbConnection conn;

        /// <summary>
        /// A dummy request object, used in the constructor tests.
        /// </summary>
        private Request request;

        /// <summary>
        /// A dummy response object, used in the constructor tests.
        /// </summary>
        private Response response;

        /// <summary>
        /// The persistence instance being tested in most of the methods.
        /// </summary>
        private DatabasePersistence persistence;

        /// <summary>
        /// The ICahcePersistence testing framework.
        /// </summary>
        private ICachePersistenceTestsHelper cacheHelper;

        /// <summary>
        /// The IResponsePersistence testing framework.
        /// </summary>
        private IResponsePersistenceTestsHelper responseHelper;
        
        /// <summary>
        /// Attempts to delete a table from the database, swollowing any errors.
        /// </summary>
        /// <param name="tableName">The name of the table to delete.</param>
        private void DeleteTable(string tableName)
        {
            try
            {
                // Perform the delete operation.
                conn.Open();
                IDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "DROP TABLE " + tableName;
                cmd.ExecuteNonQuery();
            }
            catch
            {
                // Ignore all errors.
            }
            finally
            {
                // Make sure the connection is closed.
                conn.Close();
            }
        }

        /// <summary>
        /// Before all tests and then after each test we clear the config and then load enough config
        /// data to be able to delete both database tables that the tests use.
        /// </summary>
        [TestFixtureSetUp, TearDown]
        public void RunBeforeAllTestsAndAfterEachTest()
        {
            TestHelper.Config.Clear(false);
            TestHelper.Config.LoadFile("../../test_files/CF-standard.xml");

            conn = ConnectionManager.Instance.CreatePredefinedDbConnection("standard");

            DeleteTable("response");
            DeleteTable("cache_image");
        }

        /// <summary>
        /// After all the tests are done, this makes sure the configuration is completely cleared after
        /// the databases have been dropped.
        /// </summary>
        [TestFixtureTearDown]
        public void RunAfterAllTests()
        {
            TestHelper.Config.Clear(false);
        }

        /// <summary>
        /// First, create the database tables, and then recreate all of the instance variables that
        /// will be used in the various tests.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            try
            {
                conn.Open();
                IDbCommand cmd = conn.CreateCommand();

                // Add the response table.
                cmd.CommandText = 
                    @"CREATE TABLE response (
                        identifier      VARCHAR(255) NOT NULL,  -- the request identifier
                        security_key    VARCHAR(255) NOT NULL,  -- the request security key
                        response_object IMAGE NOT NULL,         -- the response object in binary form
                        time_stamp      DATETIME NOT NULL,      -- the time at which the response is created
                        life_time       BIGINT NOT NULL         -- the intended life time of the response
                    )";
                cmd.ExecuteNonQuery();

                // Add the cache_image table.
                cmd.CommandText = 
                    @"CREATE TABLE cache_image (
                        identifier      VARCHAR(255) NOT NULL,  -- the request identifier
                        last_activity   DATETIME NOT NULL,      -- the last activity of any request with the identifier
                        in_storage      BIT NOT NULL            -- whether at least one response exists for the request
                    )";
                cmd.ExecuteNonQuery();

                // Create the test persistence and initialize the testing frameworks.
                persistence = new DatabasePersistence(conn);
                cacheHelper = new ICachePersistenceTestsHelper(persistence, 
                    new ICachePersistenceTestsHelper.ReadCache(this.ReadCacheFromDB));
                responseHelper = new IResponsePersistenceTestsHelper(persistence,
                    new IResponsePersistenceTestsHelper.ReadResponses(this.ReadResponsesFromDB));

                // Create some objects to test with.
                request = new Request("id", "key", "type");
                response = new Response("obj", new DateTime(2004, 1, 2, 3, 4, 5), new TimeSpan(1, 2, 3, 4));
            }
            finally
            {
                // Make sure the connection gets closed.
                conn.Close();
            }
        }

        /// <summary>
        /// Create two different persistences using the different constructors and make sure they can work
        /// with the same database to ensure their connections are being setup properly.
        /// </summary>
        [Test]
        public void TestCtor0()
        {
            // Load a config to use the 'standard' predefined connection and then create a persistance base on it.
            TestHelper.Config.LoadFile("../../test_files/DP-standardConnection.xml");
            DatabasePersistence defaultConstructed = new DatabasePersistence();

            // Store with one persistence and then retrieve with a manually constructed one.
            defaultConstructed.Store(request, response);
            Response fromDb = persistence.Retrieve(request);

            // Test the results.
            Assert.AreEqual(response.ResponseObject, fromDb.ResponseObject, "Wrong response object.");
            Assert.AreEqual(response.Timestamp, fromDb.Timestamp, "Wrong timestamp.");
            Assert.AreEqual(response.Life, fromDb.Life, "Wrong life time.");
        }

        /// <summary>
        /// Create two different persistences using the different constructors and make sure they can work
        /// with the same database to ensure their connections are being setup properly.
        /// </summary>
        [Test]
        public void TestCtor1()
        {
            DatabasePersistence constructed = new DatabasePersistence(conn);

            // Store with one persistence and then retrieve with a manually constructed one.
            constructed.Store(request, response);
            Response fromDb = persistence.Retrieve(request);

            // Test the results.
            Assert.AreEqual(response.ResponseObject, fromDb.ResponseObject, "Wrong response object.");
            Assert.AreEqual(response.Timestamp, fromDb.Timestamp, "Wrong timestamp.");
            Assert.AreEqual(response.Life, fromDb.Life, "Wrong life time.");
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestStoreTwoElements()
        {
            responseHelper.DoStoreTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestRetrieveTwoElements()
        {
            responseHelper.DoRetrieveTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestDeleteTwoElements()
        {
            responseHelper.DoDeleteTest(responseHelper.TwoElements, "id1");
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestGetIdentifiersTwoElements()
        {
            responseHelper.DoGetIdentifiersTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestClearTwoElements()
        {
            responseHelper.DoClearTest(responseHelper.TwoElements);
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestSaveTwoElements()
        {
            cacheHelper.DoSaveCacheTest(cacheHelper.TwoElements);
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestLoadTwoElements()
        {
            cacheHelper.DoLoadCacheTest(cacheHelper.TwoElements);
        }

        /// <summary>
        /// Reads all of the responses from the database storage and places them into the given dictionary.
        /// </summary>
        /// <param name="dict">The dictionary to place the responses into.</param>
        private void ReadResponsesFromDB(IDictionary dict)
        {
            try
            {
                conn.Open();

                IDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT * FROM response";
                
                using (IDataReader reader = cmd.ExecuteReader())
                {
                    // The response objects are stored in binary form.
                    IFormatter formatter = new BinaryFormatter();

                    while (reader.Read())
                    {
                        // Get all of the fields of this record.
                        string id = reader.GetString(0);
                        string key = reader.GetString(1);
                        byte[] bytes = (byte[]) reader.GetValue(2);
                        DateTime stamp = reader.GetDateTime(3);
                        TimeSpan life = TimeSpan.FromTicks(reader.GetInt64(4));

                        // Deserialize the response.
                        MemoryStream mem = new MemoryStream(bytes);
                        Response response = (Response) formatter.Deserialize(mem);

                        // Make sure the stamp and life fields match those of the response object.
                        Assert.AreEqual(response.Timestamp, stamp, "The timestamp does not match response's stamp.");
                        Assert.AreEqual(response.Life, life, "The life does not match response's lifetime.");

                        dict[new Request(id, key)] = response;
                    }
                }
            }
            finally
            {
                // Make sure the connection gets closed.
                conn.Close();
            }
        }

        /// <summary>
        /// Read the cache entries from the database storage into the given cache.
        /// </summary>
        /// <param name="cache">The cache to populate with the saved cache entries.</param>
        private void ReadCacheFromDB(IDistributedSimpleCache cache)
        {
            try
            {
                conn.Open();

                IDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT * FROM cache_image";
                
                using (IDataReader reader = cmd.ExecuteReader())
                {
                    // Iterate through all rows in the result set.
                    while (reader.Read())
                    {
                        // Extract the data from the current record.
                        string key = reader.GetString(0);
                        CacheEntry entry = new CacheEntry(reader.GetDateTime(1), reader.GetBoolean(2));

                        // If there is not already an entry for the key, add this entry.
                        if (cache.Get(key) == null)
                        {
                            cache.Put(key, entry);
                        }
                    }
                }
            }
            finally
            {
                // Make sure the connection gets closed.
                conn.Close();
            }
        }
    }
}
