/*
 * FileSystemCachePersistenceTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using NUnit.Framework;
using TopCoder.Cache.Distributed;
using TopCoder.Util.ConfigurationManager;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// Tests of the <see cref="FileSystemCachePersistence"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class FileSystemCachePersistenceTests
    {
        /// <summary>
        /// Default location to save/load the cache.
        /// </summary>
        private const string FileName = "../../test_files/cache.db";

        /// <summary>
        /// Instance of the class under test used throughout this fixture.
        /// </summary>
        private FileSystemCachePersistence persistence;

        /// <summary>
        /// An instance of the helper framework for testing ICachePersistence implementations.
        /// </summary>
        private ICachePersistenceTestsHelper helper;
        
        /// <summary>
        /// Before any test is run and then after each test, this method clears the ConfigManger and
        /// makes sure the default save file gets deleted.
        /// </summary>
        [TestFixtureSetUp, TearDown]
        public void RunBeforeAllTestsAndAfterEachTest()
        {
            TestHelper.Config.Clear(false);
            File.Delete(FileName);
        }

        /// <summary>
        /// Just to be safe, reinitialize the instance variables before each test.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            persistence = new FileSystemCachePersistence(FileName);
            helper = new ICachePersistenceTestsHelper(persistence, 
                new ICachePersistenceTestsHelper.ReadCache(this.ReadCacheFromFile));
        }

        /// <summary>
        /// Test the default constructor with a valid configuration file.
        /// </summary>
        [Test]
        public void TestCtor0()
        {
            // Load a config file with a filename for the FSCP.
            TestHelper.Config.LoadFile("../../test_files/FileSystemCachePersistenceTests.xml");
            string filename = TestHelper.Config.GetValue(typeof(FileSystemCachePersistence).FullName, "filename");

            // Make sure the file does not already exsist.
            Assert.IsFalse(File.Exists(filename), "The save file already exsists.");

            // Create a new instance using the default constructor.
            persistence = new FileSystemCachePersistence();

            try
            {
                // Saving a dummy cache should create the proper file.
                using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
                {
                    persistence.SaveCache(cache);
                }

                // Make sure the file was created.
                Assert.IsTrue(File.Exists(filename), "The save file was not created.");
            }
            finally
            {
                // Make sure the file gets removed, even if there is a problem with the test.
                File.Delete(filename);
            }
        }

        /// <summary>
        /// Test the 1-arg constructor with a null filename (should throw an ANE.)
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor1NullArg()
        {
            new FileSystemCachePersistence(null);
        }

        /// <summary>
        /// Test the 1-arg constructor with an empty filename (should throw an AE.)
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor1EmptyArg()
        {
            new FileSystemCachePersistence("");
        }

        /// <summary>
        /// Test the 1-arg constructor (as initialized in the [SetUp] method.)
        /// </summary>
        [Test]
        public void TestCtor1()
        {
            // Make sure the file does not already exsist.
            Assert.IsFalse(File.Exists(FileName), "The save file already exsists.");

            // Saving a dummy cache should create the proper file.
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                persistence.SaveCache(cache);
            }

            // Make sure the file was created.
            Assert.IsTrue(File.Exists(FileName), "The save file was not created.");
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestSaveTwoElement()
        {
            helper.DoSaveCacheTest(helper.TwoElements);
        }

        /// <summary>
        /// Peform a full test with two elements.
        /// </summary>
        [Test]
        public void TestLoadTwoElement()
        {
            helper.DoLoadCacheTest(helper.TwoElements);
        }

        /// <summary>
        /// Deserializes the saved objects from the file into the given cache for use in the testing framework.
        /// </summary>
        /// <param name="cache">The cache to add saved entries to.</param>
        private void ReadCacheFromFile(IDistributedSimpleCache cache)
        {
            // Make sure the destination file exsists now.
            FileInfo file = new FileInfo(FileName);
            Assert.IsTrue(file.Exists, "The file does not exsist.");

            // Get a read-only, locked (only other readers allowed) FileStream and deserialize its contents.
            using (Stream stream = file.Open(FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                // The saved data should be in binary format.
                IFormatter formatter = new BinaryFormatter();

                // While there is more in the file, keep deserializing.
                while (stream.Position < stream.Length)
                {
                    // Deserialize the key and value of one entry.
                    string key = (string) formatter.Deserialize(stream);
                    CacheEntry value = (CacheEntry) formatter.Deserialize(stream);

                    // Add the entry to the cache.
                    cache.Put(key, value);
                }
            }
        }
    }
}
