/*
 * ICachePersistenceTestsHelper.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using System.Diagnostics;
using NUnit.Framework;
using TopCoder.Cache.Distributed;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// A set of common testing methods which can be used to easily test any implementation of the
    /// ICachePersistence interface.
    /// </summary>
    /// <author>TCSDEVELOEPER</author>
    /// <version>1.0</version>
    public class ICachePersistenceTestsHelper
    {
        /// <summary>
        /// A delegate that can read the implementation's specific storage format and return the results
        /// as cache items (required so that Load and Save can be tested independently.) 
        /// </summary>
        public delegate void ReadCache(IDistributedSimpleCache cache);

        /// <summary>
        /// The callback used to read the currently stored data.
        /// </summary>
        private ReadCache readCache;

        /// <summary>
        /// The persistence object that is currently being tested.
        /// </summary>
        private ICachePersistence persistence;

        /// <summary>
        /// Creates a new testing framework for the given persistence.
        /// </summary>
        /// <param name="persistence">The persistence to test.</param>
        /// <param name="callback">The callback used to load saved cache entries.</param>
        public ICachePersistenceTestsHelper(ICachePersistence persistence, ReadCache callback)
        {
            // Check the parameters.
            Debug.Assert(persistence != null, "The persistence cannot be null.");
            Debug.Assert(callback != null, "The callback cannot be null.");

            // Set the instance fields.
            this.persistence = persistence;
            this.readCache = callback;
        }

        /// <summary>
        /// Creates and returns a test case with two entries.
        /// </summary>
        public IDictionary TwoElements
        {
            get
            {
                // Create a new dictionary.
                IDictionary dict = new Hashtable();

                // Add the elements.
                dict["key1"] = new CacheEntry(new DateTime(2004, 1, 2, 3, 4, 5), true);
                dict["key2"] = new CacheEntry(new DateTime(2004, 5, 4, 3, 2, 1), false);

                return dict;
            }
        }

        /// <summary>
        /// Tests the SaveCache method by saving the items in the given dictionary and then using the
        /// readCache callback to read them back out and compare the results.
        /// </summary>
        /// <param name="dict">The dictionary of entries to test with.</param>
        public void DoSaveCacheTest(IDictionary dict)
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                // Double-check that there the cache is initally empty.
                Assert.AreEqual(0, cache.Keys.Count, "There are already keys in the cache.");

                // Copy the dictionary entries into the cache.
                foreach (DictionaryEntry entry in dict)
                {
                    cache.Put(entry.Key, entry.Value);
                }

                // Save the cache to disk.
                persistence.SaveCache(cache);
            }

            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                // Double-check that there the cache is initally empty.
                Assert.AreEqual(0, cache.Keys.Count, "There are already keys in the cache.");

                readCache(cache);

                // Verify the contents of the cache against the inital dictionary.
                VerifyCacheContents(dict, cache);
            }
        }

        /// <summary>
        /// Tests the LoadCache method by saving the given set of entries and then loading them back in
        /// (this assumes that the SaveCache method is already working properly.)
        /// </summary>
        /// <param name="dict">The dictionary of entries to test.</param>
        public void DoLoadCacheTest(IDictionary dict)
        {
            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                // Double-check that there the cache is initally empty.
                Assert.AreEqual(0, cache.Keys.Count, "There are already keys in the cache.");

                // Copy the dictionary entries into the cache.
                foreach (DictionaryEntry entry in dict)
                {
                    cache.Put(entry.Key, entry.Value);
                }

                // Save the cache to disk.
                persistence.SaveCache(cache);
            }

            using (DistributedSimpleCache cache = new DistributedSimpleCache("http://cache", null))
            {
                // Double-check that there the cache is initally empty.
                Assert.AreEqual(0, cache.Keys.Count, "There are already keys in the cache.");

                // Load the cache from the file.
                persistence.LoadCache(cache);

                // Verify the contents of the cache against the inital dictionary.
                VerifyCacheContents(dict, cache);
            }
        }

        /// <summary>
        /// Double-checks that the results stored in the cache are equivalent to the original dictionary entries.
        /// </summary>
        /// <param name="dict">The origiinal (expected) set of entries.</param>
        /// <param name="cache">The actual entries (test results) that should be the same as the originals.</param>
        private void VerifyCacheContents(IDictionary dict, IDistributedSimpleCache cache)
        {
            // Make sure there are the right number of entries in the cache.
            Assert.AreEqual(dict.Count, cache.Keys.Count, "Wrong number of entries in the cache.");

            // Check all the elements of the cache.
            foreach (DictionaryEntry entry in dict)
            {
                // Make sure this element is found in the cache.
                CacheEntry cacheEntry = (CacheEntry) cache.Get(entry.Key);
                Assert.IsNotNull(cacheEntry, "The entry for '" + entry.Key + "' was not loaded.");

                // Make sure the entry for this key has the correct values.
                CacheEntry dictEntry = (CacheEntry) entry.Value;
                Assert.AreEqual(dictEntry.LastActivity, cacheEntry.LastActivity, 
                    "Wrong LastActivity value for '" + entry.Key + "'.");
                Assert.AreEqual(dictEntry.IsInStorage, cacheEntry.IsInStorage, 
                    "Wrong IsInStorage value for '" + entry.Key + "'.");
            }
        }
    }
}
