/*
 * IResponsePersistenceTestsHelper.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved
 */
using System;
using System.Collections;
using System.Diagnostics;
using NUnit.Framework;

namespace TopCoder.Util.Idempotency.Persistence
{
    /// <summary>
    /// Summary description for IResponsePersistenceTestsHelper.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    public class IResponsePersistenceTestsHelper
    {
        /// <summary>
        /// A delegate that can read the implementation's specific storage format and return the results
        /// as dictionary entries (required so that Store and Retrieve can be tested independently.) 
        /// </summary>
        public delegate void ReadResponses(IDictionary dict);

        /// <summary>
        /// The callback used to read the currently stored data.
        /// </summary>
        private ReadResponses readResponses;

        /// <summary>
        /// The persistence object that is currently being tested.
        /// </summary>
        private IResponsePersistence persistence;

        /// <summary>
        /// Creates a new testing framework for the given persistence.
        /// </summary>
        /// <param name="persistence">The persistence to test.</param>
        /// <param name="callback">The callback used to retrieve stored responses.</param>
        public IResponsePersistenceTestsHelper(IResponsePersistence persistence, ReadResponses callback)
        {
            // Check the arguments.
            Debug.Assert(persistence != null, "The persistence cannot be null.");
            Debug.Assert(callback != null, "The callback cannot be null.");

            // Set the instance variables.
            this.persistence = persistence;
            this.readResponses = callback;
        }

        /// <summary>
        /// Creates and returns a test case with two entries.
        /// </summary>
        public IDictionary TwoElements
        {
            get
            {
                // Creat a new dictionary.
                IDictionary dict = new Hashtable();

                // Add the elements.
                dict[new Request("id1", "key1", "type1")] = 
                    new Response("obj", new DateTime(2004, 1, 2, 3, 4, 5), new TimeSpan(1, 2, 3, 4));
                dict[new Request("id2", "key2", "type2")] = 
                    new Response("obj", new DateTime(2004, 5, 4, 3, 3, 1), new TimeSpan(4, 3, 2, 1));

                return dict;
            }
        }

        /// <summary>
        /// Tests the Store method by storing a set of responses and then reading them back in to
        /// verify that the correct data was stored.
        /// </summary>
        /// <param name="dict">The responses to store.</param>
        public void DoStoreTest(IDictionary dict)
        {
            // Add all of the entries to the persistent storage.
            StoreAll(dict);

            // Get the responses that are currently in the persistence.
            IDictionary actual = new Hashtable();
            readResponses(actual);

            // Verify the initial contents against the actual contents.
            VerifyPersistenceContents(dict, actual);
        }

        /// <summary>
        /// Tests the Retrieve method by storing a set of responses and then reading them back in to
        /// verify that the correct data was stored.
        /// </summary>
        /// <param name="dict">The responses to store.</param>
        public void DoRetrieveTest(IDictionary dict)
        {
            // Add all of the entries to the persistent storage.
            StoreAll(dict);

            // Retrieve all of the responses.
            IDictionary actual = new Hashtable();
            foreach (DictionaryEntry entry in dict)
            {
                actual[entry.Key] = persistence.Retrieve((Request) entry.Key);
            }

            // Verify the initial contents against the actual contents.
            VerifyPersistenceContents(dict, actual);
        }

        /// <summary>
        /// Loads a set of responses and then deletes an identifer and verifies that only the correct
        /// items were deleted.
        /// </summary>
        /// <param name="dict">The responses to load before the delete operation.</param>
        /// <param name="identifier">The identifier to delete.</param>
        public void DoDeleteTest(IDictionary dict, string identifier)
        {
            // Add all of the entries to the persistent storage.
            StoreAll(dict);

            // Delete the responses.
            persistence.Delete(identifier);

            // Create the expected table of results.
            IDictionary expected = new Hashtable();
            foreach (DictionaryEntry entry in dict)
            {
                // Copy over any entries that do not have the identifier that was deleted.
                if (((Request) entry.Key).Identifier != identifier)
                {
                    expected[entry.Key] = entry.Value;
                }
            }

            // Get the responses that are currently in the persistence.
            IDictionary actual = new Hashtable();
            readResponses(actual);

            // Verify the expected contents against the actual contents.
            VerifyPersistenceContents(expected, actual);
        }

        /// <summary>
        /// Loads a set of responses and then gets the set of identifiers for those entries and
        /// ensures that only the correct set of identifiers are returned.
        /// </summary>
        /// <param name="dict">The set of responses to store.</param>
        public void DoGetIdentifiersTest(IDictionary dict)
        {
            // Add all of the entries to the persistent storage.
            StoreAll(dict);

            // Create an identity dictionary (set) of the expected IDs.
            IDictionary expected = new Hashtable();
            foreach (Request request in dict.Keys)
            {
                expected[request.Identifier] = request.Identifier;
            }

            // Get the identifiers.
            IEnumerator it = persistence.GetIdentifiers();

            // Create an identity dictionary (set) of the actual IDs.
            IDictionary actual = new Hashtable();
            while (it.MoveNext())
            {
                actual[it.Current] = it.Current;
            }

            // Make sure the two sets of IDs are the same.
            Assert.AreEqual(expected.Count, actual.Count, "The ID set is not the right size.");

            foreach (string id in expected.Keys)
            {
                Assert.IsTrue(actual.Contains(id), "The ID '" + id + "' is missing from the actual results.");
            }
        }

        /// <summary>
        /// Adds a set of responses and then clears then, ensuring that all entries are cleared.
        /// </summary>
        /// <param name="dict">The responses to store before clearing.</param>
        public void DoClearTest(IDictionary dict)
        {
            // Add all of the entries to the persistent storage.
            StoreAll(dict);

            // Clear the persistence.
            persistence.Clear();

            // Get the responses that are currently in the persistence.
            IDictionary actual = new Hashtable();
            readResponses(actual);

            // Verify the expected contents (empty) against the actual contents.
            VerifyPersistenceContents(new Hashtable(), actual);
        }

        /// <summary>
        /// Helper method which simply stores all of the request/response entries in the dictionary.
        /// </summary>
        /// <param name="dict">The entries to store.</param>
        private void StoreAll(IDictionary dict)
        {
            // Add all of the entries to the persistent storage.
            foreach (DictionaryEntry entry in dict)
            {
                persistence.Store((Request) entry.Key, (Response) entry.Value);
            }

        }

        /// <summary>
        /// Verifies that the set of expected entries is equivalent to the actual entries.
        /// This process is slightly complicated by the fact that the objects cannot be directly compared
        /// since they may have been serialized and then deserialized during storage (and neither the
        /// Request nor the Response classes override Equals.)
        /// </summary>
        /// <param name="expected">The set of expected request/response entries.</param>
        /// <param name="actual">The set of actual entries which resulted from the test operation.</param>
        private void VerifyPersistenceContents(IDictionary expected, IDictionary actual)
        {
            // Make sure there are the right number of entries in the cache.
            Assert.AreEqual(expected.Count, actual.Count, "Wrong number of actual entries.");

            // Requests are not directly comparable, so we must build a special lookup table to map them.
            IDictionary lookup = new Hashtable();

            // Prevent errors by making sure all expected IDs are in the lookup table.
            foreach (Request request in expected.Keys)
            {
                if (lookup[request.Identifier] == null)
                {
                    lookup[request.Identifier] = new Hashtable();
                }
            }

            // Prevent errors by making sure all actual IDs are in the lookup table.
            foreach (Request request in actual.Keys)
            {
                if (lookup[request.Identifier] == null)
                {
                    lookup[request.Identifier] = new Hashtable();
                }

                // Identity map the request objects so they can be found by their (ID, SecurityKey) values.
                IDictionary innerDict = (IDictionary) lookup[request.Identifier];
                innerDict[request.SecurityKey] = request;
            }

            // Check for all the expected elements.
            foreach (DictionaryEntry entry in expected)
            {
                // Get the expected and actual requests.
                Request expectedRequest = (Request) entry.Key;
                IDictionary innerDict = (IDictionary) lookup[expectedRequest.Identifier];
                Request actualRequest = (Request) innerDict[expectedRequest.SecurityKey];

                // This string helps make error messages more meaningful, but keeps the code a bit shorter.
                string key = "('" + expectedRequest.Identifier + "', '" + expectedRequest.SecurityKey + "')";

                // Make sure the actual request was found.
                Assert.IsNotNull(actualRequest, "The entry for " + key + " was not found.");

                // NOTE: The request objects are not expected to match beyond the ID and SecurityKey since 
                // they may have been created when from just those files stored in the persistent storage.

                // Get the expected and actual responses.
                Response expectedResponse = (Response) entry.Value;
                Response actualResponse = (Response) actual[actualRequest];

                // Make sure the response objects are equal.
                Assert.AreEqual(expectedResponse.ResponseObject, actualResponse.ResponseObject, 
                    "Wrong response object for " + key + ".");
                Assert.AreEqual(expectedResponse.Timestamp, actualResponse.Timestamp,
                    "Wrong timestamp for " + key + ".");
                Assert.AreEqual(expectedResponse.Life, actualResponse.Life, "Wrong lifetime for " + key + ".");
            }
        }
    }
}
