/*
 * ResponseTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using NUnit.Framework;

namespace TopCoder.Util.Idempotency
{
    /// <summary>
    /// Tests of the <see cref="Response"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class ResponseTests
    {
        /// <summary>
        /// Calling the 1-arg constructor with a null response object should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor1NullArg()
        {
            new Response(null);
        }

        /// <summary>
        /// Calling the 1-arg constructor with a valid parameter should setup the response object properly.
        /// </summary>
        [Test]
        public void TestCtor1()
        {
            Response r = new Response("foo");

            Assert.AreEqual("foo", r.ResponseObject, "Wrong response object set.");
            Assert.IsTrue(DateTime.Now - r.Timestamp < TimeSpan.FromMinutes(1), "Wrong timestamp.");
            Assert.AreEqual(TimeSpan.MaxValue, r.Life, "Wrong lifetime.");
        }

        /// <summary>
        /// Calling the 2-arg constructor with a null response object should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor2NullArg()
        {
            new Response(null, TimeSpan.Zero);
        }

        /// <summary>
        /// Calling the 2-arg constructor with a negative life should throw an AOORE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestCtor2NegSpan()
        {
            new Response("foo", TimeSpan.MinValue);
        }

        /// <summary>
        /// Calling the 2-arg constructor with valid parameters should setup the response object properly.
        /// </summary>
        [Test]
        public void TestCtor2()
        {
            Response r = new Response("foo", TimeSpan.Zero);

            Assert.AreEqual("foo", r.ResponseObject, "Wrong response object set.");
            Assert.IsTrue(DateTime.Now - r.Timestamp < TimeSpan.FromMinutes(1), "Wrong timestamp.");
            Assert.AreEqual(TimeSpan.Zero, r.Life, "Wrong lifetime.");
        }

        /// <summary>
        /// Calling the 3-arg constructor with a null response object should throw an ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor3NullArg()
        {
            new Response(null, DateTime.Now, TimeSpan.Zero);
        }

        /// <summary>
        /// Calling the 3-arg constructor with a negative life should throw an AOORE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void TestCtor3NegSpan()
        {
            new Response("foo", DateTime.Now, TimeSpan.MinValue);
        }

        /// <summary>
        /// Calling the 3-arg constructor with valid parameters should setup the response object properly.
        /// </summary>
        [Test]
        public void TestCtor3()
        {
            DateTime time = DateTime.Now;
            Response r = new Response("foo", time, TimeSpan.Zero);

            Assert.AreEqual("foo", r.ResponseObject, "Wrong response object set.");
            Assert.AreEqual(time, r.Timestamp, "Wrong timestamp.");
            Assert.AreEqual(TimeSpan.Zero, r.Life, "Wrong lifetime.");
        }
    }
}
