/*
 * TypeMappingCacheStrategyTests.cs
 *
 * Copyright (c) 2004, TopCoder, Inc. All rights reserved.
 */
using System;
using System.Collections;
using NUnit.Framework;

namespace TopCoder.Util.Idempotency.Strategy
{
    /// <summary>
    /// Tests of the <see cref="TypeMappingCacheStrategy"/> class.
    /// </summary>
    /// <author>TCSDEVELOPER</author>
    /// <version>1.0</version>
    [TestFixture]
    public class TypeMappingCacheStrategyTests
    {
        /// <summary>
        /// The instance being tested.
        /// </summary>
        private TypeMappingCacheStrategy strategy;

        /// <summary>
        /// A set of decisions used by the constructor.
        /// </summary>
        private IDictionary decisions;

        /// <summary>
        /// A dummy request used to test the Accept method.
        /// </summary>
        private Request request;

        /// <summary>
        /// Recreate the instance variables for each test.
        /// </summary>
        [SetUp]
        public void RunBeforeEachTest()
        {
            decisions = new Hashtable();

            request = new Request("id", "key", "type");
        }

        /// <summary>
        /// Calling the constructor with a null dictionary should throw ANE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentNullException))]
        public void TestCtor2NullDecisions()
        {
            new TypeMappingCacheStrategy(null, CacheDecision.Accept);
        }

        /// <summary>
        /// Calling the constructor with a default decision of Default should throw AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TextCtor2DefaultDefaultDecision()
        {
            new TypeMappingCacheStrategy(decisions, CacheDecision.Default);
        }

        /// <summary>
        /// Calling the constructor with a non-defined enumeration value should throw AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2OtherDefaultDecision()
        {
            new TypeMappingCacheStrategy(decisions, unchecked((CacheDecision) (-1)));
        }

        /// <summary>
        /// Calling the constructor with a dictionary that contains an empty key should throw AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2EmptyKey()
        {
            decisions[string.Empty] = "value";

            new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);
        }

        /// <summary>
        /// Calling the constructor with a dictionary that contains a non-string key should throw an AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2NonStringKey()
        {
            decisions[5] = "value";

            new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);
        }

        /// <summary>
        /// Calling the constructor with a dictionary that contains a null value should throw an AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2NullValue()
        {
            decisions["type"] = null;

            new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);
        }

        /// <summary>
        /// Calling the constructor with with a dictionary that contains a non-CacheDecision value
        /// should throw an AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2NonCacheDecisionValue()
        {
            decisions["type"] = "value";

            new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);
        }

        /// <summary>
        /// Calling the constructor with a dictionary that contains a non-defined CacheDecision entry
        /// should throw an AE.
        /// </summary>
        [Test, ExpectedException(typeof(ArgumentException))]
        public void TestCtor2NonDefinedCacheDecisionValue()
        {
            decisions["type"] = unchecked((CacheDecision) (-1));

            new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);
        }

        /// <summary>
        /// An empty decision table and default of Accept should accept everything.
        /// </summary>
        [Test]
        public void TestCtor2DefaultDecisionAccept()
        {
            strategy = new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);

            Assert.IsTrue(strategy.Accept(request));
        }

        /// <summary>
        /// An empty decision table and default of Deny should deny everything.
        /// </summary>
        [Test]
        public void TestCtor2DefaultDecisionDeny()
        {
            strategy = new TypeMappingCacheStrategy(decisions, CacheDecision.Deny);

            Assert.IsFalse(strategy.Accept(request));
        }

        /// <summary>
        /// A decision table with an entry for this request type should allow it even with a default of Deny.
        /// </summary>
        [Test]
        public void TestCtor2Accept()
        {
            decisions[request.Type] = CacheDecision.Accept;

            strategy = new TypeMappingCacheStrategy(decisions, CacheDecision.Deny);

            Assert.IsTrue(strategy.Accept(request));
        }

        /// <summary>
        /// A decision table with an entry for this type should Deny it even with a default of Accept.
        /// </summary>
        [Test]
        public void TestCtor2Deny()
        {
            decisions[request.Type] = CacheDecision.Deny;

            strategy = new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);

            Assert.IsFalse(strategy.Accept(request));
        }

        /// <summary>
        /// A decision table with Default values should fall-through to the default value (in this case, Accept.)
        /// </summary>
        [Test]
        public void TestCtor2DefaultAccept()
        {
            decisions[request.Type] = CacheDecision.Default;

            strategy = new TypeMappingCacheStrategy(decisions, CacheDecision.Accept);

            Assert.IsTrue(strategy.Accept(request));
        }

        /// <summary>
        /// A decision table with Default values should fall-through to the default value (in this case, Deny.)
        /// </summary>
        [Test]
        public void TestCtor2DefaultDeny()
        {
            decisions[request.Type] = CacheDecision.Default;

            strategy = new TypeMappingCacheStrategy(decisions, CacheDecision.Deny);

            Assert.IsFalse(strategy.Accept(request));
        }
    }
}
